#!/usr/bin/env python3

"""
Nemesida WAF API
Copyright (c) 2017-2025 Pentestit LLC - All Rights Reserved
Author: Romanov R.
"""

import json

from base64 import b64decode
from datetime import datetime
from hashlib import md5
from time import time
from time import time_ns

from core import get_geo
from core import init_db_connection, close_db_connection
from core import memc
from core import memc_lock

from logger import log


def get_antibot_v1(waf_id):
    try:

        # init
        response = {'status': 'success', 'data': {}}

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # init query
        query = """
            SELECT
                ad.id, ad.timestamp, ad.ip, ua.data, fp.data
            FROM
                antibot_data
            AS
                ad
            INNER JOIN
                antibot_ua AS ua ON ad.ua = ua.id
            INNER JOIN
                antibot_fp AS fp ON ad.fp = fp.id
            WHERE (waf_id = %s)
            """
        cursor.execute(query, (waf_id,))

        # receive data from the database
        res = cursor.fetchall()
        close_db_connection(connect, cursor)

        # result processing
        for item in res:

            # update the response
            response['data'][item[0]] = {
                'timestamp': item[1],
                'ip': item[2],
                'ua': item[3],
                'fp': json.loads(item[4])
            }

        # del vars
        del res

        # response
        return response

    except Exception as e:
        log.error('An error occurred in get_antibot: {}'.format(e))
        return {'status': 'fail', 'description': str(e)}


def set_antibot_v1(data, waf_id):
    try:

        # init
        ts = int(time())
        memc_prefix = 'antibot_'
        data = data.get('data', [])

        # skip empty
        if not data:
            return {'status': 'fail', 'description': 'Empty or missing column: data'}

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # init processed list
        ua_processed = set()
        fp_processed = set()

        # data normalization
        for idx, item in enumerate(data.copy()):

            # updating data in the database
            for k in ['ua', 'fp']:

                # init
                v = data[idx].get(k)

                # skip empty
                if (not v) or (not isinstance(v, str)):
                    log.error('An error occurred while processing antibot data: incorrect data for {} ({})'.format(k, v))
                    del data[idx]
                    break

                # decode the data
                if k == 'ua':

                    # B64 decode
                    data[idx][k] = b64decode(v).decode('UTF-8', "ignore")

                else:

                    # removing \n and specific columns
                    v = json.loads(b64decode(v).decode('UTF-8', "ignore"))
                    if 'ts' in v:
                        del v['ts']

                    # convert to JSON
                    data[idx][k] = json.dumps(v)

                # del vars
                del v

        # processing the dbr
        for item in data:

            # updating data in the database
            for k in ['ua', 'fp']:

                # init
                v = item.get(k)
                is_processed = False
                processed = ua_processed if k == 'ua' else fp_processed

                # get the hash
                hsh = md5(v.encode('UTF-8')).hexdigest()

                # skip processed
                if hsh in processed:
                    continue

                # search in the memcached
                with memc_lock:
                    try:
                        r = memc.get(memc_prefix + hsh)
                        is_processed = True if r else False
                        del r
                    except Exception as e:
                        log.error('An error occurred while receiving data from the memcached: {}'.format(e))

                # skip if processed
                if is_processed:
                    processed.add(hsh)
                    continue

                # search data in the database
                query = """ SELECT id FROM antibot_{} WHERE id = %s """.format(k)
                cursor.execute(query, (hsh,))
                res = cursor.fetchone()

                # result processing
                if res and res[0]:

                    # update data in the memcached
                    try:
                        memc.set(memc_prefix + hsh, '-', 86400)  # 84600 - 1 day
                    except Exception as e:
                        log.error('An error occurred while placing data into the memcached: {}'.format(e))

                    # update processed list
                    processed.add(hsh)
                    continue

                # store data into the database
                query = """ INSERT INTO antibot_{} (id, timestamp, data) VALUES (%s, %s, %s) ON CONFLICT DO NOTHING """.format(k)
                cursor.execute(query, (hsh, ts, v))

                # update data in the memcached
                try:
                    memc.set(memc_prefix + hsh, '-', 86400)  # 84600 - 1 day
                except Exception as e:
                    log.error('An error occurred while placing data into the memcached: {}'.format(e))

                # update processed list
                processed.add(v)

        # del vars
        del ua_processed, fp_processed

        # prepare the data
        dbr = [
            (
                time_ns(), ts, waf_id, x['ip'], md5(x['ua'].encode('UTF-8')).hexdigest(), md5(x['fp'].encode('UTF-8')).hexdigest(),
            ) for x in data
        ]
        ar = ','.join(cursor.mogrify('(%s, %s, %s, %s, %s, %s)', i).decode('UTF-8') for i in dbr)
        del data, dbr

        # store data into the database
        query = """ INSERT INTO antibot_data (id, timestamp, waf_id, ip, ua, fp) VALUES %s ON CONFLICT DO NOTHING """
        cursor.execute(query % (ar,))
        del ar

        # close database connection
        close_db_connection(connect, cursor)

        # response
        return {'status': 'success'}

    except Exception as e:
        log.error('An error occurred in set_antibot: {}'.format(e))
        return {'status': 'fail', 'description': str(e)}


def del_antibot_v1(data, waf_id):
    try:

        # init
        ids = data.get('id')

        # skip empty
        if not ids:
            return {'status': 'fail', 'description': 'Empty or missing column: id'}

        # delete the data
        connect = init_db_connection()
        cursor = connect.cursor()
        query = """ DELETE FROM antibot_data WHERE (waf_id = %s AND id IN %s) RETURNING id """
        cursor.execute(query, (waf_id, tuple(ids),))
        res = cursor.fetchall()
        close_db_connection(connect, cursor)

        # processing result
        response = {'status': 'success', 'deleted': len(res)}

        # response
        return response

    except Exception as e:
        log.error('An error occurred in del_antibot: {}'.format(e))
        return {'status': 'fail', 'description': str(e)}


def get_antibot_summary_v1(data, waf_id):
    try:

        # init
        group_by_cols = ['ua', 'fp']
        response = {'status': 'success', 'data': []}

        # grouping processing
        group_by = data.get('group_by').lower() if isinstance(data.get('group_by'), str) else None
        if (not group_by) or (group_by not in group_by_cols):
            group_by = None

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # receive the data according to WAF ID
        query = """
            SELECT
                ad.id, ad.timestamp, ad.ip, ad.ua, ua.data, ad.fp, fp.data
            FROM
                antibot_data
            AS
                ad
            INNER JOIN
                antibot_ua AS ua ON ad.ua = ua.id
            INNER JOIN
                antibot_fp AS fp ON ad.fp = fp.id
            WHERE (waf_id = %s)
        """
        cursor.execute(query, (waf_id,))

        # result processing
        antibot_data = [
            {
                'id': x[0],
                'ts': x[1],
                'ip': x[2],
                'ua': x[3],
                'ud': x[4],
                'fp': x[5],
                'fd': x[6]
            } for x in cursor.fetchall()
        ]

        # close database connection
        close_db_connection(connect, cursor)

        # skip empty
        if not antibot_data:
            return response

        # receive the Geo data
        ips = set(x['ip'] for x in antibot_data)
        ipsg = get_geo(ips, waf_id)
        del ips

        # receive unique combination of UA/FP
        ucomb = set((x['ua'], x['fp'],) if not group_by else x[group_by] for x in antibot_data)

        # receive list of IPs for each unique combination
        for item in ucomb:

            # convert item to dict if not group_by
            if not group_by:
                item = {'ua': item[0], 'fp': item[1]}

            # init the result
            if not group_by:
                result = {'ua': '', 'fp': '', 'ips': []}
            else:
                result = {
                    **{group_by: ''},
                    **{x: [] for x in [x for x in group_by_cols if x != group_by]},
                    **{'ips': []}
                }

            # result processing
            if not group_by:

                # update the FP data in result
                for i in antibot_data:
                    t = 'fp'
                    if i[t] == item[t]:
                        result[t] = json.loads(i[t[:-1] + 'd'])
                        break

                # update the UA data in result
                for i in antibot_data:
                    t = 'ua'
                    if i[t] == item[t]:
                        result[t] = i[t[:-1] + 'd']
                        break

            else:

                # update group_by column data
                for i in antibot_data:
                    if i[group_by] == item:
                        result[group_by] = json.loads(i[group_by[:-1] + 'd']) if group_by == 'fp' else i[group_by[:-1] + 'd']
                        break

                # update other data
                for col in group_by_cols:

                    # skip group_by
                    if col == group_by:
                        continue

                    # init
                    tc = col[:-1] + 'd'

                    # update the result
                    result[col] = [json.loads(x[tc]) if col == 'fp' else x[tc] for x in antibot_data if x[group_by] == item]
                    
                    # sort the result
                    if col != 'fp':
                        result[col] = sorted(set(result[col]))

            # receive the IPs with the timestamp
            if not group_by:
                ipsd = set((x['ip'], x['ts'],) for x in antibot_data if (x['ua'] == item['ua'] and x['fp'] == item['fp']))
            else:
                ipsd = set((x['ip'], x['ts'],) for x in antibot_data if x[group_by] == item)

            # init
            result['ips'] = {}
            ips = set(x[0] for x in ipsd)

            # result processing
            for ip in ips:

                # update the result
                result['ips'][ip] = {
                    'cc': ipsg[ip].get('cc'),
                    'bl': ipsg[ip].get('feed', []),
                    'privacy': [
                        x for x in ipsg[ip] if x in ['mobile', 'hosting', 'proxy', 'tor', 'crawler'] and ipsg[ip].get(x)
                    ],
                    'date': [
                        datetime.fromtimestamp(x[1]).strftime('%Y-%m-%d %H:%M:%S') for x in ipsd if x[0] == ip
                    ]
                }

            # update the response
            response['data'].append(result)
            del result, item, ipsd, ips

        # sort data by IPs count
        response['data'] = sorted(response['data'], key=lambda x: len(x['ips']), reverse=True)

        # del vars
        del antibot_data

        # response
        return response

    except Exception as e:
        log.error('An error occurred in get_antibot_summary_v1: {}'.format(e))
        return {'status': 'fail', 'description': str(e)}
