#!/usr/bin/env python3

"""
Nemesida WAF API
Copyright (c) 2017-2025 Pentestit LLC - All Rights Reserved
Author: Romanov R.
"""

import json

from time import time
from time import time_ns

from core import init_db_connection, close_db_connection
from logger import log


def get_agent_monitoring_v1(waf_id):
    try:

        # init
        ts = int(time())
        max_inactive_time = 60               # 1 min.
        max_nws_inactive_time = 60 * 60 * 3  # 3 h.
        response = {'status': 'success', 'agents': []}

        # receive data from the database
        query = """ SELECT id, agent, ip, uuid, last_activity, systemMonitor FROM agent_monitoring WHERE (waf_id = %s) """
        connect = init_db_connection()
        cursor = connect.cursor()
        cursor.execute(query, (waf_id,))
        res = cursor.fetchall()
        close_db_connection(connect, cursor)

        # result processing
        for item in res:

            # extract the data
            agent = item[1]
            last_activity = item[4]
            sysmon = item[5]

            # active processing
            ttl = max_nws_inactive_time if agent == 'nws' else max_inactive_time
            active = True if last_activity > ts - ttl else False

            # SysMon processing
            sysmon = json.loads(sysmon) if sysmon else None

            # result processing
            response['agents'].append({
                'id': item[0],
                'active': active,
                'agent': agent,
                'ip': item[2],
                'uuid': item[3],
                'lastActivity': last_activity,
                'systemMonitor': sysmon
            })

        # response
        return response

    except Exception as e:
        log.error('An error occurred in get_agent_monitoring: {}'.format(e))
        return {'status': 'fail', 'description': str(e)}


def set_agent_monitoring_v1(agent, ip, uuid, sysmon, waf_id):
    try:

        # init
        rid = time_ns()
        ts = int(time())

        # sysmon normalization
        sysmon = json.dumps(sysmon)

        # update data in the database
        query = """
            INSERT INTO agent_monitoring (
                id, waf_id, agent, ip, uuid, last_activity, systemMonitor
            ) VALUES (
                %s, %s, %s, %s, %s, %s, %s
            )
            ON CONFLICT (
                waf_id, agent, uuid
            ) DO UPDATE SET
                (ip, last_activity, systemMonitor) = (%s, %s, %s)
        """
        connect = init_db_connection()
        cursor = connect.cursor()
        cursor.execute(query, (rid, waf_id, agent, ip, uuid, ts, sysmon, ip, ts, sysmon,))
        close_db_connection(connect, cursor)

        # response
        return {'status': 'success'}

    except Exception as e:
        log.error('An error occurred in set_agent_monitoring: {}'.format(e))
        return {'status': 'fail', 'description': str(e)}


def del_agent_monitoring_v1(ids, waf_id):
    try:

        # skip empty
        if not ids:
            return {'status': 'fail', 'description': 'Empty or missing column: id'}

        # ID type validation
        if not isinstance(ids, list):
            return {'status': 'fail', 'description': 'Value of "{}" contains incorrect data'.format(ids)}
        if any(x for x in ids if not isinstance(x, int) or isinstance(x, bool)):
            return {'status': 'fail', 'description': 'Value of "{}" contains incorrect data'.format(ids)}

        # convert to tuple
        ids = tuple(ids)

        # delete by id
        query = """ DELETE FROM agent_monitoring WHERE (waf_id = %s AND id IN %s) RETURNING id """
        connect = init_db_connection()
        cursor = connect.cursor()
        cursor.execute(query, (waf_id, ids,))
        res = cursor.fetchall()
        close_db_connection(connect, cursor)

        # processing result
        response = {'status': 'success', 'deleted': len(res)}

        # response
        return response

    except Exception as e:
        log.error('An error occurred in del_agent_monitoring: {}'.format(e))
        return {'status': 'fail', 'description': str(e)}
