#!/usr/bin/env python3

"""
Nemesida WAF API
Copyright (c) 2017-2025 Pentestit LLC - All Rights Reserved
Author: Romanov R.
"""

import base64
import pandas as pd
import re

from binascii import hexlify
from time import time
from urllib.parse import unquote

from core import attack_zone_replace
from core import cc_upd_db
from core import geo_token_update
from core import get_rldesc
from core import get_waf_id
from core import init_db_connection, close_db_connection
from core import memc
from core import memc_lock
from core import waf_id_ce

from logger import log

##

attack_not_null_cols = ['id', 'timestamp', 'lm', 'vhost', 'ip', 'method', 'bt',]

# BT description
bt_description = {
    0: 'Request logger',
    1: 'Signature analysis',
    2: 'Signature analysis',
    3: 'Nemesida AI MLA',
    4: 'Antivirus analysis',
    5: 'Internal error',
    6: 'Autoban',
    7: 'Brute-force',
    8: 'Nemesida AI MLC',
    9: 'Flood',
    10: 'DDoS',
    14: 'API Firewall',
    15: 'RFC violation',
    16: 'Blocked IP'
}

##


def get_attack(url, data, agent_ip, waf_id):
    try:

        # Update the CC
        cc_upd_db('attack')

        ##
        # Receive data from the database
        ##

        # init
        response = {'status': 'success', 'count': 0, 'totalCount': 0, 'waf_id': waf_id, 'attacks': []}
        
        # init query filter headers/values
        qhdrs = ['waf_id = %s']
        qargs = [waf_id]

        # processing deprecated
        if 'signaturetype' in data:
            k = 'description'
            data[k] = data['signaturetype'].copy() if k not in data else data[k]
            del data['signaturetype']

        # init parameters type
        mgmt_param = [
            'limit',
            'offset',
            'bodylimit',
        ]
        int_param = [
            'starttime',
            'endtime',
        ]
        bool_param = [
            'reverseorder',
            'blocked',
            'possible',
        ]
        list_s_param = [
            'bt',
            'ip',
            'cc',
            'vhost',
            'method',
            'rule_id',
            'description',
            'long_description',
            'mz',
            'request_id',
            'request_type',
            'group_id',
        ]
        list_x_param = [
            'url',
            'args',
            'body',
            'cookie',
            'other_headers',
            'referer',
            'ua',
            'search',
        ]
        list_param = list_s_param + list_x_param
        all_param = mgmt_param + int_param + bool_param + list_param

        # remove unknown parameter
        for k in data.copy().keys():
            if k not in all_param:
                del data[k]

        # parameters processing
        for k in all_param:
            try:

                # skip non-exists key
                if k not in data:
                    continue

                # delete empty
                if (k in data) and (data.get(k) is None):
                    del data[k]
                    continue

                # int validation
                if k in mgmt_param + int_param:
                    if (isinstance(data[k], bool)) or (not isinstance(data[k], int)):
                        response = {'status': 'fail', 'description': 'Incorrect value for {}'.format(k)}
                        return response
                    if k in mgmt_param:
                        continue

                # bool validation
                elif k in bool_param:
                    if not isinstance(data[k], bool):
                        response = {'status': 'fail', 'description': 'Incorrect value for {}'.format(k)}
                        return response

                # list validation and processing
                elif k in list_param:

                    # remove empty
                    if not data[k]:
                        del data[k]
                        continue

                    # type validation
                    if not isinstance(data[k], list):
                        response = {'status': 'fail', 'description': 'Incorrect value for {}'.format(k)}
                        return response

                    # CC normalization
                    if k == 'cc':

                        # value validation
                        if [x for x in data[k] if not (x is None or isinstance(x, str))]:
                            response = {'status': 'fail', 'description': 'Value of "{}" contains incorrect data'.format(k)}
                            return response

                        # value normalization
                        data[k] = [None if (not x or not str(x).strip()) else x.lower() for x in data[k]]
                        data[k].append('  ') if None in data[k] else data[k]

                        # value processing
                        q = """ AND (LOWER({}) IN %s OR {} IN %s) """.format(k, k)
                        qhdrs.append(q)
                        qargs.extend([tuple(data[k]), tuple(data[k])])
                        continue

                    # null value processing
                    null = True if None in data[k] else False
                    qnull = ' OR {} IS NULL'.format(k) if null else ''

                    # fail if null is not allowed
                    if null and k in attack_not_null_cols:
                        response = {'status': 'fail', 'description': 'Null is not allowed for {}'.format(k)}
                        return response

                    # remove null values
                    data[k] = [x for x in data[k] if x is not None]

                    # bt/rule_id
                    if k in ['bt', 'rule_id']:

                        # value validation
                        if [x for x in data[k] if (not isinstance(x, int)) or (isinstance(x, bool))]:
                            response = {'status': 'fail', 'description': 'Value of "{}" contains incorrect data'.format(k)}
                            return response

                        # value processing
                        q = """ AND ({} IN %s{}) """.format(k, qnull)
                        qhdrs.append(q)
                        qargs.append(tuple(data[k]))
                        continue

                    # mz
                    elif k == 'mz':

                        # value validation
                        if [x for x in data[k] if (not isinstance(x, str)) or (not re.match(r'[\w-]', x))]:
                            response = {'status': 'fail', 'description': 'Value of "{}" contains incorrect symbol'.format(k)}
                            return response

                        # init
                        use_other_headers = False

                        # zone normalization
                        for idx, item in enumerate(data[k].copy()):

                            # user-agent normalization
                            if item == 'ua':
                                data[k][idx] = 'user-agent'

                            # other_headers normalization
                            elif item == 'other_headers':
                                use_other_headers = True
                                del data[k][idx]

                            # remove unknown zone
                            elif item not in ['url', 'args', 'body', 'cookie', 'referer', 'user-agent']:
                                del data[k][idx]

                        # update the query
                        if use_other_headers:
                            s1 = '|'.join([str(x) for x in data[k]])
                            s2 = '|'.join([str(x) for x in ['url', 'args', 'body', 'cookie', 'referer', 'user-agent']])
                            q = """
                                AND (
                                    LOWER(mz) SIMILAR TO %s
                                        OR
                                    mz != '' AND LOWER(mz) NOT SIMILAR TO %s
                                    {}
                                )
                            """.format(qnull)
                            qhdrs.append(q)
                            qargs.append(s1)
                            qargs.append(s2)
                        else:
                            s1 = '|'.join([str(x) for x in data[k]])
                            q = """ AND (LOWER(mz) SIMILAR TO %s{}) """.format(qnull)
                            qhdrs.append(q)
                            qargs.append(s1)

                        # finish parameter processing
                        continue

                    # other keys
                    else:

                        # convert value to lowercase
                        data[k] = [str(x).lower() if x or x == 0 else x for x in data[k]]

                        # request_type normalization
                        if k == 'request_type':
                            if 'generic' in data[k]:
                                data[k].append('')

                        # remove duplicate
                        data[k] = list(set(data[k]))

                        # final value normalization
                        if k in list_s_param:
                            data[k] = '|'.join(
                                ''.join('\\x' + hexlify(c.encode('UTF-8')).decode('UTF-8') for c in [x for x in i]) for i in data[k]
                            )
                        else:
                            data[k] = '|'.join(
                                '%' + ''.join('\\x' + hexlify(c.encode('UTF-8')).decode('UTF-8') for c in [x for x in i]) + '%' for i in data[k]
                            )

                        # body processing
                        if k == 'body':
                            q = """ AND (LOWER(CONVERT_FROM({}, \'utf-8\')) SIMILAR TO %s{}) """.format(k, qnull)
                            qhdrs.append(q)
                            qargs.append(data[k])
                            continue

                        # description type processing
                        if k == 'description':

                            r = [
                                k1 for k1, v1 in bt_description.items() if v1.lower() in [
                                    x.lower() for x in data[k] if isinstance(x, str)
                                ]
                            ]

                            if r:
                                q = """ AND (LOWER({}) SIMILAR TO %s{} OR bt IN %s) """.format(k, qnull)
                                qhdrs.append(q)
                                qargs.append(data[k])
                                qargs.append(tuple(r))
                            else:
                                q = """ AND (LOWER({}) SIMILAR TO %s{}) """.format(k, qnull)
                                qhdrs.append(q)
                                qargs.append(data[k])

                            continue

                        # long_description type processing
                        if k == 'long_description':
                            q = """ AND (LOWER({}) SIMILAR TO %s{}) """.format(k, qnull)
                            qhdrs.append(q)
                            qargs.append(data[k])
                            continue

                        # search processing
                        if k == 'search':

                            # init
                            ql = []
                            zones = ['url', 'param', 'body', 'cookie', 'other_headers', 'referer', 'ua']

                            # generate query list
                            for z in zones:
                                qni = ' OR {} IS NULL'.format(z) if null else ''
                                f = 'LOWER(CONVERT_FROM({}, \'utf-8\'))' if z == 'body' else 'LOWER({})'
                                f = f.format(z)
                                ql.append(""" ({} SIMILAR TO %s{}) """.format(f, qni))
                                del qni

                            # generate the query
                            q = ' AND (' + ' OR '.join(ql) + ')'
                            del ql

                            # update the query
                            qhdrs.append(q)

                            # update query args
                            for _ in range(0, 7):
                                qargs.append(data[k])

                            # finish parameter processing
                            continue

                ##
                #  Update the query
                ##

                # start time
                if k == 'starttime':
                    q = """ AND (timestamp >= to_timestamp(%s)) """
                    qhdrs.append(q)

                # end time
                elif k == 'endtime':
                    q = """ AND (timestamp <= to_timestamp(%s)) """
                    qhdrs.append(q)

                # int
                elif k in int_param:
                    q = """ AND ({} = %s) """.format(k)
                    qhdrs.append(q)

                # bool
                elif k in bool_param:
                    
                    if k in ['reverseorder']:
                        continue

                    q = """ AND ({} IS %s) """.format(k)
                    qhdrs.append(q)

                # list
                elif k in list_param:

                    # null value processing
                    qnull = ' OR {} IS NULL'.format(k) if any(x for x in data[k] if x is None) else ''

                    # query processing
                    if k == 'args':
                        q = """ AND (LOWER(param) SIMILAR TO %s{}) """.format(qnull)
                    else:
                        q = """ AND (LOWER({}) SIMILAR TO %s{}) """.format(k, qnull)

                    qhdrs.append(q)

                else:
                    log.error('An error occurred while extracting data from {} to {}: unknown parameter ({})'.format(agent_ip, url, k))
                    continue

                # update query args
                qargs.append(data[k])

            except Exception as e:
                log.error('An error occurred while extracting data from {} to {}: {}'.format(agent_ip, url, e))
                response = {'status': 'fail', 'description': str(e)}
                return response

        # extract the limit
        limit = data.get('limit')
        limit = limit if isinstance(limit, int) and (not isinstance(limit, bool)) and (limit > 0) else 0

        # offset wo limit
        if ('offset' in data) and (limit < 1):
            response = {'status': 'fail', 'description': 'Using parameter "offset" with parameter "limit" < 1 is not allowed'}
            return response

        # extract the offset
        offset = data.get('offset')
        offset = offset * limit if (limit and offset) else 0

        # order processing
        order = 'DESC' if data.get('reverseorder') else 'ASC'

        # body processing
        k = 'bodylimit'
        if data.get(k):
            qbody = 'CONVERT_FROM(SUBSTRING(body FOR {}), \'utf-8\')'.format(data[k])
        else:
            qbody = 'CONVERT_FROM(body, \'utf-8\')'

        ##
        # Search data in the database
        ##

        # convert query filter headers/values
        qhdrs = ' '.join(qhdrs)
        qargs = tuple(qargs)

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # total count
        query = """ SELECT COUNT(*) FROM attack WHERE {} """.format(qhdrs)
        cursor.execute(query, qargs)
        response['totalCount'] = cursor.fetchone()[0]

        # search the data

        query = """
            SELECT
                id, CAST(EXTRACT(EPOCH FROM timestamp) as integer), request_id, request_type, group_id, rule_id,
                description, long_description, blocked, possible, lm, LOWER(mz), bt, ip, UPPER(cc), vhost, method, url, param,
                {}, cookie, other_headers, ua, referer
            FROM
                attack
            WHERE {}
            ORDER BY id {}
        """.format(qbody, qhdrs, order)

        if limit:
            query += ' LIMIT %s'
            qargs += (limit,)

        if offset:
            query += ' OFFSET %s'
            qargs += (offset,)

        cursor.execute(query, qargs)
        res = cursor.fetchall()

        # del vars
        del qhdrs, qargs, data

        # processing result
        if res:

            # update the response data
            response['count'] = len(res)

            # processing the result
            for item in res:

                # item processing
                r = {
                    'id': item[0],
                    'timestamp': item[1],
                    'request_id': item[2],
                    'request_type': item[3],
                    'group_id': item[4],
                    'rule_id': item[5],
                    'description': item[6],
                    'long_description': item[7],
                    'blocked': item[8],
                    'possible': item[9],
                    'lm': item[10],
                    'mz': item[11],
                    'bt': item[12],
                    'ip': item[13],
                    'cc': item[14],
                    'vhost': item[15],
                    'method': item[16],
                    'url': item[17],
                    'args': item[18],
                    'body': item[19],
                    'cookie': item[20],
                    'other_headers': item[21],
                    'ua': item[22],
                    'referer': item[23]
                }

                # CC processing
                if not r['cc'].strip():
                    r['cc'] = None

                # update the response data
                response['attacks'].append(r)

                # del vars
                del item, r

        # close database connection
        close_db_connection(connect, cursor)

        # response
        return response

    except Exception as e:
        log.error('An error occurred in get_attack: {}'.format(e))


def set_attack(data, agent_ip):
    try:

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        ##
        # Processing dictionary of the attacks
        ##

        for item in data:
            try:

                # extract lic. key
                key = item.get('waf_lic_key', '')

                # WAF ID processing
                if key == 'none':
                    waf_id = waf_id_ce
                else:
                    waf_id = get_waf_id(key, agent_ip)
                    geo_token_update(key, agent_ip)
                    if not waf_id:
                        continue

                # init
                result = {}

                # missing column fix
                for k in ['description', 'long_description', 'group_id', 'request_type']:
                    if k not in item:
                        item[k] = ''

                # url fix
                if 'uri' in item:
                    item['url'] = item['uri']
                    item.pop('uri')

                # rename content to boy
                try:
                    if 'content' in item:
                        item['body'] = item.pop('content')
                except Exception as e:
                    log.error('An error occurred while renaming column "content" to "body": {}'.format(e))

                ##
                # b64decode
                ##

                # init
                b64_keys = (
                    'ip',
                    'request_id',
                    'request_type',
                    'vhost',
                    'url',
                    'param',
                    'mz',
                    'method',
                    'ua',
                    'referer',
                    'other_headers',
                    'body',
                    'cookie',
                    'description',
                    'long_description',
                    'group_id',
                )

                # decode custom field
                for k in b64_keys:
                    try:

                        # skip empty/null
                        if item.get(k) is None:
                            result[k] = '' if k == 'mz' else None
                            continue

                        # base64 decode
                        result[k] = base64.b64decode(item[k]).decode('UTF-8', 'ignore')

                    except Exception as e:
                        log.error('An error occurred while decoding the value of {}: {} ({})'.format(k, e, item.get(k)))
                        result[k] = 'API DECODE ERROR'

                ##
                # Extract data from the JSON item
                ##

                # extract items
                bt = int(item.get('bt', 0))
                rule_id = int(item.get('rule_id', 0))
                ip = result.get('ip', '')
                vhost = result.get('vhost', '')
                mz_list = result.get('mz', '').split('|')
                request_id = result.get('request_id', '0')
                request_type = result.get('request_type')
                body = result.get('body', '')


                # logging
                log.info('Request {} (BT: {}) received from {}'.format(request_id, bt, agent_ip))

                # mz normalization
                mz_list = attack_zone_replace(mz_list)

                # extract body
                try:
                    if body:
                        body = bytearray(body, 'UTF-8')
                except Exception as e:
                    log.error('An error occurred while extracting the "body" data from request {}: {}'.format(request_id, e))
                    body = ''

                ##
                # Processing for each zone in mz
                ##

                for mz in mz_list:

                    # init
                    description = ''

                    # processing specific bt
                    if bt in [1, 2] and rule_id != 0:

                        ##
                        # Receive description
                        ##

                        # init
                        memc_key = 'rldsc_' + str(rule_id)

                        # Search description in the memcached
                        try:

                            # get data from the memcached
                            with memc_lock:
                                try:
                                    memc_bin = memc.get(memc_key)
                                except Exception as e:
                                    memc_bin = None
                                    log.error('An error occurred while receiving data from the memcached: {}'.format(e))

                            # if data exist in memcached
                            if memc_bin:
                                description = memc_bin.decode('UTF-8')

                        except Exception as e:
                            log.error(
                                'An error occurred while processing rule description from the memcached: {}'
                                .format(e)
                            )
                            description = ''

                        # Receive the description from the remote server if description is empty
                        if not description or description == '-':

                            # receive the description
                            description = get_rldesc(rule_id)

                            # update data in memcached
                            try:
                                if description:
                                    with memc_lock:
                                        try:
                                            memc.set(memc_key, description, 86400)  # 84600 - 1 day
                                        except Exception as e:
                                            log.error('An error occurred while placing data into the memcached: {}'.format(e))
                            except Exception as e:
                                log.error(
                                    'An error occurred while storing description "{}" for rule ID {} in the memcached: {}'
                                    .format(description, rule_id, e)
                                )

                    elif bt == 4:
                        description = 'Virus (' + result['description'] + ')'

                    elif bt == 5:
                        mz = ''
                        description = result['description']

                    elif bt == 7:
                        description = result['description']
                        if not description or description == '-':
                            description = 'Brute-force'

                    elif bt == 9:
                        description = result['description']
                        if not description or description == '-':
                            description = 'Flood'

                    elif bt == 10:
                        mz = ''
                        description = result['description']

                    elif bt == 16:
                        description = 'Blocked IP'

                    else:
                        description = result['description']
                        if not description:
                            description = ''

                    ##
                    # Store attack to the database
                    ##

                    try:

                        # preparing a query
                        query = """
                            INSERT INTO attack (
                                timestamp, lm, bt, mz, ip, vhost, rule_id, request_id, request_type,
                                description, long_description, waf_id, url, method,
                                referer, param, ua, cookie, other_headers, body,
                                group_id, blocked, possible, cc
                            ) VALUES (
                                to_timestamp(%s), %s, %s, %s, %s, %s, %s,
                                %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s,
                                %s, %s, %s, %s, %s, %s
                            ) ON CONFLICT DO NOTHING
                        """

                        dbr = (
                            item.get('timestamp', ''),
                            item.get('lm', ''),
                            bt,
                            mz,
                            ip,
                            vhost,
                            rule_id,
                            request_id,
                            request_type,
                            description,
                            result['long_description'],
                            waf_id,
                            result['url'],
                            result['method'],
                            result['referer'],
                            result['param'],
                            result['ua'],
                            result['cookie'],
                            result['other_headers'],
                            body,
                            result['group_id'],
                            item.get('blocked', None),
                            item.get('possible', None),
                            item.get('cc', None),
                        )

                        # insert data to the DB
                        cursor.execute(query, dbr)

                    except Exception as e:
                        log.error(
                            'An error occurred while storing the data of request {} into the database: {} ({})'
                            .format(request_id, e, item)
                        )

            except Exception as e:
                log.error('An error occurred while processing JSON item: {} ({})'.format(e, item))

        # close database connection
        close_db_connection(connect, cursor)

    except Exception as e:
        log.error('An error occurred in set_attack: {}'.format(e))


def del_attack(data, waf_id):
    try:

        # extract the id
        ids = data.get('id')

        # skip empty
        if not ids:
            return {'status': 'fail', 'description': 'Empty or missing column: id'}

        # id processing
        if (not isinstance(ids, list)) or ([x for x in ids if (not isinstance(x, int)) or (isinstance(x, bool))]):
            response = {'status': 'fail', 'description': 'Value of "id" contains incorrect data or is missing'}
            return response

        # convert to tuple
        ids = tuple(ids)

        ##
        # Processing JSON
        ##

        # init
        response = {'status': 'success', 'deleted': 0}

        # insert the data
        query = """ DELETE FROM attack WHERE (id IN %s) AND (waf_id = %s) RETURNING id """
        connect = init_db_connection()
        cursor = connect.cursor()
        cursor.execute(query, (ids, waf_id,))
        res = cursor.fetchall()
        close_db_connection(connect, cursor)

        # processing result
        response['deleted'] = len(res)

        # response
        return response

    except Exception as e:
        log.error('An error occurred in del_attack: {}'.format(e))


def get_attack_stats(data, waf_id):
    try:

        # init
        st = 'starttime'
        et = 'endtime'
        ts = int(time())
        use_period = False
        d01_ts = ts - 86400
        d07_ts = ts - (86400 * 7)
        d30_ts = ts - (86400 * 30)
        response = {'status': 'success'}
        params = ['bt', 'ip', 'cc', 'vhost']

        # init param in response
        for param in params:
            response[param] = []

        # Update the CC
        cc_upd_db('attack')

        ##
        # Receive data from the database
        ##

        # processing the parameters
        if st in data and et in data:
            if (
                (isinstance(data[st], int)) and (not isinstance(data[st], bool))
                    and
                (isinstance(data[et], int)) and (not isinstance(data[et], bool))
            ):
                use_period = True

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # search the data
        if use_period:

            # select for specific period
            query = """
                SELECT
                    json_agg(
                        json_build_object(
                            'timestamp', CAST(EXTRACT(EPOCH FROM timestamp) as integer),
                            'bt', bt,
                            'ip', ip,
                            'cc', UPPER(cc),
                            'vhost', vhost
                        )
                    )
                FROM
                    attack
                WHERE (
                    waf_id = %s
                        AND
                    timestamp >= to_timestamp(%s)
                        AND
                    timestamp <= to_timestamp(%s)
                )
            """
            cursor.execute(query, (waf_id, data[st], data[et],))

        else:

            # select for 1 month
            query = """
                SELECT
                    json_agg(
                        json_build_object(
                            'timestamp', CAST(EXTRACT(EPOCH FROM timestamp) as integer),
                            'bt', bt,
                            'ip', ip,
                            'cc', UPPER(cc),
                            'vhost', vhost
                        )
                    )
                FROM
                    attack
                WHERE (
                    waf_id = %s
                        AND
                    timestamp > to_timestamp(%s)
                )
            """
            cursor.execute(query, (waf_id, d30_ts,))

        # get the result
        res = cursor.fetchall()

        # close database connection
        close_db_connection(connect, cursor)

        """
        If the result is not empty, returned a list of tuples, e.g.:
        [([{'bt': 1, ...}, {'bt': 2, ...},...],)]
        """

        # processing result
        if res and res[0][0]:

            # import to pd
            df = pd.DataFrame(res[0][0])
            del res

            # CC normalization
            df.loc[df['cc'].isnull(), 'cc'] = ''
            df.loc[df['cc'].str.strip() == '', 'cc'] = ''

            # processing for each param from the list
            for param in params:

                # processing for each unique value of param
                for v in sorted(df[param].unique()):

                    # extract param data
                    pdf = df.loc[df[param] == v]

                    # value normalization
                    if param == 'bt':
                        v = int(v)
                    elif param == 'cc':
                        v = v if v else None

                    # attack by period
                    if use_period:
                        response[param].append({
                            param: v,
                            'period': len(pdf)
                        })
                    else:
                        response[param].append({
                            param: v,
                            'lastDay': len(pdf.loc[pdf['timestamp'] > d01_ts]),
                            'last7Days': len(pdf.loc[pdf['timestamp'] > d07_ts]),
                            'last30Days': len(pdf.loc[pdf['timestamp'] > d30_ts])
                        })

                    # del vars
                    del pdf

            # del vars
            del df

        # response
        return response

    except Exception as e:
        log.error('An error occurred in get_attack_stats: {}'.format(e))


def get_attack_vhost_stats(data, waf_id):
    try:

        # init
        vhosts = []
        response = {'status': 'success', 'vhost': {}}
        filters = ['starttime', 'endtime', 'vhost', 'description', 'ip', 'cc', 'bt',]

        # query base
        query_base = """ {} FROM attack WHERE (waf_id = %s) """
        query_base_param = (waf_id,)
        default_query_base = query_base
        default_query_base_param = query_base_param

        # Update the CC
        cc_upd_db('attack')

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # filters processing
        for k in filters:

            ##
            # Period processing
            ##

            # start period time
            if k == 'starttime':

                # non-empty data processing
                if data.get(k):

                    # data type validation
                    if isinstance(data[k], int) and (not isinstance(data[k], bool)):

                        # update query base
                        query_base += """ AND (timestamp >= to_timestamp(%s)) """
                        query_base_param += (data.get(k),)

                    else:

                        # validation failed
                        response = {'status': 'fail', 'description': 'Incorrect value for {}'.format(k)}
                        return response

            # end period time
            elif k == 'endtime':

                # non-empty data processing
                if data.get(k):

                    # data type validation
                    if isinstance(data[k], int) and (not isinstance(data[k], bool)):

                        # update query base
                        query_base += """ AND (timestamp <= to_timestamp(%s)) """
                        query_base_param += (data.get(k),)

                    else:

                        # validation failed
                        response = {'status': 'fail', 'description': 'Incorrect value for {}'.format(k)}
                        return response

            # vhost processing
            elif k == 'vhost':

                # get vhosts list
                vhosts = data.get(k)

                # data type validation
                if vhosts and (not isinstance(vhosts, list)):
                    response = {'status': 'fail', 'description': 'Incorrect value for {}'.format(k)}
                    return response

                # empty vhosts list
                if not vhosts:

                    # search for vhosts list
                    query = query_base.format('SELECT DISTINCT(vhost)')
                    cursor.execute(query, query_base_param)
                    del query

                    # vhost list processing
                    vhosts = [x[0] for x in cursor.fetchall()]

                # skip for empty vhosts list
                if not vhosts:
                    return response

                # update query base string
                if data.get(k):
                    query_base = query_base + """ AND ({} IN %s) """.format(k)
                    query_base_param += (tuple(vhosts),)

            # other filter processing
            else:

                # update query base string
                if k in data:

                    # non-empty data processing
                    if data.get(k):

                        # type validation
                        if not isinstance(data[k], list):

                            # validation failed
                            response = {'status': 'fail', 'description': 'Incorrect value for {}'.format(k)}
                            return response

                        # fail if null is not allowed
                        if (None in data[k]) and (k in attack_not_null_cols):
                            response = {'status': 'fail', 'description': 'Null is not allowed for {}'.format(k)}
                            return response

                        # update query base
                        if k == 'cc':
                            query_base = query_base + """ AND (LOWER({}) IN %s OR {} IN %s) """.format(k, k)
                            data[k] = [None if (not x or not str(x).strip()) else x.lower() for x in data[k]]
                            data[k].append('  ') if None in data[k] else data[k]
                            query_base_param += (tuple(data[k]), tuple(data[k]),)

                        elif k == 'description':

                            r = [
                                k1 for k1, v1 in bt_description.items() if v1.lower() in [
                                    x.lower() for x in data[k] if isinstance(x, str)
                                ]
                            ]

                            if r:
                                query_base = query_base + """ AND ({} IN %s OR bt IN %s) """.format(k)
                                query_base_param += (tuple(data[k]), tuple(r),)
                            else:
                                query_base = query_base + """ AND ({} IN %s) """.format(k)
                                query_base_param += (tuple(data[k]),)

                        else:
                            query_base = query_base + """ AND ({} IN %s) """.format(k)
                            query_base_param += (tuple(data[k]),)

        ##
        #  Generate the result
        ##

        # processing for each vhost
        for vhost in vhosts:

            # create vhost structure
            response['vhost'][vhost] = {
                'summary': {
                    'total': 0,
                    'blocked': 0,
                    'detected': 0
                },
                'bt': {},
                'ip': {},
                'cc': {},
                'description': {},
                'attacks': []
            }

            # get the ids list
            query = query_base.format('SELECT id') + """ AND vhost = %s """
            query_param = query_base_param + (vhost,)
            cursor.execute(query, query_param)
            ids = tuple(x[0] for x in cursor.fetchall())

            # skip empty
            if not ids:
                continue

            # init the vhost query base
            vqb = default_query_base + """ AND (id IN %s) """
            vqbp = default_query_base_param + (ids,)

            # summary: total
            query = vqb.format('SELECT COUNT(*)')
            cursor.execute(query, vqbp)
            response['vhost'][vhost]['summary']['total'] = cursor.fetchone()[0]

            # summary: blocked
            query = vqb.format('SELECT COUNT(*)') + """ AND blocked IS TRUE """
            cursor.execute(query, vqbp)
            response['vhost'][vhost]['summary']['blocked'] = cursor.fetchone()[0]

            # summary: detected
            query = vqb.format('SELECT COUNT(*)') + """ AND blocked IS FALSE """
            cursor.execute(query, vqbp)
            response['vhost'][vhost]['summary']['detected'] = cursor.fetchone()[0]

            # description, bt, ip and cc processing
            for column in ['bt', 'ip', 'cc', 'description']:

                # receive the data from database
                query = vqb.format('SELECT {}, COUNT(*)').format(column) + """ AND {} IN (SELECT DISTINCT({})) GROUP BY {} ORDER BY {} """.format(column, column, column, column)
                cursor.execute(query, vqbp)

                # processing the data
                for x in cursor.fetchall():

                    # init
                    k, v = x[0], x[1]

                    # cc normalization
                    if column == 'cc':
                        k = None if not k.strip() else k.upper()

                    # updating the result
                    response['vhost'][vhost][column][k] = v

            ##
            # Attacks
            ##

            # search the data
            query = vqb.format(
                """
                    SELECT
                        id, CAST(EXTRACT(EPOCH FROM timestamp) as integer), request_id,
                        rule_id, description, request_type, blocked, possible, lm, mz, bt,
                        ip, UPPER(cc), vhost, method
                """
            )
            cursor.execute(query, vqbp)
            res = cursor.fetchall()

            # processing the result
            for item in res:

                # item processing
                r = {
                    'id': item[0],
                    'timestamp': item[1],
                    'request_id': item[2],
                    'rule_id': item[3],
                    'description': item[4],
                    'request_type': item[5],
                    'blocked': item[6],
                    'possible': item[7],
                    'lm': item[8],
                    'mz': item[9],
                    'bt': item[10],
                    'ip': item[11],
                    'cc': item[12],
                    'vhost': item[13],
                    'method': item[14]
                }

                # CC processing
                if not r['cc'].strip():
                    r['cc'] = None

                # update the response data
                response['vhost'][vhost]['attacks'].append(r)

                # del vars
                del item, r

        # close database connection
        close_db_connection(connect, cursor)

        # response
        return response

    except Exception as e:
        log.error('An error occurred in get_attack_vhost_stats: {}'.format(e))


def get_attack_spec(waf_id):
    try:

        # init
        result = {'status': 'success', 'specification': {'analyzer': {}}}

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # result processing
        for bt in bt_description.keys():

            # BT: 1
            if bt == 1:

                # init
                query = """ SELECT DISTINCT(description) FROM attack WHERE (waf_id = %s) AND (bt IN (1, 2)) """
                cursor.execute(query, (waf_id,))

                # update the result
                result['specification']['analyzer'][bt_description[bt]] = {
                    'bt': [1, 2],
                    'description': [x[0] for x in cursor.fetchall()]
                }

            # BT: 2 (processing by BT 1)
            elif bt == 2:
                continue

            # other BT processing
            else:

                 # init
                query = """ SELECT DISTINCT(description) FROM attack WHERE (waf_id = %s) AND (bt = %s) """
                cursor.execute(query, (waf_id, bt,))

                # update the result
                result['specification']['analyzer'][bt_description[bt]] = {
                    'bt': [bt],
                    'description': [x[0] for x in cursor.fetchall()]
                }

        # close database connection
        close_db_connection(connect, cursor)

        # response
        return result

    except Exception as e:
        log.error('An error occurred in get_attack_spec: {}'.format(e))


def set_scan(url, rsc, data, agent_ip):
    try:

        # processing data from the JSON
        for item in data:

            # decode WAF ID
            try:
                key = base64.b64decode(item['waf_lic_key']).decode('UTF-8', 'ignore')
                if not key:
                    log.error('An error occurred while processing data in {} from client {}: no product key'.format(url, agent_ip))
                    return 401
            except Exception as e:
                log.error(
                    'An error occurred while processing product key from request to {} (client IP: {}): {}'
                    .format(url, agent_ip, e)
                )
                return 401

            # decode other fields
            domain_scan = base64.b64decode(item['domain']).decode('UTF-8', 'ignore')
            method_scan = base64.b64decode(item['method']).decode('UTF-8', 'ignore')
            param_scan = base64.b64decode(item['param']).decode('UTF-8', 'ignore')
            data_scan = base64.b64decode(item['data']).decode('UTF-8', 'ignore')
            scan_date_scan = base64.b64decode(item['date']).decode('UTF-8', 'ignore')
            payload_scan = base64.b64decode(item['payload']).decode('UTF-8', 'ignore')
            type_scan = base64.b64decode(item['type']).decode('UTF-8', 'ignore')
            url_scan = base64.b64decode(item['URL']).decode('UTF-8', 'ignore')

            # URL decode
            url_scan = unquote(unquote(url_scan))
            param_scan = unquote(unquote(param_scan))

            # mz processing
            if 'mz' in item:
                mz = base64.b64decode(item['mz']).decode('UTF-8', 'ignore')
            else:
                mz = ''

            # severity processing
            if 'severity' in item:
                severity_scan = base64.b64decode(item['severity']).decode('UTF-8', 'ignore')
            else:
                if type_scan.lower() in ['ba', 'lfi', 'rce', 'rfi', 'sqli', 'ssti']:
                    severity_scan = 'High'
                elif type_scan.lower() in ['sde', 'ssrf', 'vc', 'xss']:
                    severity_scan = 'Medium'
                else:
                    severity_scan = 'Low'

            # WAF ID check
            waf_id = get_waf_id(key, agent_ip)
            if not waf_id:
                return 401

            ##
            # Storing scanner result to the database
            ##

            # prepare the query
            query = """
                INSERT INTO scan_report (
                    domain, method, scan_date, mz, param, data,
                    payload, type, url, content, waf_id
                ) VALUES (
                    %s, %s, %s, %s, %s, %s,
                    %s, %s, %s, %s, %s
                )
            """

            # prepare the array
            dbr = (
                domain_scan, method_scan, scan_date_scan, mz, param_scan, data_scan,
                payload_scan, type_scan, url_scan, severity_scan, waf_id,
            )

            try:

                # insert the data
                connect = init_db_connection()
                cursor = connect.cursor()
                cursor.execute(query, dbr)
                close_db_connection(connect, cursor)

                # logging
                log.info(
                    'Scanner result for domain {} has been successfully stored in the database (client IP: {})'
                    .format(domain_scan, agent_ip)
                )

            except Exception as e:
                log.error(
                    'An error occurred while storing scanner data in the database (client IP: {}): {}'
                    .format(e, agent_ip)
                )
                return 400

        # response
        return rsc

    except Exception as e:
        log.error('An error occurred in set_scan: {}'.format(e))
