#!/usr/bin/env python3

"""
Nemesida WAF API
Copyright (c) 2017-2025 Pentestit LLC - All Rights Reserved
Author: Romanov R.
"""

import hashlib
import json

from netaddr import IPAddress

from core import init_db_connection, close_db_connection
from core import memc
from core import memc_lock

from ws import dyn_sync_ban_cols
from ws import dyn_sync_ban_memc_ttl
from ws import dyn_sync_ban_value_validation

from logger import log

##

memc_prefix = 'dyn_sync_ban_'

##


def get_dyn_sync_ban(url, waf_id):
    try:

        ##
        # Get related keys
        ##

        # init
        relations = []

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        try:

            # search the data
            query = """ SELECT key_id FROM dyn_sync_ban WHERE (waf_id = %s) """
            cursor.execute(query, (waf_id,))
            res = cursor.fetchall()

            # processing result
            if res:

                # extract relations as list
                relations = [i for sub in res for i in sub]

        except Exception as e:
            log.error('An error occurred while processing relations from the database in {}: {}' .format(url, e))

        ##
        # Processing relations
        ##

        # init
        result = []

        # processing key from relations
        for key_id in relations:
            try:

                # get data from the memcached
                memc_key = memc_prefix + str(key_id)
                with memc_lock:
                    try:
                        data = memc.get(memc_key)
                    except Exception as e:
                        data = None
                        log.error('An error occurred while receiving data from the memcached: {}'.format(e))

                # processing data
                if data:

                    # load data
                    data = json.loads(data.decode('UTF-8'))

                    # update response result if ret is not empty
                    if data:

                        # data normalization
                        ban_ts = data['timestamp']
                        ban_data = data['ban_data']

                        # add timestamp to ban_data
                        for d in ban_data:
                            try:

                                # set the timestamp
                                d['timestamp'] = ban_ts

                                # domain normalization
                                if d['domain'] == 'default':
                                    d['domain'] = d['domain'].upper()

                                # add TT column
                                d['tt'] = int(d['TTL']) + int(d['timestamp'])

                                # update result dictionary
                                result.append(d)

                            except Exception as e:
                                log.error(
                                    'An error occurred while GetDynSyncBan memcached data normalization for KeyID {}: {} ({})'
                                    .format(key_id, e, d)
                                )

                        # del vars
                        del ban_ts, ban_data, data

            except Exception as e:
                log.error(
                    'An error occurred while processing GetDynSyncBan memcached data data normalization for KeyID {}: {}'
                    .format(key_id, e)
                )

        # init
        response = {'status': 'success', 'ban_data': []}

        # result processing
        if result:
            try:

                # init
                r = []
                uc = ['ip-address', 'is_banned', 'domain']  # unique column list

                # sort the result based on the largest tt value
                result = sorted(result, key=lambda x: x['tt'], reverse=True)

                # for rows with same ip-address+is_banned+domain save row with greater TTL+timestamp
                for item in result:

                    # unique item
                    if (item[uc[0]], item[uc[1]], item[uc[2]]) not in [(x[uc[0]], x[uc[1]], x[uc[2]]) for x in r]:

                        # remove extra columns
                        for k in ['tt', 'unban_exp']:
                            if k in item:
                                del item[k]

                        # update the result
                        r.append(item)

            except Exception as e:
                log.error('An error occurred while processing the result in GetDynSyncBan for WAF ID {}: {}'.format(waf_id, e))

            # update response
            response['ban_data'] = r

            # del vars
            del result, r

        # close database connection
        close_db_connection(connect, cursor)

        # response
        return response

    except Exception as e:
        log.error('An error occurred in get_dyn_sync_ban: {}'.format(e))


def set_dyn_sync_ban(url, data, key, agent_ip, waf_id):
    try:

        # convert all keys to lowercase
        data = {k.lower(): v for k, v in data.items()}

        ##
        # Processing request data
        ##

        # remove unused keys
        data = {k: data[k] for k in dyn_sync_ban_cols if k in data}

        # check columns
        for k in dyn_sync_ban_cols:
            if k not in data.keys():
                response = {'status': 'fail', 'description': 'Required key missing'}
                return response

        # value validation
        for k in dyn_sync_ban_cols:

            # init
            v = data[k]

            # value validations
            if not dyn_sync_ban_value_validation(agent_ip, k, v):

                # logging
                log.error('Incorrect value for {} received from {} to {}: {}'.format(k, agent_ip, url, v))

                # response
                response = {'status': 'fail', 'description': 'Incorrect value for {}'.format(k)}
                return response

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # convert key to md5
        key_id = hashlib.md5(key.encode('UTF-8')).hexdigest()

        ##
        # Insert key to relations table
        ##

        try:

            # insert the data
            query = """
                INSERT INTO dyn_sync_ban (
                    key_id, waf_id
                ) VALUES (
                    %s, %s
                ) ON CONFLICT (
                    key_id
                ) DO UPDATE SET
                    waf_id = %s
            """
            cursor.execute(query, (key_id, waf_id, waf_id,))

        except Exception as e:
            log.error('An error occurred while inserting data to dyn_sync_ban: {}'.format(e))
            close_db_connection(connect, cursor)
            return

        ##
        # Get related keys
        ##

        try:

            # search the data
            query = """ SELECT key_id FROM dyn_sync_ban WHERE (waf_id = %s) """
            cursor.execute(query, (waf_id,))
            res = cursor.fetchall()

            # extract relations as list
            relations = [i for sub in res for i in sub]

        except Exception as e:
            log.error('An error occurred while processing relations from the database: {}' .format(e))
            close_db_connection(connect, cursor)
            return

        # close database connection
        close_db_connection(connect, cursor)

        # init
        dyn_data = data.copy()
        del data

        # receive the data from the memcached if ban_data is not empty
        if dyn_data.get('ban_data'):
            try:

                # get data from the memcached
                memc_key = memc_prefix + str(key_id)
                with memc_lock:
                    try:
                        memc_bin = memc.get(memc_key)
                    except Exception as e:
                        memc_bin = None
                        log.error('An error occurred while receiving data from the memcached: {}'.format(e))

                # processing data
                if memc_bin:

                    # load data
                    api_data = json.loads(memc_bin.decode('UTF-8'))

                    # update dyn ban data with IP having is_banned status == 0
                    if api_data:

                        # processing ban_data from Dyn
                        for idx, dd in enumerate(dyn_data['ban_data'].copy()):

                            # skip empty
                            if not dd:
                                continue

                            # extract the data
                            ip = dd['ip-address']
                            domain = dd['domain']
                            unban_exp = int(dd.get('unban_exp', '0'))

                            # processing ban_data from API
                            for ad in api_data['ban_data']:

                                # search item from dyn_ban_data in api_ban_data
                                if (
                                        dd['is_banned'] == '1'
                                    ) and (
                                        ad['is_banned'] == '0'
                                    ) and (
                                        ip == ad['ip-address']
                                    ) and (
                                        domain == ad['domain']
                                    ) and (
                                        int(dd['TTL']) > unban_exp
                                ):

                                    # set extra column
                                    dyn_data['ban_data'][idx]['unban_exp'] = dd['TTL']

                                    # update dyn_ban_data for IP
                                    dyn_data['ban_data'][idx]['is_banned'] = '0'
                                    dyn_data['ban_data'][idx]['TTL'] = str(int(dd['TTL']) + 60)  # notify dyn developer if value (60) change

                    # del vars
                    del api_data

            except Exception as e:
                log.error('An error occurred while processing SetDynSyncBan data for key id {}: {}'.format(key_id, e))
                return

        # init memcached
        memc_key = memc_prefix + key_id

        # store data to memcached
        with memc_lock:
            try:
                memc.set(memc_key, json.dumps(dyn_data).encode('UTF-8'), dyn_sync_ban_memc_ttl)
            except Exception as e:
                log.error(
                    'An error occurred while placing dyn_sync_ban data "{}" into the memcached (client {}): {} (data: {})'
                    .format(memc_key, agent_ip, e, dyn_data)
                )

        ##
        # Processing relations
        ##

        # init
        result = []

        # processing key from relations
        for key_id in relations:
            try:

                # get data from the memcached
                memc_key = memc_prefix + str(key_id)
                with memc_lock:
                    try:
                        memc_bin = memc.get(memc_key)
                    except Exception as e:
                        memc_bin = None
                        log.error('An error occurred while receiving data from the memcached: {}'.format(e))

                # processing data
                if memc_bin:

                    # load data
                    r = json.loads(memc_bin.decode('UTF-8'))

                    # remove extra columns
                    for idx, dd in enumerate(r['ban_data'].copy()):
                        for k in ['unban_exp']:
                            if k in dd:
                                del r['ban_data'][idx][k]

                    # update response result if ret is not empty
                    if r:
                        result.append(json.dumps(r))

            except Exception as e:
                log.error(
                    'An error occurred while processing SetDynSyncBan data from the memcached (KeyID {}): {}'
                    .format(key_id, e)
                )

        # init
        response = json.dumps({'status': 'success'})

        # if ban_data exists
        if result:
            response += '\n' + '\n'.join(result)

        # response
        return response

    except Exception as e:
        log.error('An error occurred in set_dyn_sync_ban: {}'.format(e))


def del_dyn_sync_ban(url, data, key, agent_ip, waf_id):
    try:

        # convert all keys to lowercase
        data = {k.lower(): v for k, v in data.items()}

        ##
        # Processing request data
        ##

        # init
        param = 'ip-address'

        # extract IP address
        try:

            if param in data:

                # extract an IP
                ip = data[param].strip()

                ##
                # IP validation
                ##

                # Only IPv4/IPv6/*
                if ip != '*':
                    IPAddress(ip)

            else:

                # response
                err_msg = 'Key missing: {}'.format(param)
                log.error('An error occurred while processing {} from {} to {}: {}'.format(param, agent_ip, url, err_msg))
                response = {'status': 'fail', 'description': err_msg}
                return response

        except Exception as e:
            log.error('An error occurred while processing {} from {} to {}: {}'.format(param, agent_ip, url, e))
            response = {'status': 'fail', 'description': str(e)}
            return response

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # convert key to md5
        key_id = hashlib.md5(key.encode('UTF-8')).hexdigest()

        ##
        # Insert key to relations table
        ##

        try:

            # insert the data
            query = """
                INSERT INTO dyn_sync_ban (
                    key_id, waf_id
                ) VALUES (
                    %s, %s
                ) ON CONFLICT (
                    key_id
                ) DO UPDATE SET
                    waf_id = %s
            """
            cursor.execute(query, (key_id, waf_id, waf_id,))

        except Exception as e:
            log.error('An error occurred while inserting data to dyn_sync_ban: {}'.format(e))
            close_db_connection(connect, cursor)
            return

        ##
        # Get related keys
        ##

        try:

            # search data
            query = """ SELECT key_id FROM dyn_sync_ban WHERE (waf_id = %s) """
            cursor.execute(query, (waf_id,))
            res = cursor.fetchall()

            # extract relations as list
            relations = [i for sub in res for i in sub]

        except Exception as e:
            log.error('An error occurred while processing relations from the database: {}'.format(e))
            close_db_connection(connect, cursor)
            return

        # close database connection
        close_db_connection(connect, cursor)

        ##
        # Processing relations
        ##

        # processing key from relations
        for key_id in relations:
            try:

                # get data from the memcached
                memc_key = memc_prefix + str(key_id)
                with memc_lock:
                    try:
                        memc_bin = memc.get(memc_key)
                    except Exception as e:
                        memc_bin = None
                        log.error('An error occurred while receiving data from the memcached: {}'.format(e))

                # processing data
                if memc_bin:

                    # load data
                    data = json.loads(memc_bin.decode('UTF-8'))

                    # unban an IP from the memcached result
                    if data:

                        # extract ban_data
                        ban_data = data['ban_data']

                        # processing for each IP-address
                        for i, d in enumerate(ban_data):

                            # skip if IP does not match and IP is not a '*'
                            if ip != d['ip-address'] and ip != '*':
                                continue

                            # unban an IP
                            ban_data[i]['is_banned'] = '0'

                        # update ban_data
                        data['ban_data'] = ban_data

                    # update data in memcached
                    with memc_lock:
                        try:
                            memc.set(memc_key, json.dumps(data).encode('UTF-8'), dyn_sync_ban_memc_ttl)
                        except Exception as e:
                            log.error(
                                'An error occurred while placing dyn_sync_ban data "{}" into the memcached (client {}): {} ({})'
                                .format(memc_key, agent_ip, e, data)
                            )

            except Exception as e:
                log.error(
                    'An error occurred while processing DelDynSyncBan data for key id {}: {}'
                    .format(key_id, e)
                )

        # response
        response = {'status': 'success'}
        return response

    except Exception as e:
        log.error('An error occurred in del_dyn_sync_ban: {}'.format(e))
