#!/usr/bin/env python3

"""
Nemesida WAF API
Copyright (c) 2017-2025 Pentestit LLC - All Rights Reserved
Author: Romanov R.
"""

import base64
import json
import psycopg2
import re
import requests
import validators

from decouple import Config, RepositoryEnv
from multiprocessing import Lock
from netaddr import IPAddress
from pymemcache.client.base import Client
from time import sleep

from urls import set_dyn_sync_ban_uri
from logger import log

##
# URLs
##

# Remote server URI
auth_uri = 'https://nemesida-security.com/nw/auth/v3'
rl_desc_uri = 'https://nemesida-security.com/nw/rl/desc'
geoip_batch_uri = 'https://geoip.nemesida-security.com/geoip/batch/v1?geo_token='
geocc_memc_key = 'geocc_'
geoip_v3_memc_key = 'geoip_v3_'

# Misc.
waf_id_ce = 'COMMUNITYEDITION'
agents = ['dyn', 'mla', 'mlc', 'nws']
headers = {'User-agent': 'Nemesida WAF API'}

##
# Export settings
##

# init
proxy = {}
ec_storage_period = 30
settings_file = '/var/www/nw-api/settings.py'

try:

    # init config
    env_config = Config(RepositoryEnv(settings_file))

    # init database settings
    db_host = env_config.get('DB_HOST', default='127.0.0.1')
    db_port = env_config.get('DB_PORT', default='5432')
    db_name = env_config.get('DB_NAME', default='waf')
    db_user = env_config.get('DB_USER', default='nw_api')
    db_pass = env_config.get('DB_PASS', default='')
    ro_mode = env_config.get('RO_MODE', default=False, cast=bool)

    # database settings processing
    if not db_pass:
        log.warning(
            'Option "db_pass" is not set, make changes to the configuration file and restart the services'
        )
        exit()

    # init memcached
    memc_lock = Lock()
    memcached_host = env_config.get('MEMCACHED_HOST', default='127.0.0.1')
    memcached_port = env_config.get('MEMCACHED_PORT', default='11211')
    memc = Client((memcached_host, int(memcached_port)))

    # init proxy
    proxy_settings = env_config.get('PROXY', default='')

    # proxy settings processing
    if proxy_settings:
        try:

            # init
            proxy = proxy_settings

            # set schema if not exists
            if not re.match(r'^https?://', proxy_settings.lower()):
                proxy = 'http://' + proxy_settings

            # set proxy
            proxy = {'http': proxy, 'https': proxy}

        except Exception as err:
            log.error(
                'An error occurred while adding the schema for proxy-server: {}'.format(err)
            )

    # EC storage period
    ec_storage_period = int(env_config.get('EC_STORAGE_PERIOD', default=30))
    ec_storage_period = ec_storage_period if ec_storage_period > 0 else 0

except Exception as err:
    log.error(
        'An error occurred while exporting settings from file {}: {}'
        .format(settings_file, err)
    )
    exit()


##
# Functions
##

async def request_form_preprocess(url, request):
    try:

        # extract client IP
        agent_ip = get_agent_ip(request)

        # extract lic. key
        key = dict(await request.form()).get('key')
        if not key:
            log.error('An error occurred while extracting product key from {} to {}: no key found'.format(agent_ip, url))
            return False, 401

        # lic. key validation
        waf_id = get_waf_id(key, agent_ip)
        if not waf_id:
            return False, 401

        # response
        return True, [key, waf_id, agent_ip]

    except Exception as e:
        log.error('A error occurred while preprocessing request to {} in JSON format: {}'.format(url, e))
        return False, 503


async def request_json_preprocess(url, request):
    try:

        # extract client IP
        agent_ip = get_agent_ip(request)

        # extract data from the request
        try:
            data = json.loads(await request.body())
        except Exception:
            response = {'status': 'fail', 'description': 'The data is not in JSON format'}
            return False, response

        # convert all keys to lowercase
        try:
            data = {k.lower(): v for k, v in data.items()}
        except Exception as e:
            log.error('An error occurred while converting keys to lowercase from {} to {}: {} ({})'.format(agent_ip, url, e, data))

        # extract lic. key
        if url in [set_dyn_sync_ban_uri]:

            # waf_lic_key field
            try:
                param = 'waf_lic_key'
                if data.get(param):
                    key = data[param]
                else:
                    return False, 401
            except Exception as e:
                log.error('An error occurred while extract product key from {} to {}: {}' .format(agent_ip, url, e))
                return False, 401

        else:

            # extract lic. key
            key = data.get('key')
            if not key:
                log.error('An error occurred while extracting product key from {} to {}: no key found'.format(agent_ip, url))
                return False, 401

        # lic. key validation
        waf_id = get_waf_id(key, agent_ip)
        if not waf_id:
            return False, 401

        # response
        return True, [data, key, waf_id, agent_ip]

    except Exception as e:
        log.error('A error occurred while preprocessing request to {} in JSON format: {}'.format(url, e))
        return False, 503


def get_log_key(key):
    try:
        return 'none' if key == 'none' else 'x' * 26 + key[-6:]
    except Exception as e:
        log.error('An error occurred while processing key {}: {}'.format(key, e))
        return key


def get_agent_ip(request):

    # extract client IP
    try:
        return request.client.host
    except Exception:
        return ''


def init_db_connection():
    try:

        connect = psycopg2.connect(
            database=db_name,
            user=db_user,
            password=db_pass,
            host=db_host,
            port=db_port,
            connect_timeout=10
        )
        connect.set_session(autocommit=True)
        return connect

    except Exception as e:
        log.error('An error occurred while connecting to the database: {}'.format(e))
        return False


def close_db_connection(connect, cursor):
    try:
        cursor.close()
        connect.close()
    except Exception as e:
        log.error('An error occurred while closing database connection: {}'.format(e))


def get_auth(key, agent_ip):

    # init
    waf_id = ''
    tarif_id = 0
    models_num = 0
    geo_token = ''

    # init memcached
    memc_key = 'auth_{}'.format(key)

    ##
    # Search data in memcached
    ##

    try:

        # get data from the memcached
        with memc_lock:
            try:
                memc_bin = memc.get(memc_key)
            except Exception as e:
                memc_bin = None
                log.error('An error occurred while receiving data from the memcached: {}'.format(e))

        # if data in memcached
        if memc_bin:

            # processing memcached data
            data = memc_bin.decode('UTF-8')
            data = json.loads(data)

            # If lic. key is valid
            if data['status'] == 'success':

                # extract data
                waf_id = str(data['waf_id'])
                tarif_id = data['tarif_id']
                models_num = data['models_num']
                geo_token = data['geo_token']

        # return
        if waf_id:
            return waf_id, tarif_id, models_num, geo_token

    except Exception as e:
        log.error('An error occurred while processing WAF ID from the memcached: {}'.format(e))

    ##
    # If WAF ID not in memcached
    # receiving from the Nemesida WAF Auth
    ##

    while True:
        try:

            # init
            s = init_session()

            # send request
            result = s.post(
                auth_uri,
                headers=headers,
                proxies=proxy,
                data={'key': key},
                verify=True,
                timeout=15
            )
            status_code = result.status_code

            # processing
            if status_code == 201:
                try:

                    # init
                    data = json.loads(result.text)

                    # extract data
                    waf_id = str(data['waf_id'])
                    tarif_id = data['tarif_id']
                    models_num = data['models_num']
                    geo_token = data['geo_token']

                    # update WAF ID in memcached
                    if waf_id:
                        with memc_lock:
                            try:
                                memc.set(memc_key, json.dumps(data), 600)
                            except Exception as e:
                                log.error(
                                    'An error occurred while placing auth data for product key {} into the memcached (client {}): {}'
                                    .format(get_log_key(key), agent_ip, e)
                                )

                    # return
                    return waf_id, tarif_id, models_num, geo_token

                except Exception as e:
                    log.error(
                        'An error occurred while processing auth data for product key {} from client {}: {}'
                        .format(get_log_key(key), agent_ip, e)
                    )
                    sleep(1)

            elif status_code == 401:
                log.warning('Invalid product key {} received from client {}'.format(get_log_key(key), agent_ip))
                return waf_id, tarif_id, models_num, geo_token

        except Exception as e:
            log.error('An error occurred while processing the request from client {}: {}'.format(agent_ip, e))
            sleep(5)


def get_waf_id(key, agent_ip):
    try:
        waf_id, _, _, _ = get_auth(key, agent_ip)
        return waf_id
    except Exception as e:
        log.error('An error occurred while processing WAF ID: {}'.format(e))
        return False


def init_db():

    ##
    # Create new column if not exist
    ##

    try:

        log.info('The database is being initialized')

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        """
        Rename the sequence
        """

        # attack_ids
        query = """ ALTER SEQUENCE IF EXISTS attack_id_seq RENAME TO attack_ids """
        cursor.execute(query,)

        # scan_report_ids
        try:
            query = """ ALTER SEQUENCE IF EXISTS scan_report_id_seq RENAME TO scan_report_ids """
            cursor.execute(query,)
        except Exception as e:
            log.error('An error occurred while renaming sequence scan_report_id_seq: {}'.format(e))
            try:
                query = """ DROP SEQUENCE IF EXISTS scan_report_id_seq """
                cursor.execute(query,)
            except Exception as e:
                log.error('An error occurred while deleting sequence scan_report_id_seq: {}'.format(e))
                try:
                    query = """ DROP SEQUENCE IF EXISTS scan_report_ids """
                    cursor.execute(query,)
                    query = """ ALTER SEQUENCE IF EXISTS scan_report_id_seq RENAME TO scan_report_ids """
                    cursor.execute(query,)
                except Exception as e:
                    log.error('An error occurred while deleting sequence scan_report_ids: {}'.format(e))

        # modify sequence type
        for sequence in ['dyn_wl_ids', 'dyn_erl_ids']:
            try:
                query = """ ALTER SEQUENCE IF EXISTS {} AS integer """.format(sequence)
                cursor.execute(query,)
            except Exception as e:
                log.error('An error occurred while processing sequence {}: {}'.format(sequence, e))

        """
        Create new tables if not exists
        """

        fname = 'api.sql'
        log.info('Database processing using {}'.format(fname))
        cursor.execute(open(fname, 'r').read())

        # logging
        # log.info('The database has been processed successfully using {}'.format(fname))

        ##
        # Updating the tables
        ##

        log.info('Updating the database tables')

        """
        Table: attack
        """

        # Receive columns name
        query = """
            SELECT
                column_name
            FROM
                INFORMATION_SCHEMA.COLUMNS
            WHERE
                TABLE_NAME = 'attack'
            """
        cursor.execute(query,)
        res = cursor.fetchall()

        # rename content to body
        if res and ('content',) in res:
            query = """ ALTER TABLE attack RENAME COLUMN content TO body """
            cursor.execute(query,)

        # rename uri to url
        if res and ('uri',) in res:
            query = """ ALTER TABLE attack RENAME COLUMN uri TO url """
            cursor.execute(query,)

        # modify the columns
        query = """
            ALTER TABLE
                attack
            DROP COLUMN IF EXISTS
                bot,
            ADD COLUMN IF NOT EXISTS
                group_id text,
            ADD COLUMN IF NOT EXISTS
                blocked boolean,
            ADD COLUMN IF NOT EXISTS
                possible boolean,
            ADD COLUMN IF NOT EXISTS
                cc char(2) DEFAULT NULL,
            ALTER COLUMN
                url DROP NOT NULL,
            ALTER COLUMN
                param DROP NOT NULL,
            ALTER COLUMN
                other_headers DROP NOT NULL,
            ALTER COLUMN
                referer DROP NOT NULL,
            ALTER COLUMN
                mz DROP NOT NULL,
            ALTER COLUMN
                description DROP NOT NULL,
            ALTER COLUMN
                rule_id DROP NOT NULL,
            ALTER COLUMN
                rule_id SET DEFAULT 0,
            ADD COLUMN IF NOT EXISTS
                long_description text,
            ADD COLUMN IF NOT EXISTS
                request_type text
        """
        cursor.execute(query,)

        # drop indexes
        query = """ DROP INDEX IF EXISTS idx_mtp """
        cursor.execute(query,)

        # logging
        # log.info('Table "attack" has been successfully updated or no migration is required')

        """
        Table: scan_report
        """

        # add new column
        query = """ ALTER TABLE scan_report ADD COLUMN IF NOT EXISTS mz text """
        cursor.execute(query,)

        # drop NOT NULL for specific columns
        query = """
            ALTER TABLE
                scan_report
            ALTER COLUMN
                param DROP NOT NULL,
            ALTER COLUMN
                payload DROP NOT NULL,
            ALTER COLUMN
                type DROP NOT NULL,
            ALTER COLUMN
                url DROP NOT NULL,
            ALTER COLUMN
                data DROP NOT NULL
        """
        cursor.execute(query,)

        # logging
        # log.info('Table "scan_report" has been successfully updated or no migration is required')

        """
        Table: dyn_settings
        """

        # add new column
        query = """
            ALTER TABLE
                dyn_settings
            ADD COLUMN IF NOT EXISTS
                nwaf_b64_decode_disable text,
            ADD COLUMN IF NOT EXISTS
                nwaf_geoip_mla_disable text,
            ADD COLUMN IF NOT EXISTS
                nwaf_rfc_violation_wl text,
            ADD COLUMN IF NOT EXISTS
                nwaf_rfc_violation_lm text,
            ADD COLUMN IF NOT EXISTS
                nwaf_body_bin_exclude text,
            ADD COLUMN IF NOT EXISTS
                nwaf_log_request text,
            ADD COLUMN IF NOT EXISTS
                nwaf_antibot_mem integer,
            ADD COLUMN IF NOT EXISTS
                nwaf_antibot_url text,
            ADD COLUMN IF NOT EXISTS
                nwaf_antibot_url_wl text,
            ADD COLUMN IF NOT EXISTS
                nwaf_antibot_ip_wl text,
            ADD COLUMN IF NOT EXISTS
                nwaf_antibot_cc_wl text,
            ADD COLUMN IF NOT EXISTS
                nwaf_av_ip_lm text,
            ADD COLUMN IF NOT EXISTS
                nwaf_av_ip_wl text,
            ADD COLUMN IF NOT EXISTS
                nwaf_av_url_lm text,
            ADD COLUMN IF NOT EXISTS
                nwaf_av_url_wl text
        """
        cursor.execute(query,)

        # logging
        # log.info('Table "dyn_settings" has been successfully updated or no migration is required')

        """
        Table: openapi_url
        """

        # modify the columns
        query = """
            ALTER TABLE
                openapi_url
            ALTER COLUMN
                timestamp TYPE timestamp without time zone USING timestamp::timestamp without time zone
        """
        cursor.execute(query,)
        # log.info('Table "openapi_url" has been successfully updated or no migration is required')

        """
        Table: openapi_schemas
        """

        # modify the columns
        query = """
            ALTER TABLE
                openapi_schemas
            ALTER COLUMN
                timestamp TYPE timestamp without time zone USING timestamp::timestamp without time zone
        """
        cursor.execute(query,)

        # modify the openapi_url_idx index
        try:

            query = """
                SELECT
                    indexdef
                FROM
                    pg_indexes
                WHERE
                    schemaname = 'public'
                        AND
                    indexname = 'openapi_url_idx'
                        AND
                    (
                        indexdef LIKE '%USING btree (waf_id)'
                            OR
                        indexdef LIKE '%USING btree (waf_id, url)'
                    )
            """
            cursor.execute(query,)

            if cursor.fetchone():

                query = """ DROP INDEX IF EXISTS new_openapi_url_idx """
                cursor.execute(query,)

                query = """ CREATE UNIQUE INDEX new_openapi_url_idx ON openapi_url USING btree (waf_id, md5(url)) """
                cursor.execute(query,)

                query = """ DROP INDEX openapi_url_idx """
                cursor.execute(query,)

                query = """ ALTER INDEX new_openapi_url_idx RENAME TO openapi_url_idx """
                cursor.execute(query,)

                # logging
                # log.info('Table "openapi_schemas" has been successfully updated or no migration is required')

        except Exception as e:
            log.error('An error occurred while processing openapi_url_idx: {}'.format(e))

        """
        Table: openapi_schema_paths
        """

        # modify the columns
        query = """
            ALTER TABLE
                openapi_schema_paths
            ALTER COLUMN
                timestamp TYPE timestamp without time zone USING timestamp::timestamp without time zone
        """
        cursor.execute(query,)

        # modify the openapi_schema_paths_idx index
        try:

            query = """
                SELECT
                    indexdef
                FROM
                    pg_indexes
                WHERE
                    schemaname = 'public'
                        AND
                    indexname = 'openapi_schema_paths_idx'
                        AND
                    indexdef LIKE '%USING btree (waf_id, url, schema_name)'
            """
            cursor.execute(query,)

            if cursor.fetchone():

                query = """ DROP INDEX IF EXISTS new_openapi_schema_paths_idx """
                cursor.execute(query,)

                query = """ CREATE UNIQUE INDEX new_openapi_schema_paths_idx ON openapi_schema_paths USING btree (waf_id, md5(url), schema_name) """
                cursor.execute(query,)

                query = """ DROP INDEX openapi_schema_paths_idx """
                cursor.execute(query,)

                query = """ ALTER INDEX new_openapi_schema_paths_idx RENAME TO openapi_schema_paths_idx """
                cursor.execute(query,)

        except Exception as e:
            log.error('An error occurred while processing the openapi_schema_paths_idx: {}'.format(e))

        # logging
        # log.info('Table "openapi_schemas_path" has been successfully updated or no migration is required')

        """
        Table: openapi_exclude_paths
        """

        # modify the columns
        query = """
            ALTER TABLE
                openapi_exclude_paths
            ALTER COLUMN
                timestamp TYPE timestamp without time zone USING timestamp::timestamp without time zone,
            DROP COLUMN IF EXISTS
                id
        """
        cursor.execute(query,)

        # sequence processing
        try:
            query = """ DROP SEQUENCE IF EXISTS openapi_exclude_paths_ids """
            cursor.execute(query,)
        except Exception as e:
            log.error('An error occurred while deleting sequence openapi_exclude_paths_ids: {}'.format(e))

        # index processing
        try:

            # modify the index
            query = """
                SELECT
                    indexdef
                FROM
                    pg_indexes
                WHERE (
                    schemaname = 'public'
                        AND
                    indexname = 'openapi_exclude_paths_idx'
                        AND
                    indexdef LIKE '%USING btree (id)'
                )
            """
            cursor.execute(query,)

            if cursor.fetchone():

                query = """ DROP INDEX IF EXISTS new_openapi_exclude_paths_idx """
                cursor.execute(query,)

                query = """ CREATE UNIQUE INDEX new_openapi_exclude_paths_idx ON openapi_exclude_paths USING btree (waf_id) """
                cursor.execute(query,)

                query = """ DROP INDEX openapi_exclude_paths_idx """
                cursor.execute(query,)

                query = """ ALTER INDEX new_openapi_exclude_paths_idx RENAME TO openapi_exclude_paths_idx """
                cursor.execute(query,)

        except Exception as e:
            log.error('An error occurred while updating index openapi_exclude_paths_idx: {}'.format(e))

        # logging
        # log.info('Table "openapi_exclude_paths" has been successfully updated or no migration is required')

        """
        Table: ml
        """

        # drop table
        query = """ DROP TABLE IF EXISTS ml """
        cursor.execute(query,)

        # logging
        # log.info('Table "ml" has been successfully dropped or no migration is required')

        """
        Table: nginx_global
        """

        # modify the columns
        query = """
            ALTER TABLE
                nginx_global
            ALTER COLUMN
                resolver_timeout TYPE text,
            ALTER COLUMN
                client_max_body_size TYPE text
        """
        cursor.execute(query,)

        # logging
        # log.info('Table "nginx_global" has been successfully dropped or no migration is required')

        """
        Table: nginx_vhost
        """

        # modify the columns
        query = """
            ALTER TABLE
                nginx_vhost
            ALTER COLUMN
                ssl_session_timeout TYPE text,
            ADD COLUMN IF NOT EXISTS
                add_header text,
            ADD COLUMN IF NOT EXISTS
                error_page text
        """
        cursor.execute(query,)

        # logging
        # log.info('Table "nginx_vhost" has been successfully dropped or no migration is required')

        ##

        # modify the columns
        query = """
            ALTER TABLE
                mt_main
            ALTER COLUMN
                cc TYPE char(2),
            ALTER COLUMN
                cc SET DEFAULT NULL
        """
        cursor.execute(query,)

        # logging
        # log.info('Table "mt_main" has been successfully dropped or no migration is required')

        """
        Table: dyn_erl
        """

        # modify the columns
        query = """
            ALTER TABLE
                dyn_erl
            ADD COLUMN IF NOT EXISTS
                block_tor boolean DEFAULT false,
            ADD COLUMN IF NOT EXISTS
                block_proxy boolean DEFAULT false,
            ADD COLUMN IF NOT EXISTS
                no_ua boolean DEFAULT false,
            ADD COLUMN IF NOT EXISTS
                no_referer boolean DEFAULT false,
            ADD COLUMN IF NOT EXISTS
                url_x text,
            ADD COLUMN IF NOT EXISTS
                args_x text,
            ADD COLUMN IF NOT EXISTS
                body_x text,
            ADD COLUMN IF NOT EXISTS
                ua_x text,
            ADD COLUMN IF NOT EXISTS
                cookie_x text,
            ADD COLUMN IF NOT EXISTS
                referer_x text,
            ADD COLUMN IF NOT EXISTS
                other_headers_x text,
            ADD COLUMN IF NOT EXISTS
                zone_max_size text
        """
        cursor.execute(query,)

        # logging
        # log.info('Table "dyn_erl" has been successfully dropped or no migration is required')

        """
        Table: mlc_settings
        """

        # add new column
        query = """ ALTER TABLE mlc_settings ADD COLUMN IF NOT EXISTS brute__wl_ip text """
        cursor.execute(query,)

        # add new column
        query = """ ALTER TABLE mlc_settings ADD COLUMN IF NOT EXISTS brute__bf_detect text """
        cursor.execute(query,)

        # logging
        # log.info('Table "mlc_settings" has been successfully dropped or no migration is required')

        """
        Table: agent_monitoring
        """

        # add new column
        query = """ ALTER TABLE agent_monitoring ADD COLUMN IF NOT EXISTS systemMonitor text """
        cursor.execute(query,)

        # logging
        # log.info('Table "agent_monitoring" has been successfully dropped or no migration is required')

        ##

        log.info('The database initialization is complete')

        # close database connection
        connect.close()

    except Exception as e:
        log.error('An error occurred while updating the table structure: {}'.format(e))


def init_session():

    s = requests.Session()
    s.trust_env = False
    return s


def get_rldesc(rule_id):

    # init
    rldesc = ''
    rule_id = str(rule_id)

    try:

        # logging
        log.info('Updating description for RuleID {}'.format(rule_id))

        # init
        s = init_session()
        data = {'rule_id': rule_id}
        data = json.dumps(data)

        # receiving rule description
        state = s.post(
            rl_desc_uri,
            proxies=proxy,
            headers=headers,
            data=data,
            timeout=120
        )
        status_code = state.status_code

        # processing result
        if status_code != 200 and status_code != 201:
            log.error(
                'Incorrect server answer for RuleID {}: {}, text: {}'.format(rule_id, state.status_code, state.text)
            )
            return rldesc

        desc_dict = state.text
        desc_dict = json.loads(desc_dict)

        return desc_dict['desc']

    except Exception as e:
        log.error(
            'An error occurred in while processing description for rule {}: {}'.format(rule_id, e)
        )
        return rldesc


def get_geo(ips, waf_id):

    # init
    result = {}
    chunk_size = 100

    try:

        # receiving from the memcached
        for ip in ips:

            # search data in the memcached
            with memc_lock:
                try:
                    r = memc.get(geoip_v3_memc_key + str(ip))
                    result[ip] = json.loads(r) if r else {}
                except Exception as e:
                    log.error('An error occurred while receiving data from the memcached: {}'.format(e))

        # remove exists IP
        ips = [x for x in ips if not result.get(x)]

        # exit if all IPs with data in the result
        if not ips:
            return result

        # skip empty/CE WAF ID
        if not waf_id.strip() or (waf_id == waf_id_ce):
            return None

        # get GeoToken
        geo_token = get_geo_token(waf_id)
        if not geo_token:
            return None

        # init
        s = init_session()

        # chunk processing
        for x in range(0, len(ips), chunk_size):

            # init
            chunk = ips[x: x + chunk_size]

            # skip empty chunk
            if not chunk:
                continue

            # receiving GeoIP data
            response = s.post(
                geoip_batch_uri + geo_token,
                proxies=proxy,
                headers=headers,
                data=str(chunk),
                verify=True,
                timeout=30,
            )

            # processing status code
            if not re.match(r'^20', str(response.status_code)):
                log.error('An error occurred while processing data from GeoIP server: {}'.format(response))
                return None

            ##
            # JSON processing
            ##

            # load to JSON
            try:
                data = json.loads(response.text)
            except Exception as e:
                log.error('An error occurred while processing data from GeoIP server: {} ({})'.format(e, response.text))
                return None

            # if result is not empty/False
            if data:

                # update data in the memcached
                for k, v in data.items():

                    # set the memcached key
                    memc_key = geoip_v3_memc_key + str(k)

                    # skip private
                    if v == 'Private IP':
                        with memc_lock:
                            try:
                                memc.set(memc_key, '{}', 86400)  # 86400 - 1 d.
                            except Exception as e:
                                log.error('An error occurred while placing data into the memcached: {}'.format(e))

                    else:

                        # skip non-success
                        if (not isinstance(v, dict)) or (v.get('status') != 'success'):
                            continue

                        # update the result
                        result[k] = v.copy()

                        # update data in the memcached
                        try:
                            with memc_lock:
                                try:
                                    memc.set(memc_key, json.dumps(v), 86400)  # 86400 - 1 d.
                                except Exception as e:
                                    log.error('An error occurred while placing data into the memcached: {}'.format(e))
                        except Exception as e:
                            log.error('An error occurred while storing data for {}: {}'.format(k, e))

    except Exception as e:
        log.error('An error occurred in get_geo: {}'.format(e))

    # return the result
    return result


def geo_token_update(key, agent_ip):
    try:
        waf_id, _, _, geo_token = get_auth(key, agent_ip)
        put_geo_token_to_db(waf_id, geo_token)
    except Exception as e:
        log.error('An error occurred while processing GeoToken: {}'.format(e))


def get_geo_token(waf_id):
    try:

        # search GeoToken in the database
        query = """ SELECT geo_token FROM geo_tokens WHERE waf_id = %s """
        connect = init_db_connection()
        cursor = connect.cursor()
        cursor.execute(query, (waf_id,))
        row = cursor.fetchone()
        connect.close()

        # processing result
        return row[0] if row else ''

    except Exception as e:
        log.error(
            'An error occurred while processing GeoToken for WAF ID {} in the database: {}'.format(waf_id, e)
        )
        return ''


def put_geo_token_to_db(waf_id, geo_token):
    try:

        # store GeoToken to the database
        query = """ INSERT INTO geo_tokens (waf_id, geo_token) VALUES (%s, %s) ON CONFLICT DO NOTHING """
        connect = init_db_connection()
        cursor = connect.cursor()
        cursor.execute(query, (waf_id, geo_token,))
        connect.close()

    except Exception as e:
        log.error('An error occurred during storing data to {}: {}'.format(geo_token, e))


def attack_zone_replace(mz):
    try:

        # init
        result = []

        # replace conditions
        repl = {
            "uri": "URL",
            "args": "ARGS",
            "body": "BODY",
            "ua": "User-Agent",
            "cookie": "Cookie",
            "referer": "Referer",
            "other_headers": "HEADERS",
        }

        # processing for each zone in mz list
        for z in mz:

            # zone normalization
            for k, v in repl.items():
                z = z.replace(k, v)

            # update result list
            result.append(z)

        return result

    except Exception as e:
        log.error('An error occurred while mz normalization: {} ({})'.format(e, mz))
        return mz


def models_vhost_normalization(vhost):
    try:

        # init
        dot = r'^\.'
        sub = r'^\*\.'

        # *
        if vhost == '*':
            vhost = 'all'

        # .example.com
        elif re.search(dot, vhost):
            vhost = re.sub(dot, '', vhost)

        # *.example.com
        elif re.search(sub, vhost):
            vhost = re.sub(sub, 'sub_', vhost)

    except Exception as e:
        log.error('An error occurred while models_vhost_normalization (vhost: {}): {}'.format(vhost, e))

    return vhost


def idna_normalization(vhost):
    try:

        # convert to lowercase
        vhost = vhost.lower().strip()

        # dot
        if vhost.startswith('.'):

            # processing for non-latin vhost
            try:
                vhost = '.' + vhost[1:].encode('idna').decode('UTF-8')
            except Exception as e:
                log.error('An error occurred while IDNA normalization for virtual host {}: {}'.format(vhost, e))

        # other
        else:

            # processing for non-latin vhost
            try:
                vhost = vhost.encode('idna').decode('UTF-8')
            except Exception as e:
                log.error('An error occurred while IDNA normalization for virtual host {}: {}'.format(vhost, e))

    except Exception as e:
        log.error('An error occurred while normalizing vhost {}: {}'.format(vhost, e))

    return vhost


def is_base64(s):

    ##
    # Check is data in base64
    ##

    # data is not in Base64 if contains only numbers
    if re.match(r'^\d+={,2}$', s):
        return False

    # check is data in Base64
    try:
        return base64.b64encode(base64.b64decode(s.encode('UTF-8'))).decode('UTF-8', 'ignore') in [s, s + '=', s + '==']
    except Exception:
        return False


def is_private(ip):
    try:

        # init
        addr = IPAddress(ip)

        # multicast IP
        if addr.is_multicast():
            return True

        # if not global IP
        if not addr.is_global():
            return True

        # loopback/multicast validation
        if addr.is_loopback() or addr.is_multicast():
            return True

    except Exception as e:
        log.error('An error occurred while processing IP {}: {}'.format(ip, e))

    return False


def domain_validation(domain):
    try:

        # processing '*'
        if domain == '*':
            return True

        # remove '*.' or '.' for future domain validation
        domain = re.sub(r'^[*]?\.', '', domain)

        # processing localhost
        if domain == 'localhost':
            return True

        # check value
        try:
            return False if not validators.domain(domain) else True
        except Exception as e:
            log.error('An error occurred while validation domain {}: {}'.format(domain, e))
            return False

    except Exception as e:
        log.error('An error occurred in domain_validation: {}'.format(e))
        return False


def cc_upd_db(table):

    ##
    # Processing CC in attack
    ##

    try:

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # get the data
        query = """
            SELECT DISTINCT
                ip, waf_id
            FROM
                {}
            WHERE (
                cc IS NULL
                    AND
                waf_id != %s
                    AND
                waf_id != ''
            )
        """.format(table)
        cursor.execute(query, (waf_id_ce,))
        res = cursor.fetchall()

        # skip empty
        if not res:
            return

        # preprocessing the result
        waf_ids = set(x[1] for x in res if x)

        ##
        # Updating CC
        ##

        # processing IPs for each WAF ID
        for waf_id in waf_ids:

            # init
            private_ips = []

            # extrack IPs
            all_ips = set(x[0] for x in res if x and x[1] == waf_id)

            # get private IPs
            for ip in all_ips:
                if is_private(ip):
                    private_ips.append(ip)

            # remove private IPs from the public IPs
            public_ips = [x for x in all_ips if x not in private_ips]

            # put GeoIP CC to the memcached
            ipsp = get_geo(public_ips, waf_id)

            # update CC for each IP in the database
            for ip in all_ips:

                # private IP
                if ip in private_ips:
                    cc = ''

                else:

                    # get the CC
                    cc = ipsp.get(ip)

                    # skip empty
                    if not cc:
                        continue

                    # private IP in the memcached
                    if cc == 'Private IP':
                        cc = ''

                # update
                query = """ UPDATE {} SET cc = %s WHERE (ip = %s AND waf_id = %s) """.format(table)
                cursor.execute(query, (cc, ip, waf_id,))

            # del vars
            del all_ips, public_ips, ipsp

        # close database connection
        close_db_connection(connect, cursor)

    except Exception as e:
        log.error('An error occurred while updating CountryCode in "{}" table: {}'.format(table, e))


def k2lower(data):
    try:

        # dict processing
        if isinstance(data, dict):

            # convert key name to lowercase
            for k, v in data.copy().items():

                # dict
                if isinstance(v, dict):
                    del data[k]
                    data[str(k).lower()] = k2lower(v)

                # list
                elif isinstance(v, list):
                    del data[k]
                    data[k.lower()] = [x.lower() if isinstance(x, str) else k2lower(x) for x in v]

                # other
                else:
                    del data[k]
                    data[k.lower()] = v

        # list processing
        elif isinstance(data, list):
            data = [x.lower() if isinstance(x, str) else k2lower(x) for x in data]

    except Exception as e:
        log.error('An error occurred in k2lower: {}'.format(e))

    return data
