#!/usr/bin/env python3

"""
Nemesida WAF API
Copyright (c) 2017-2025 Pentestit LLC - All Rights Reserved
Author: Romanov R.
"""

import base64
import ipaddress

from core import agents
from core import ec_storage_period
from core import init_db_connection, close_db_connection
from core import is_base64

from logger import log

# Set event columns
set_event_cols = ['timestamp', 'ip', 'uuid', 'level', 'agent', 'category', 'description']

# Set the available values
levels = ['debug', 'info', 'warn', 'error', 'fatal']


def event_validation(k, v):
    try:

        if k == 'timestamp':
            if (not isinstance(v, int)) or (isinstance(v, bool)):
                return False, 'The value {} of parameter {} is not an integer'.format(v, k)

        elif k == 'ip':
            if not isinstance(v, str):
                return False, 'The value {} of parameter {} is not an IP address'.format(v, k)
            try:
                ipaddress.ip_address(v)
            except Exception:
                return False, 'The value {} of parameter {} is not an IP address'.format(v, k)

        elif k == 'uuid' or k == 'category':
            if not isinstance(v, str):
                return False, 'The value {} of parameter {} is not a string'.format(v, k)

        elif k == 'level':
            if v not in levels:
                return False, 'The value {} of parameter {} is incorrect'.format(v, k)

        elif k == 'agent':
            if v not in agents:
                return False, 'The value {} of parameter {} is incorrect'.format(v, k)

        elif k == 'description':
            if not is_base64(v):
                return False, 'The value {} of parameter {} is not in Base64'.format(v, k)

    except Exception as e:
        err_msg = 'An error occurred during event validation: {}/{}: {}'.format(k, v, e)
        log.error(err_msg)
        return False, err_msg

    return True, ''


def get_event(url, data, agent_ip, waf_id):
    try:

        ##
        # Receive data from the database
        ##

        # init
        response = {'status': 'success', 'count': 0, 'totalCount': 0, 'waf_id': waf_id, 'events': []}

        try:

            # other parameters processing
            for k in ['starttime', 'endtime', 'reverseorder', 'limit', 'offset', 'ip', 'uuid', 'agent', 'level', 'category']:
                try:

                    # limit/offset validation
                    if k in ['limit', 'offset']:
                        if k in data:
                            if (isinstance(data[k], bool)) or (not isinstance(data[k], int)):
                                response = {'status': 'fail', 'description': 'Incorrect value for {}'.format(k)}
                                return response

                    # reverseorder validation
                    if k in ['reverseorder']:
                        if (k in data) and (not isinstance(data[k], bool)):
                            response = {'status': 'fail', 'description': 'Incorrect value for {}'.format(k)}
                            return response

                    # timestamp validation and normalization
                    if k in ['starttime', 'endtime']:
                        if k in data:
                            if (isinstance(data[k], bool)) or (not isinstance(data[k], int)):
                                response = {'status': 'fail', 'description': 'Incorrect value for {}'.format(k)}
                                return response
                        else:
                            data[k] = 0 if k == 'starttime' else 9999999999

                    # list validation and normalization
                    if k in ['ip', 'uuid', 'agent', 'level', 'category']:
                        if (k in data) and (not isinstance(data[k], list)):
                            response = {'status': 'fail', 'description': 'Incorrect value for {}'.format(k)}
                            return response
                        else:
                            data[k] = '|'.join([str(x) for x in data[k]]) if data.get(k) else '%%'

                except Exception as e:
                    log.error('An error occurred while extracting data from {} to {}: {}'.format(agent_ip, url, e))
                    response = {'status': 'fail', 'description': '{}'.format(e)}
                    return response

            # extract the limit
            limit = data.get('limit')
            limit = limit if isinstance(limit, int) and (not isinstance(limit, bool)) and (limit > 0) else 0

            # offset wo limit
            if ('offset' in data) and (limit < 1):
                response = {'status': 'fail', 'description': 'Using parameter "offset" with parameter "limit" < 1 is not allowed'}
                return response

            # extract the offset
            offset = data.get('offset')
            offset = offset * limit if (limit and offset) else 0

            # order processing
            order = 'DESC' if data.get('reverseorder') else 'ASC'

            # init database connection
            connect = init_db_connection()
            cursor = connect.cursor()

            # init
            response['events'] = []

            # init
            qhdrs = """
                waf_id = %s
                    AND
                timestamp >= %s
                    AND
                timestamp <= %s
                    AND
                ip SIMILAR TO %s
                    AND
                uuid SIMILAR TO %s
                    AND
                agent SIMILAR TO %s
                    AND
                level SIMILAR TO %s
                    AND
                category SIMILAR TO %s
            """
            qargs = (
                waf_id,
                int(data['starttime']),
                int(data['endtime']),
                data['ip'],
                data['uuid'],
                data['agent'],
                data['level'],
                data['category'],
            )

            # total count
            query = """ SELECT COUNT(*) FROM events WHERE {} """.format(qhdrs)
            cursor.execute(query, qargs)
            response['totalCount'] = cursor.fetchone()[0]

            # search the data

            query = """
                SELECT
                    id, ip, uuid, agent, level, category, timestamp, CONVERT_FROM(description, 'utf-8')
                FROM
                    events
                WHERE {}
                ORDER BY id {}
            """.format(qhdrs, order)

            if limit:
                query += ' LIMIT %s'
                qargs += (limit,)

            if offset:
                query += ' OFFSET %s'
                qargs += (offset,)

            cursor.execute(query, qargs)
            res = cursor.fetchall()

            # del vars
            del qhdrs, qargs, data

            # processing result
            if res:

                # update the response data
                response['count'] = len(res)

                # processing the result
                for item in tuple(res):

                    # item processing
                    r = {
                        'id': item[0],
                        'ip': item[1],
                        'uuid': item[2],
                        'agent': item[3],
                        'level': item[4],
                        'category': item[5],
                        'timestamp': item[6],
                        'description': item[7]
                    }

                    # update the response data
                    response['events'].append(r)

                    # del var
                    del r

            # del vars
            del res

            # close database connection
            close_db_connection(connect, cursor)

        except Exception as e:
            log.error('An error occurred while processing data from the database in {}: {}'.format(url, e))

        # response
        return response

    except Exception as e:
        log.error('An error occurred in get_event: {}'.format(e))


def set_event(events, waf_id):
    try:

        ##
        # Processing JSON
        ##

        # Storage period more than 0 day
        if ec_storage_period > 0:

            # init
            cols = set_event_cols

            # init database connection
            connect = init_db_connection()
            cursor = connect.cursor()

            # processing each item in the list
            for event in events:
                try:

                    # convert all keys to lowercase
                    event = {k.lower(): v for k, v in event.items()}

                    # required key check
                    for req_k in cols:
                        if req_k not in event.keys():
                            err_msg = 'The required key "{}" is missing'.format(req_k)
                            log.error(err_msg)
                            response = {'status': 'fail', 'description': err_msg}
                            return response

                    # processing item
                    for k in cols:

                        # init
                        v = event[k]

                        # value validations
                        r = event_validation(k, v)
                        if not r[0]:
                            response = {'status': 'fail', 'description': r[1]}
                            return response

                    # add WAF ID
                    event['waf_id'] = waf_id

                    # decode description
                    try:
                        event['description'] = base64.b64decode(event['description']).decode('UTF-8', 'ignore').encode('UTF-8', 'ignore')
                    except Exception as e:
                        err = 'An error occurred while decoding the description: {}'.format(e)
                        response = {'status': 'fail', 'description': err}
                        return response

                    # init
                    k = list(event.keys())
                    v = list(event.values())

                    # insert the data
                    query = """ INSERT INTO events ({}) VALUES %s """.format(','.join(k))
                    cursor.execute(query, (tuple(v),))

                except Exception as e:
                    log.error('An error occurred while processing the event: {} ({})'.format(e, event))
                    close_db_connection(connect, cursor)
                    response = {'status': 'fail', 'description': str(e)}
                    return response

            # close database connection
            close_db_connection(connect, cursor)

        # response
        response = {'status': 'success'}
        return response

    except Exception as e:
        log.error('An error occurred in set_event: {}'.format(e))


def del_event(url, data, agent_ip, waf_id):
    try:

        ##
        # Data validation
        ##

        # processing the parameters
        for param in ['timestamp', 'id', 'ip', 'uuid', 'agent', 'level', 'category']:
            try:

                # int validation and normalization
                if param in ['timestamp']:
                    if data.get(param):
                        if isinstance(data[param], bool):
                            response = {'status': 'fail', 'description': 'Value of "{}" contains incorrect data'.format(param)}
                            return response
                        if not isinstance(data[param], int):
                            response = {'status': 'fail', 'description': 'Value of "{}" contains incorrect data'.format(param)}
                            return response
                    else:
                        data[param] = 9999999999

                # list validation and normalization
                if param in ['id', 'ip', 'uuid', 'agent', 'level', 'category']:
                    if (param in data) and (not isinstance(data[param], list)):
                        response = {'status': 'fail', 'description': 'Value of "{}" contains incorrect data'.format(param)}
                        return response
                    else:
                        if not data.get(param):
                            data[param] = '%%'
                        else:
                            data[param] = '|'.join([str(x) for x in data[param]])

            except Exception as e:
                log.error('An error occurred while extracting data from {} to {}: {}'.format(agent_ip, url, e))
                response = {'status': 'fail', 'description': str(e)}
                return response

        ##
        # Processing JSON
        ##

        # init
        response = {'status': 'success', 'deleted': 0}

        try:

            # init database connection
            connect = init_db_connection()
            cursor = connect.cursor()

            # insert the data
            query = """
                DELETE FROM
                    events
                WHERE (
                    waf_id = %s
                        AND
                    timestamp <= %s
                        AND
                    CAST (id AS TEXT) SIMILAR TO %s
                        AND
                    ip SIMILAR TO %s
                        AND
                    uuid SIMILAR TO %s
                        AND
                    agent SIMILAR TO %s
                        AND
                    level SIMILAR TO %s
                        AND
                    category SIMILAR TO %s
                )
                RETURNING id
            """
            cursor.execute(
                query, (
                    waf_id,
                    data['timestamp'],
                    data['id'],
                    data['ip'],
                    data['uuid'],
                    data['agent'],
                    data['level'],
                    data['category'],
                )
            )
            res = cursor.fetchall()

            # close database connection
            close_db_connection(connect, cursor)

            # processing result
            response['deleted'] = len(res)

        except Exception as e:
            log.error('An error occurred while deleting the events: {}'.format(e))
            response = {'status': 'fail', 'description': str(e)}
            return response

        # close database connection
        close_db_connection(connect, cursor)

        # response
        return response

    except Exception as e:
        log.error('An error occurred in del_event: {}'.format(e))
