#!/usr/bin/env python3

"""
Nemesida WAF API
Copyright (c) 2017-2025 Pentestit LLC - All Rights Reserved
Author: Romanov R.
"""

import base64
import json
import os
import psycopg2
import re
import shutil

from ast import literal_eval
from time import time_ns
from zipfile import ZipFile

from core import domain_validation
from core import idna_normalization
from core import init_db_connection, close_db_connection
from core import models_vhost_normalization

from logger import log


def training_vhost_processing(vhost, waf_id,):
    try:

        ##
        # Search vhost in local DB
        ##

        query = """ SELECT vhosts_list FROM vhosts_list WHERE waf_id = %s LIMIT 1 """
        connect = init_db_connection()
        cursor = connect.cursor()
        cursor.execute(query, (waf_id,))
        res = cursor.fetchone()
        connect.close()

        # response
        return True if res and (vhost in literal_eval(res[0])) else False

    except Exception as e:
        log.error('An error occurred in training_vhost_processing: {}'.format(e))

    # default response
    return False


def get_models(data, waf_id):
    try:

        # load models list from agent
        a_models_dict = json.loads(data)

        # if check status request
        if 'is_active' in a_models_dict.keys():
            response = {'is_active': True}
            return 200, json.dumps(response)

        ##
        # Get models list from the database
        ##

        # init
        l_models_dict = {}

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # receiving vhosts and hashes for all models by WAF ID
        query = """ SELECT vhost, hash FROM models WHERE (waf_id = %s) """
        cursor.execute(query, (waf_id,))
        res = cursor.fetchall()

        # generate db models dict
        for row in res:
            vhost = str(row[0])
            shash = str(row[1])
            l_models_dict[vhost] = shash

        ##
        # Convert dict to list
        ##

        # list from remote agent
        a_models_list = []
        for k in a_models_dict.keys():
            k = idna_normalization(k)
            a_models_list.append(str(k) + ':' + str(a_models_dict[k]))

        # list from the database
        l_models_list = []
        for k in l_models_dict.keys():
            k = idna_normalization(k)
            l_models_list.append(str(k) + ':' + str(l_models_dict[k]))

        # if models exists in the database
        if l_models_dict:

            # get models list, that the remote client does not have
            models_list = list(set(l_models_list) - set(a_models_list))

            # if unique models list is not empty, get models for first list element by hash
            if models_list:

                # init
                vhost = str(models_list[0].split(':')[0])
                shash = str(models_list[0].split(':')[1])

                # get models by vhost and hash
                query = """ SELECT model FROM models WHERE (vhost = %s AND hash = %s AND waf_id = %s) """
                cursor.execute(query, (vhost, shash, waf_id,))
                res = cursor.fetchone()
                close_db_connection(connect, cursor)

                # if models exist
                if res:
                    return 202, base64.b64encode(res[0])
                else:
                    return 304, ''

            # if the all models from the database are already on the remote side (MLA/MLC)
            else:

                # close database connection
                close_db_connection(connect, cursor)

                # get unique models list, that the db does not have
                models_list = list(set(a_models_list) - set(l_models_list))

                # response
                if len(models_list):
                    return 203, str(models_list)
                else:
                    return 304, ''

        # models don't exist in the database
        else:

            # close database connection
            close_db_connection(connect, cursor)

            # get unique models list, that the db does not have
            models_list = list(set(a_models_list) - set(l_models_list))

            # if there are models on a remote server. (MLA/MLC)
            if a_models_dict:
                return 203, str(models_list)
            else:
                return 304, ''

    except Exception as e:
        log.error('An error occurred in get_models: {}'.format(e))
        return 503


def set_models(url, vhost, shash, model, agent_ip, waf_id):
    try:

        # if models in the request:
        if model:
            try:

                # idna normalization
                vhost = idna_normalization(vhost)

                # init database connection
                connect = init_db_connection()
                cursor = connect.cursor()

                # search in the database
                query = """ SELECT vhost FROM models WHERE (vhost = %s AND waf_id = %s) """
                cursor.execute(query, (vhost, waf_id,))
                res = cursor.fetchone()

                # if models are existed
                if res:

                    # update in the database
                    query = """
                        UPDATE
                            models
                        SET
                            (vhost, waf_id, model, hash) = (%s, %s, %s, %s)
                        WHERE
                            (vhost = %s AND waf_id = %s)
                    """
                    cursor.execute(query, (vhost, waf_id, psycopg2.Binary(model), shash, vhost, waf_id,))

                    # logging
                    log.info('Models for {} with WAF ID {} received from client {}'.format(vhost, waf_id, agent_ip,))

                # if models doesn't exists
                else:

                    # insert into the database
                    query = """ INSERT INTO models (vhost, waf_id, model, hash) VALUES (%s, %s, %s, %s) """
                    cursor.execute(query, (vhost, waf_id, psycopg2.Binary(model), shash,))

                    # logging
                    log.info('Models for {} with WAF ID {} received from client {}'.format(vhost, waf_id, agent_ip,))

                # close database connection
                close_db_connection(connect, cursor)

                # response
                return 202

            except Exception as e:
                log.error('An error occurred in {} while updating/storing models in the database received from client {}: {}'.format(url, agent_ip, e))
                return 500

        else:

            # response
            return 401

    except Exception as e:
        log.error('An error occurred in set_models: {}'.format(e))
        return 503


def del_models(vhost, waf_id):
    try:

        # idna normalization
        vhost = idna_normalization(vhost)

        # vhost normalization
        vhostn = models_vhost_normalization(vhost)

        # delete models
        query = """ DELETE FROM models WHERE (vhost = %s) AND (waf_id = %s) """
        connect = init_db_connection()
        cursor = connect.cursor()
        cursor.execute(query, (vhostn, waf_id,))
        close_db_connection(connect, cursor)

        # logging
        log.info('The model of virtual host {} has been deleted'.format(vhost))

        # processing result
        response = {'status': 'success'}
        return response

    except Exception as e:
        log.error('An error occurred in del_models: {}'.format(e))


def rep_models(src, dst, agent_ip, waf_id):
    try:

        # idna normalization
        src = idna_normalization(src)
        dst = idna_normalization(dst)

        ##
        # Processing vhost status
        ##

        is_vhost_exists = training_vhost_processing(dst, waf_id)
        if not is_vhost_exists:
            response = {'status': 'fail', 'description': 'The virtual host was not found, set virtual host using Cabinet/API'}
            return response

        ##
        # Copy vhost to dst model with the same WAF ID
        ##

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # vhost normalization
        srcn = models_vhost_normalization(src)
        dstn = models_vhost_normalization(dst)

        # logging
        log.info(
            'Request to copy models for {} to {} with WAF ID {} received from client {}'
            .format(src, dst, waf_id, agent_ip)
        )

        # search the model
        query = """ SELECT model, hash, timestamp FROM models WHERE (vhost = %s AND waf_id = %s) """
        cursor.execute(query, (srcn, waf_id,))
        res = cursor.fetchone()

        # processing result
        if not res:
            close_db_connection(connect, cursor)
            response = {'status': 'fail', 'description': 'No model found'}
            return response

        ##
        # Update filenames in models.zip
        ##

        # init
        smodels = res[0]
        shash = res[1]
        sts = res[2]
        del res

        # init models path
        tmp_dir = os.path.join('/tmp', str(time_ns()))
        smodels_path = os.path.join(tmp_dir, 'smodels.zip')
        dmodels_path = os.path.join(tmp_dir, 'dmodels.zip')

        # create temporary directory
        if not os.path.exists(tmp_dir):
            os.mkdir(tmp_dir)

        # save src models
        with open(smodels_path, 'wb') as f:
            f.write(smodels)

        # init
        smodels = ZipFile(smodels_path, 'r')
        dmodels = ZipFile(dmodels_path, 'w')

        # update model's files
        for file in smodels.filelist:

            # save files
            if file.filename.endswith('_rf.ml'):
                dmodels.writestr(dstn + '_rf.ml', smodels.read(file.filename))
            if file.filename.endswith('_vect.ml'):
                dmodels.writestr(dstn + '_vect.ml', smodels.read(file.filename))

        # close file
        smodels.close()
        dmodels.close()

        # load dst models
        with open(dmodels_path, 'rb') as f:
            dmodels = f.read()

        # insert new model
        query = """
            INSERT INTO models (
                vhost, waf_id, model, hash, timestamp
            ) VALUES
                (%s, %s, %s, %s, %s)
            ON CONFLICT
                (vhost, waf_id)
            DO UPDATE SET
                (model, hash, timestamp) = (%s, %s, %s)
        """
        cursor.execute(
            query, (
                dstn, waf_id,
                psycopg2.Binary(dmodels), shash, sts,
                psycopg2.Binary(dmodels), shash, sts,
            )
        )

        # logging
        log.info('The model of virtual host {} is created from the model of virtual host {}'.format(dst, src))

        # del vars
        del smodels, dmodels

        # del tmp_dir
        if os.path.exists(tmp_dir):
            shutil.rmtree(tmp_dir)

        ##
        # Update training status
        ##

        # init
        duration = 0
        complete = True
        progress = 100

        # get training status
        query = """ SELECT duration, complete, progress FROM training WHERE (vhost = %s AND waf_id = %s) """
        cursor.execute(query, (srcn, waf_id,))
        res = cursor.fetchone()

        # processing result
        if res:
            duration = res[0]
            complete = res[1]
            progress = res[2]

        # insert a training status
        query = """
            INSERT INTO
                training (vhost, waf_id, duration, complete, progress)
            VALUES
                (%s, %s, %s, %s, %s)
            ON CONFLICT
                (vhost, waf_id)
            DO UPDATE SET
                (duration, complete, progress) = (%s, %s, %s)
        """
        cursor.execute(
            query, (
                dstn, waf_id,
                duration, complete, progress,
                duration, complete, progress,
            )
        )

        # del vars
        del res

        # close database connection
        close_db_connection(connect, cursor)

        # processing result
        response = {'status': 'success'}
        return response

    except Exception as e:
        log.error('An error occurred in rep_models: {}'.format(e))


def get_models_list(waf_id):
    try:

        # receiving vhost for all models by waf_id
        query = """ SELECT vhost FROM models WHERE (waf_id = %s) """
        connect = init_db_connection()
        cursor = connect.cursor()
        cursor.execute(query, (waf_id,))
        res = cursor.fetchall()
        close_db_connection(connect, cursor)

        # get models list as list
        response = [i[0] for i in res]

        # model's name normalization
        for i, r in enumerate(response.copy()):
            if r == 'all':
                response[i] = '*'
            elif r.startswith('sub_'):
                response[i] = r.replace('sub_', '*.')

        # sort result
        response.sort()

        # processing result
        response = [] if not response else response

        # response
        return response

    except Exception as e:
        log.error('An error occurred in get_models_list: {}'.format(e))


def get_training(vhost, waf_id):
    try:

        # get from the database
        query = """ SELECT duration, complete, progress FROM training WHERE (vhost = %s AND waf_id = %s) """
        connect = init_db_connection()
        cursor = connect.cursor()
        cursor.execute(query, (vhost, waf_id,))
        res = cursor.fetchone()
        close_db_connection(connect, cursor)

        # result processing
        if not res:
            response = {'status': 'none'}
            return response

        else:

            # processing duration (default 4 days)
            duration = 4 if res[0] is None else res[0]

            # processing complete/progress
            complete = literal_eval(str(res[1]))
            progress = 0 if res[2] is None else res[2]

            # response
            response = {'status': 'success', 'duration': duration, 'complete': complete, 'progress': progress}
            return response

    except Exception as e:
        log.error('An error occurred in get_training: {}'.format(e))


def get_training_v2(waf_id):
    try:

        ##
        # Extract training info of vhost from the database
        ##

        # init response
        response = {'status': 'success', 'training': []}

        # get training status
        query = """ SELECT vhost, duration, complete, progress FROM training WHERE (waf_id = %s) """
        connect = init_db_connection()
        cursor = connect.cursor()
        cursor.execute(query, (waf_id,))
        res = cursor.fetchall()
        close_db_connection(connect, cursor)

        # processing result
        if not res:
            return response

        else:

            # processing the items
            for item in res:

                # extract vhost
                vhost = item[0]

                # skip items with non-normalized vhost
                if vhost != idna_normalization(vhost):
                    continue

                # processing duration (default 4 days)
                duration = 4 if item[1] is None else int(item[1])

                # extract complete
                complete = item[2]

                # processing progress
                progress = 0 if item[3] is None else int(item[3])

                # update the result
                response['training'].append({
                    'vhost': vhost,
                    'duration': duration,
                    'complete': complete,
                    'progress': progress
                })

            # response
            return response

    except Exception as e:
        log.error('An error occurred in get_training_v2: {}'.format(e))


def set_training(url, data, vhost, agent_ip, waf_id):
    try:

        ##
        # vhost normalization
        ##

        # init
        ovhost = vhost

        # idna normalization
        vhost = idna_normalization(vhost)

        # all
        if vhost == 'all':
            vhost = '*'

        # *.example.com
        if re.search(r'^sub_', vhost):
            vhost = re.sub(r'^sub_', '*.', vhost)

        # update training_data
        training_data = {'vhost': vhost, 'waf_id': waf_id}

        ##
        # Extract training status/duration
        ##

        # init
        rkeys = ['duration', 'complete']

        # extract data
        for k in rkeys:
            try:

                # extract data
                v = data.get(k)

                # skip empty
                if v is None:
                    continue

                # data validation
                if k == 'duration':

                    try:
                        duration = int(str(v))
                    except Exception:
                        response = {'status': 'fail', 'description': 'Parameter "duration" is not an integer'}
                        return response

                    # check duration
                    if duration < 1:
                        response = {'status': 'fail', 'description': 'Parameter "duration" cannot be less than 1'}
                        return response

                # update training_data if value is not empty
                training_data[k] = v

            except Exception:
                pass

        ##
        # Check required keys in training_data
        ##

        if not [k for k in list(training_data.keys()) if k in rkeys]:
            log.error('An error occurred in {} from {}: complete and duration are missing'.format(url, agent_ip,))
            return

        ##
        # Processing vhost status
        ##

        is_vhost_exists = training_vhost_processing(vhost, waf_id)
        if not is_vhost_exists:
            response = {'status': 'fail', 'description': 'The virtual host was not found, set virtual host using Cabinet/API'}
            return response
        ##
        # Insert/update data in the database
        ##

        try:

            # delete keys with None value
            training_data = {k: v for k, v in training_data.items() if v is not None}

            # extract data as tuple
            k = ','.join(training_data.keys())
            v = tuple(training_data.values())

            # init database connection
            connect = init_db_connection()
            cursor = connect.cursor()

            # insert the data
            query = """ INSERT INTO training ({}) VALUES %s ON CONFLICT (vhost, waf_id) DO UPDATE SET ({}) = %s """.format(k, k)
            cursor.execute(query, (v, v,))

            # delete non-normalized vhost
            if ovhost != vhost:
                query = """ DELETE FROM training WHERE (vhost = %s) AND (waf_id = %s) """
                cursor.execute(query, (ovhost, waf_id,))

            # close database connection
            close_db_connection(connect, cursor)

            # response
            response = {'status': 'success'}
            return response

        except Exception as e:
            log.error('An error occurred in {} while storing data: {}'.format(url, e))

    except Exception as e:
        log.error('An error occurred in set_training: {}'.format(e))


def set_training_v2(data, vhost, waf_id):
    try:

        # idna normalization
        vhost = idna_normalization(vhost)

        ##
        # Extract training status/duration
        ##

        # init
        training_data = {'vhost': vhost, 'waf_id': waf_id}

        # data processing
        for k in data.keys():
            try:

                # extract data
                v = data.get(k)

                # skip empty
                if v is None:
                    continue

                # duration
                if k == 'duration':
                    if (not isinstance(v, bool)) and (isinstance(v, int)) and (v > 0):
                        training_data[k] = v
                    else:
                        response = {'status': 'fail', 'description': 'Incorrect value for {}'.format(k)}
                        return response

                # complete
                elif k == 'complete':
                    if isinstance(v, bool):
                        training_data[k] = v
                    else:
                        response = {'status': 'fail', 'description': 'Incorrect value for {}'.format(k)}
                        return response

                # other
                else:
                    continue

            except Exception:
                pass

        ##
        # Check required keys in training_data
        ##

        # init
        rkeys = ['duration', 'complete']

        # is any of rkeys in the data
        if not [k for k in training_data.keys() if k in rkeys]:
            msg = 'A required key ({}) is missing'.format('/'.join(rkeys))
            response = {'status': 'fail', 'description': msg}
            return response

        ##
        # vhost normalization
        ##

        # init
        ovhost = vhost

        # idna normalization
        vhost = idna_normalization(vhost)

        # all
        if vhost == 'all':
            vhost = '*'

        # *.example.com
        if re.search(r'^sub_', vhost):
            vhost = re.sub(r'^sub_', '*.', vhost)

        ##
        # Processing vhost status
        ##

        is_vhost_exists = training_vhost_processing(vhost, waf_id)
        if not is_vhost_exists:
            response = {'status': 'fail', 'description': 'The virtual host was not found, set virtual host using Cabinet/API'}
            return response

        ##
        # Insert/update data in the database
        ##

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # extract data as tuple
        k = ','.join(training_data.keys())
        v = tuple(training_data.values())

        # insert the data
        query = """ INSERT INTO training ({}) VALUES %s ON CONFLICT (vhost, waf_id) DO UPDATE SET ({}) = %s """.format(k, k)
        cursor.execute(query, (v, v,))

        # delete non-normalized vhost
        if ovhost != vhost:
            query = """ DELETE FROM training WHERE (vhost = %s) AND (waf_id = %s) """
            cursor.execute(query, (ovhost, waf_id,))

        # close database connection
        close_db_connection(connect, cursor)

        # response
        response = {'status': 'success'}
        return response

    except Exception as e:
        log.error('An error occurred in set_training_v2: {}'.format(e))


def set_training_progress(url, progress, vhost, waf_id):
    try:

        # idna normalization
        vhost = idna_normalization(vhost)

        # all
        if vhost == 'all':
            vhost = '*'

        # *.example.com
        if re.search(r'^sub_', vhost):
            vhost = re.sub(r'^sub_', '*.', vhost)

        ##
        # Processing vhost status
        ##

        is_vhost_exists = training_vhost_processing(vhost, waf_id)
        if not is_vhost_exists:
            response = {'status': 'fail', 'description': 'The virtual host was not found, set virtual host using Cabinet/API'}
            return response

        ##
        # Insert/update data in the database
        ##

        try:

            # update in the database
            query = """ UPDATE training SET progress = %s WHERE (vhost = %s AND waf_id = %s) """
            connect = init_db_connection()
            cursor = connect.cursor()
            cursor.execute(query, (progress, vhost, waf_id,))
            close_db_connection(connect, cursor)

            # response
            response = {'status': 'success'}
            return response

        except Exception as e:
            log.error('An error occurred in {} while storing data: {}'.format(url, e))

    except Exception as e:
        log.error('An error occurred in set_training_progress: {}'.format(e))


def get_vhosts_list(waf_id):
    try:

        # search the data in database
        query = """ SELECT vhosts_list FROM vhosts_list WHERE (waf_id = %s) """
        connect = init_db_connection()
        cursor = connect.cursor()
        cursor.execute(query, (waf_id,))
        res = cursor.fetchone()
        close_db_connection(connect, cursor)

        # processing result
        response = {
            'status': 'success',
            'vhosts_list': str(res[0]) if res else '[]'
        }

        # response
        return response

    except Exception as e:
        log.error('An error occurred in get_vhosts_list: {}'.format(e))


def set_vhosts_list(url, data, models_num, agent_ip, waf_id):
    try:

        # extract vhosts_list
        try:

            # extract value
            vhosts_list = data['vhosts_list']

            # vhosts_list size validation
            if len(vhosts_list) > models_num:
                desc = (
                    'The number of virtual hosts ({}) exceeds the number of available models ({}) for your tariff plan'
                    .format(len(vhosts_list), models_num)
                )
                response = {'status': 'fail', 'description': desc}
                return response

            # block using example.com with .example.com
            res = [x for x in vhosts_list if '.' + x in vhosts_list]
            if res:
                desc = 'Simultaneous use of virtual host {} and .{} is prohibited'.format(res[0], res[0])
                response = {'status': 'fail', 'description': desc}
                return response

            # vhost validation
            for vhost in vhosts_list:
                if (not vhost) or (not domain_validation(vhost)):
                    desc = 'Incorrect virtual host: {}'.format(vhost)
                    response = {'status': 'fail', 'description': desc}
                    return response

            ##
            # vhosts_list normalization
            ##

            # init
            nl = []

            # processing vhosts_list
            for v in vhosts_list:

                # skip empty
                if not v:
                    continue

                # idna normalization
                v = idna_normalization(v)

                # *
                if v == 'all':
                    nl.append('*')

                # *.example.com
                elif re.search(r'^sub_', v):
                    nl.append(re.sub(r'^sub_', '*.', v))

                else:
                    nl.append(v)

            # update vhosts_list
            vhosts_list = list(set(nl))

            # then sort list by length
            vhosts_list.sort(key=len, reverse=True)

            # replace single quotes
            vhosts_list = str(vhosts_list).replace("'", '"')

        except Exception as e:
            log.error('An error occurred while processing vhosts_list from {} to {}: {}'.format(agent_ip, url, e))
            return

        ##
        # Save vhosts_list to the database
        ##

        try:

            # insert the data into database
            query = """
                INSERT INTO vhosts_list (
                    vhosts_list, waf_id
                ) VALUES (
                    %s, %s
                ) ON CONFLICT (
                    waf_id
                ) DO UPDATE SET
                    vhosts_list = %s
            """
            connect = init_db_connection()
            cursor = connect.cursor()
            cursor.execute(query, (vhosts_list, waf_id, vhosts_list,))
            close_db_connection(connect, cursor)

            # response
            response = {'status': 'success'}
            return response

        except Exception as e:

            # response
            response = {'status': 'fail', 'description': str(e)}
            return response

    except Exception as e:
        log.error('An error occurred in set_vhosts_list: {}'.format(e))
