#!/usr/bin/env python3

"""
Nemesida WAF API
Copyright (c) 2017-2025 Pentestit LLC - All Rights Reserved
Author: Romanov R.
"""

import re

from binascii import hexlify
from datetime import datetime
from psycopg2.extras import RealDictCursor
from time import time

from core import cc_upd_db
from core import init_db_connection, close_db_connection

from logger import log


def lower_dict_keys(data_dict):
    return {key.lower() if isinstance(key, str) else key: value for key, value in data_dict.items()}


def mt_get_export(waf_id):

    # init
    res = ()

    try:

        query = """ SELECT * FROM mt_export WHERE waf_id = %s """
        connect = init_db_connection()
        cursor = connect.cursor(cursor_factory=RealDictCursor)
        cursor.execute(query, (waf_id,))
        res = cursor.fetchall()
        connect.close()

    except Exception as e:
        log.error('An error occurred while processing data in mt_export for WAF ID {}: {}'.format(waf_id, e))

    return res


def mt_get_export_md5(waf_id):

    # init
    res = '0'

    try:

        query = """ SELECT md5(CAST(array_agg(mt_export.*) AS TEXT)) FROM mt_export WHERE waf_id = %s """
        connect = init_db_connection()
        cursor = connect.cursor(cursor_factory=RealDictCursor)
        cursor.execute(query, (waf_id,))
        res = cursor.fetchall()
        connect.close()

        if res and res[0]['md5']:
            res = str(res[0]['md5'])
        else:
            res = '0'

    except Exception as e:
        log.error('An error occurred while processing hash in mt_export for WAF ID {}: {}'.format(waf_id, e))

    return res


def get_mt_attack(url, data, agent_ip, waf_id):
    try:

        # Update the CC
        cc_upd_db('mt_main')

        ##
        # Receive data from the database
        ##

        # init
        response = {'status': 'success', 'count': 0, 'totalCount': 0, 'waf_id': waf_id, 'attacks': []}

        # init query filter headers/values
        qhdrs = ['waf_id = %s']
        qargs = [waf_id]

        # remove the key
        del data['key']

        # init parameters type
        mgmt_param = [
            'limit',
            'offset',
            'bodylimit',
        ]
        int_param = [
            'starttime',
            'endtime',
        ]
        bool_param = [
            'reverseorder',
        ]
        list_s_param = [
            'bt',
            'ip',
            'cc',
            'vhost',
            'method',
            'agent',
            'request_id',
        ]
        list_x_param = [
            'mz',
            'url',
            'args',
            'body',
            'cookie',
            'other_headers',
            'referer',
            'ua',
            'search',
        ]
        list_param = list_s_param + list_x_param
        all_param = mgmt_param + int_param + bool_param + list_param

        # remove unknown parameter
        for k in data.copy().keys():
            if k not in all_param:
                del data[k]

        # parameters processing
        for k in all_param:
            try:

                # skip non-exists key
                if k not in data:
                    continue

                # delete empty
                if (k in data) and (data.get(k) is None):
                    del data[k]
                    continue

                # int validation
                if k in mgmt_param + int_param:
                    if (isinstance(data[k], bool)) or (not isinstance(data[k], int)):
                        response = {'status': 'fail', 'description': 'Incorrect value for {}'.format(k)}
                        return response
                    if k in mgmt_param:
                        continue

                # bool validation
                elif k in bool_param:
                    if not isinstance(data[k], bool):
                        response = {'status': 'fail', 'description': 'Incorrect value for {}'.format(k)}
                        return response

                # list validation and normalization
                elif k in list_param:

                    # remove empty
                    if not data[k]:
                        del data[k]
                        continue

                    # type validation
                    if not isinstance(data[k], list):
                        response = {'status': 'fail', 'description': 'Incorrect value for {}'.format(k)}
                        return response

                    # mz validation
                    if (k == 'mz') and ([x for x in data[k] if (not isinstance(x, str)) or (not re.match(r'[\w-]', x))]):
                        response = {'status': 'fail', 'description': 'Value of "{}" contains incorrect symbol'.format(k)}
                        return response

                    # bt/rule_id
                    if k == 'bt':

                        # value validation
                        if [x for x in data[k] if (not isinstance(x, int)) or (isinstance(x, bool))]:
                            response = {'status': 'fail', 'description': 'Value of "{}" contains incorrect data'.format(k)}
                            return response

                        # value processing
                        q = 'AND {} in %s'.format(k)
                        qhdrs.append(q)
                        qargs.append(tuple(data[k]))
                        continue

                    # other keys
                    else:

                        # convert value to lowercase
                        data[k] = [str(x).lower() if x or x == 0 else x for x in data[k]]

                        # CC normalization
                        if k == 'cc':
                            data[k] = ['' if x is None else str(x) for x in data[k]]

                        # remove duplicate
                        data[k] = list(set(data[k]))

                        # mz normalization
                        if k == 'mz':

                            # zone normalization
                            for idx, item in enumerate(data[k].copy()):

                                # skip none/empty
                                if item in [None, '']:
                                    continue

                                # URL normalization
                                if item == 'url':
                                    data[k][idx] = 'uri'
                                    continue

                                # remove unknown zone
                                if item and (item not in ['uri', 'args', 'body', 'cookie', 'other_headers', 'referer', 'ua']):
                                    del data[k][idx]

                            # remove empty
                            if not data[k]:
                                del data[k]
                                continue

                        # final value normalization
                        if k in list_s_param:
                            data[k] = '|'.join(
                                ''.join('\\x' + hexlify(c.encode('UTF-8')).decode('UTF-8') for c in [x for x in i]) for i in data[k]
                            )
                        else:
                            data[k] = '|'.join(
                                '%' + ''.join('\\x' + hexlify(c.encode('UTF-8')).decode('UTF-8') for c in [x for x in i]) + '%' for i in data[k]
                            )

                        # search processing
                        if k == 'search':

                            # init the query
                            q = 'AND (' \
                                + 'lower(uri) SIMILAR TO %s' \
                                + ' OR ' \
                                + 'lower(args) SIMILAR TO %s' \
                                + ' OR ' \
                                + 'lower(body) SIMILAR TO %s' \
                                + ' OR ' \
                                + 'lower(cookie) SIMILAR TO %s' \
                                + ' OR ' \
                                + 'lower(other_headers) SIMILAR TO %s' \
                                + ' OR ' \
                                + 'lower(referer) SIMILAR TO %s' \
                                + ' OR ' \
                                + 'lower(ua) SIMILAR TO %s' \
                                + ')'

                            # update the query
                            qhdrs.append(q)

                            # update query args
                            for _ in range(0, 7):
                                qargs.append(data[k])

                            # finish parameter processing
                            continue

                # update the query
                if k in ['reverseorder']:
                    continue
                elif k == 'starttime':
                    q = 'AND timestamp::integer >= %s'
                    qhdrs.append(q)
                elif k == 'endtime':
                    q = 'AND timestamp::integer <= %s'
                    qhdrs.append(q)
                elif k in int_param:
                    q = 'AND {} = %s'.format(k)
                    qhdrs.append(q)
                elif k in list_param:
                    if k == 'url':
                        q = 'AND lower(uri) SIMILAR TO %s'
                    else:
                        q = 'AND lower({}) SIMILAR TO %s'.format(k)
                    qhdrs.append(q)
                else:
                    log.error('An error occurred while extracting data from {} to {}: unknown parameter ({})'.format(agent_ip, url, k))
                    continue

                # update query args
                qargs.append(data[k])

            except Exception as e:
                log.error('An error occurred while extracting data from {} to {}: {}'.format(agent_ip, url, e))
                response = {'status': 'fail', 'description': str(e)}
                return response

        # body processing
        k = 'bodylimit'
        if data.get(k):
            qbody = 'SUBSTRING(body FOR {})'.format(data[k])
        else:
            qbody = 'body'

        # extract the limit
        limit = data.get('limit')
        limit = limit if isinstance(limit, int) and (not isinstance(limit, bool)) and (limit > 0) else 0

        # offset wo limit
        if ('offset' in data) and (limit < 1):
            response = {'status': 'fail', 'description': 'Using parameter "offset" with parameter "limit" < 1 is not allowed'}
            return response

        # extract the offset
        offset = data.get('offset')
        offset = offset * limit if (limit and offset) else 0

        # order processing
        order = 'DESC' if data.get('reverseorder') else 'ASC'

        ##
        # Search data in the database
        ##

        # convert query filter headers/values
        qhdrs = ' '.join(qhdrs)
        qargs = tuple(qargs)

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # total count
        query = """ SELECT COUNT(*) FROM mt_main WHERE {} """.format(qhdrs)
        cursor.execute(query, qargs)
        response['totalCount'] = cursor.fetchone()[0]

        # search the data

        query = """
            SELECT
                id, timestamp, request_id, agent, mz, bt, ip, cc,
                vhost, method, uri, args, {}, cookie, other_headers, ua, referer, mz_origin
            FROM
                mt_main
            WHERE {}
            ORDER BY id {}
        """.format(qbody, qhdrs, order)

        if limit:
            query += ' LIMIT %s'
            qargs += (limit,)

        if offset:
            query += ' OFFSET %s'
            qargs += (offset,)

        cursor.execute(query, qargs)
        res = cursor.fetchall()

        # del vars
        del qhdrs, qargs, data

        # processing result
        if res:

            # update the response data
            response['count'] = len(res)

            # processing the result
            for item in res:

                # init
                mz = item[4]

                # mz normalization:
                mz = mz.lower().replace('uri', 'url') if isinstance(mz, str) else mz

                # item processing
                r = {
                    'id': item[0],
                    'timestamp': item[1],
                    'request_id': item[2],
                    'agent': item[3],
                    'mz': mz,
                    'bt': item[5],
                    'ip': item[6],
                    'cc': item[7],
                    'vhost': item[8],
                    'method': item[9],
                    'url': item[10],
                    'args': item[11],
                    'body': item[12],
                    'cookie': item[13],
                    'other_headers': item[14],
                    'ua': item[15],
                    'referer': item[16],
                    'mz_origin': item[17]
                }

                # CC processing
                if not r['cc'].strip():
                    r['cc'] = None

                # update the response data
                response['attacks'].append(r)

                # del vars
                del item, r

        # del vars
        del res

        # close database connection
        close_db_connection(connect, cursor)

        # response
        return response

    except Exception as e:
        log.error('An error occurred in get_mt_attack: {}'.format(e))


def set_mt_attack(url, request, data, agent_ip, waf_id):
    try:

        # extract user-agent
        try:
            agent = request.headers.get('user-agent', '')
        except Exception as e:
            log.error('An error occurred while extracting User-Agent from request to {} from client {}: {}'.format(url, agent_ip, e))
            agent = ''

        # processing for each item
        for i in range(len(data)):

            # set WAF ID
            data[i]['waf_id'] = waf_id

            # timestamp processing and data normalization
            if len(str(data[i]['timestamp'])) > 10:
                data[i]['timestamp'] = str(data[i]['timestamp'])[:10]
            try:
                datetime.fromtimestamp(int(data[i]['timestamp'])).strftime('%H:%M:%S %d.%m.%Y')
            except Exception as e:
                log.error('An error occurred in {} while converting date {}: {}'.format(url, data[i]['timestamp'], e))
                continue

            # add agent field
            if 'MLA' in agent:
                data[i]['agent'] = 'MLA'
            elif 'MLC' in agent:
                data[i]['agent'] = 'MLC'
            else:
                data[i]['agent'] = '-'

            # data normalization
            for z in ['uri', 'args', 'cookie', 'referer', 'other_headers', 'body', 'ua']:
                if isinstance(data[i][z], str):
                    data[i][z] = data[i][z].replace('\x00', '\\x00')

            ##
            # Zone normalization
            ##

            # convert to lowercase
            data[i]['mz'] = str(data[i]['mz']).lower()

            # init
            r = []

            # rename zones
            for mz in data[i]['mz'].split('|'):

                # rename the zone
                if mz == 'url':
                    mz = 'uri'
                elif mz == 'user-agent':
                    mz = 'ua'

                # update the list
                r.append(mz)

            # update zones list
            data[i]['mz'] = '|'.join(r)
            del r

            # CC processing
            data[i]['cc'] = data[i].get('cc', None)

            ##
            # Save data to the database
            ##

            # set the fields
            fields_list = (
                'bt', 'timestamp', 'mz', 'ip', 'vhost', 'method',
                'uri', 'args', 'cookie', 'referer', 'other_headers', 'body',
                'mz_origin', 'ua', 'request_id', 'waf_id', 'agent', 'cc'
            )

            query = """
                INSERT INTO mt_main (
                    {}
                ) VALUES (
                    %s, %s, %s, %s, %s, %s,
                    %s, %s, %s, %s, %s, %s,
                    %s, %s, %s, %s, %s, %s
                )
            """.format(','.join(fields_list))

            dbr = (data[i].get(field) for field in fields_list)
            connect = init_db_connection()
            cursor = connect.cursor()
            cursor.execute(query, tuple(dbr))
            close_db_connection(connect, cursor)

        # response
        return True

    except Exception as e:
        log.error('An error occurred in set_mt_attack: {}'.format(e))


def del_mt_attack(data, waf_id):
    try:

        # extract the id
        ids = data.get('id')

        # skip empty
        if not ids:
            return {'status': 'fail', 'description': 'Empty or missing column: id'}

        # id processing
        if (not isinstance(ids, list)) or ([x for x in ids if (not isinstance(x, int)) or (isinstance(x, bool))]):
            response = {'status': 'fail', 'description': 'Value of "id" contains incorrect data or is missing'}
            return response

        # convert to tuple
        ids = tuple(ids)

        ##
        # Processing JSON
        ##

        # init
        response = {'status': 'success', 'deleted': 0}

        # insert the data
        query = """ DELETE FROM mt_main WHERE (id IN %s) AND (waf_id = %s) RETURNING id """
        connect = init_db_connection()
        cursor = connect.cursor()
        cursor.execute(query, (ids, waf_id,))
        res = cursor.fetchall()
        close_db_connection(connect, cursor)

        # processing result
        response['deleted'] = len(res)

        # response
        return response

    except Exception as e:
        log.error('An error occurred in del_mt_attack: {}'.format(e))


def get_mt_fp(url, data, agent_ip, waf_id):
    try:

        ##
        # Receive data from the database
        ##

        # init
        response = {'status': 'success', 'count': 0, 'totalCount': 0, 'waf_id': waf_id, 'fp': []}

        # init query filter headers/values
        qhdrs = ['waf_id = %s']
        qargs = [waf_id]

        # init parameters type
        mgmt_param = [
            'limit',
            'offset',
        ]
        bool_param = [
            'reverseorder',
        ]
        list_i_param = [
            'id',
        ]
        list_s_param = [
            'vhost',
        ]
        list_param = list_i_param + list_s_param
        all_param = mgmt_param + bool_param + list_param

        # remove unknown parameter
        for k in data.copy().keys():
            if k not in all_param:
                del data[k]

        # parameters processing
        for k in all_param:
            try:

                # skip non-exists key
                if k not in data:
                    continue

                # delete empty
                if data.get(k) is None:
                    del data[k]
                    continue

                # int validation
                if k in mgmt_param:
                    if (isinstance(data[k], bool)) or (not isinstance(data[k], int)):
                        response = {'status': 'fail', 'description': 'Incorrect value for {}'.format(k)}
                        return response
                    if k in mgmt_param:
                        continue

                # bool validation
                elif k in bool_param:
                    if not isinstance(data[k], bool):
                        response = {'status': 'fail', 'description': 'Incorrect value for {}'.format(k)}
                        return response

                # list validation and normalization
                if k in list_param:

                    # remove empty
                    if not data[k]:
                        del data[k]
                        continue

                    # type validation
                    if not isinstance(data[k], list):
                        response = {'status': 'fail', 'description': 'Incorrect value for {}'.format(k)}
                        return response

                    # id validation
                    if (k == 'id') and ([x for x in data[k] if (isinstance(x, bool)) or (not isinstance(x, int))]):
                        response = {'status': 'fail', 'description': 'Value of "{}" contains incorrect data'.format(k)}
                        return response

                    # vhost validation
                    if (k == 'vhost') and ([x for x in data[k] if (not isinstance(x, str)) or (not re.match(r'^[a-zA-Z0-9.-]+$', x))]):
                        response = {'status': 'fail', 'description': 'Value of "{}" contains incorrect symbol'.format(k)}
                        return response

                    # convert value to lowercase
                    if k not in list_i_param:
                        data[k] = [str(x).lower() if x or x == 0 else x for x in data[k]]

                    # remove duplicate
                    data[k] = list(set(data[k]))

                    # convert to str
                    if k in list_s_param:
                        data[k] = '|'.join(data[k])

                # update the query
                if k in ['reverseorder']:
                    continue
                elif (k == 'id') and k in data:
                    q = 'AND {} in %s'.format(k)
                    qhdrs.append(q)
                    qargs.append(tuple(data[k]))
                    continue
                elif (k == 'vhost') and data.get(k):
                    q = 'AND lower({}) SIMILAR TO %s'.format(k)
                    qhdrs.append(q)
                    qargs.append(data[k])
                    continue
                else:
                    log.error('An error occurred while extracting data from {} to {}: unknown parameter ({})'.format(agent_ip, url, k))
                    continue

            except Exception as e:
                log.error('An error occurred while extracting data from {} to {}: {}'.format(agent_ip, url, e))
                response = {'status': 'fail', 'description': str(e)}
                return response

        # extract the limit
        limit = data.get('limit')
        limit = limit if isinstance(limit, int) and (not isinstance(limit, bool)) and (limit > 0) else 0

        # offset wo limit
        if ('offset' in data) and (limit < 1):
            response = {'status': 'fail', 'description': 'Using parameter "offset" with parameter "limit" < 1 is not allowed'}
            return response

        # extract the offset
        offset = data.get('offset')
        offset = offset * limit if (limit and offset) else 0

        # order processing
        order = 'DESC' if data.get('reverseorder') else 'ASC'

        ##
        # Search data in the database
        ##

        # convert query filter headers/values
        qhdrs = ' '.join(qhdrs)
        qargs = tuple(qargs)

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # total count
        query = """ SELECT COUNT(*) FROM mt_export WHERE {} """.format(qhdrs)
        cursor.execute(query, qargs)
        response['totalCount'] = cursor.fetchone()[0]

        # search the data

        query = """
            SELECT
                id, timestamp, lower(mz), vhost, uri, args, body, cookie, other_headers, ua, referer
            FROM
                mt_export
            WHERE {}
            ORDER BY id {}
        """.format(qhdrs, order)

        if limit:
            query += ' LIMIT %s'
            qargs += (limit,)

        if offset:
            query += ' OFFSET %s'
            qargs += (offset,)

        cursor.execute(query, qargs)
        res = cursor.fetchall()

        # del vars
        del qhdrs, qargs, data

        # processing result
        if res:

            # update the response data
            response['count'] = len(res)

            # processing the result
            for item in res:

                # init
                mz = item[2]

                # mz normalization:
                mz = mz.replace('uri', 'url') if isinstance(mz, str) else mz

                # item processing
                r = {
                    'id': item[0],
                    'timestamp': item[1],
                    'mz': mz,
                    'vhost': item[3],
                    'url': item[4],
                    'args': item[5],
                    'body': item[6],
                    'cookie': item[7],
                    'other_headers': item[8],
                    'ua': item[9],
                    'referer': item[10]
                }

                # update the response data
                response['fp'].append(r)

                # del vars
                del item, r

        # del vars
        del res

        # close database connection
        close_db_connection(connect, cursor)

        # response
        return response

    except Exception as e:
        log.error('An error occurred in get_mt_fp: {}'.format(e))


def set_mt_fp(url, data, agent_ip, waf_id):
    try:

        ##
        # Receive data from the database
        ##

        # remove the key
        del data['key']

        # init parameters type
        str_param = [
            'vhost',
            'mz',
            'mz_data',
        ]
        all_param = str_param

        # remove unknown parameter
        for k in data.copy().keys():
            if k not in all_param:
                del data[k]

        # parameters processing
        for k in all_param:
            try:

                # skip non-exists key
                if k not in data:
                    continue

                # string validation
                if (not isinstance(data.get(k), str)) or (data.get(k) == ''):
                    response = {'status': 'fail', 'description': 'Value of "{}" contains incorrect data or is missing'.format(k)}
                    return response

                # convert value to lowercase
                data[k] = data[k].lower()

                # mz validation
                if (k == 'mz') and (data[k].lower() not in ['url', 'args', 'body', 'cookie', 'other_headers', 'referer', 'ua']):
                    response = {'status': 'fail', 'description': 'Value of "{}" contains incorrect data'.format(k)}
                    return response

            except Exception as e:
                log.error('An error occurred while extracting data from {} to {}: {}'.format(agent_ip, url, e))
                response = {'status': 'fail', 'description': str(e)}
                return response

        ##
        # Search data in the database
        ##

        # init
        dbr = {
            'waf_id': waf_id,
            'bt': 12,
            'timestamp': int(time()),
            'request_id': '',
            'mz': data['mz'],
            'vhost': data['vhost'],
            'method': 'POST',
            'uri': '',
            'args': '',
            'cookie': '',
            'ua': '',
            'referer': '',
            'body': '',
            'other_headers': ''
        }

        # update the data
        if data['mz'] == 'url':
            dbr['uri'] = data['mz_data']
        else:
            dbr[data['mz']] = data['mz_data']

        # insert the data
        query = """ INSERT INTO mt_export ({}) VALUES %s RETURNING id """.format(','.join(dbr.keys()))
        connect = init_db_connection()
        cursor = connect.cursor()
        cursor.execute(query, (tuple(dbr.values()),))
        res = cursor.fetchone()
        close_db_connection(connect, cursor)

        # del vars
        del data, dbr

        # set response
        if res:
            response = {'status': 'success', 'id': res[0]}
        else:
            response = {'status': 'fail', 'description': 'ID was not returned during the data saving process'}

        # response
        return response

    except Exception as e:
        log.error('An error occurred in get_mt_fp: {}'.format(e))


def del_mt_fp(data, waf_id):
    try:

        # extract the id
        ids = data.get('id')

        # skip empty
        if not ids:
            return {'status': 'fail', 'description': 'Empty or missing column: id'}

        # id processing
        if (not isinstance(ids, list)) or ([x for x in ids if (not isinstance(x, int)) or (isinstance(x, bool))]):
            response = {'status': 'fail', 'description': 'Value of "id" contains incorrect data or is missing'}
            return response

        # convert to tuple
        ids = tuple(ids)

        ##
        # Processing JSON
        ##

        # init
        response = {'status': 'success', 'deleted': 0}

        # insert the data
        query = """ DELETE FROM mt_export WHERE (id IN %s) AND (waf_id = %s) RETURNING id """
        connect = init_db_connection()
        cursor = connect.cursor()
        cursor.execute(query, (ids, waf_id,))
        res = cursor.fetchall()
        close_db_connection(connect, cursor)

        # processing result
        response['deleted'] = len(res)

        # response
        return response

    except Exception as e:
        log.error('An error occurred in del_mt_fp: {}'.format(e))
