#!/usr/bin/env python3

"""
Nemesida WAF API
Copyright (c) 2017-2025 Pentestit LLC - All Rights Reserved
Author: Romanov R.
"""

import base64
import hashlib
import json
import OpenSSL.crypto
import re
import validators

from ast import literal_eval
from cryptography.hazmat.primitives.serialization import load_pem_parameters
from netaddr import IPAddress, IPNetwork, IPRange

from urls import get_nginx_certificate_uri
from urls import get_nginx_dhparam_uri
from urls import get_nginx_global_uri

from core import init_db_connection, close_db_connection
from core import is_base64
from core import log
from core import memc
from core import memc_lock

# Command
nginx_cmd = ['set', 'del']

##
# Columns
##

nginx_global_cols = [
    'active', 'resolver', 'resolver_timeout', 'set_real_ip_from', 'real_ip_header',
    'client_max_body_size', 'limit_req_zone',
]

nginx_vhost_cols = [
    'active', 'name', 'listen', 'server_name', 'certificate_name', 'dhparam_name', 'ssl_session_timeout',
    'ssl_ciphers', 'ssl_protocols', 'root', 'index', 'add_header', 'error_page', 'locations',
]

nginx_upstream_cols = [
    'name', 'balancing_method', 'servers',
]

nginx_certificate_cols = [
    'name', 'certificate', 'certificate_key',
]

nginx_dhparam_cols = [
    'name', 'dhparam',
]

##
# Keys and types
##

# name
nginx_name = {
    'certificate_name': 'name', 'dhparam_name': 'name',
}

# non-empty name
nginx_ne_name = {
    'name': 'name',
}

# global
nginx_global = {
    'resolver': 'str', 'resolver_timeout': 'str', 'set_real_ip_from': 'list',
    'real_ip_header': 'str', 'client_max_body_size': 'str',
    'limit_req_zone': 'list',
}

# vhost
nginx_vhost = {
    'active': 'bool', 'listen': 'list', 'server_name': 'str',
    'certificate_name': 'str', 'dhparam_name': 'str', 'ssl_session_timeout': 'str', 'ssl_ciphers': 'str',
    'ssl_protocols': 'str', 'root': 'str', 'index': 'str', 'add_header': 'dict', 'error_page': 'dict',
    'locations': 'lod',
}

# locations
nginx_vhost_locations = {
    'location': 'str', 'pass_type': 'str', 'add_header': 'dict', 'error_page': 'dict',
    'limit_req': 'str', 'limit_req_status': 'int',
}

# static_pass
nginx_vhost_locations_static_pass = {
    'path': 'str', 'index': 'str', 'autoindex': 'str',
}

# proxy_pass
nginx_vhost_locations_proxy_pass = {
    'proxy_pass': 'str', 'proxy_http_version': 'float', 'proxy_set_header': 'dict', 'proxy_connect_timeout': 'str',
    'proxy_send_timeout': 'str', 'proxy_read_timeout': 'str', 'proxy_intercept_errors': 'str', 'nwaf_ws_proxy_pass': 'str'
}

# memcached_pass
nginx_vhost_locations_memcached_pass = {
    'set': 'str', 'memcached_pass': 'str',
}

# upstream
nginx_upstream = {
    'balancing_method': 'str', 'servers': 'lod',
}

# upstream
nginx_upstream_servers = {
    'server': 'str', 'fail_timeout': 'str',
    'weight': 'int', 'max_fails': 'int',
    'backup': 'bool', 'down': 'bool',
}

# certificate
nginx_certificate = {
    'certificate': 'b64', 'certificate_key': 'b64',
}

# dhparam
nginx_dhparam = {
    'dhparam': 'b64',
}

##
# All entities
##

nginx_all_items = {}
nginx_all_items.update(nginx_name)
nginx_all_items.update(nginx_ne_name)
nginx_all_items.update(nginx_global)
nginx_all_items.update(nginx_vhost)
nginx_all_items.update(nginx_vhost_locations)
nginx_all_items.update(nginx_vhost_locations_static_pass)
nginx_all_items.update(nginx_vhost_locations_proxy_pass)
nginx_all_items.update(nginx_vhost_locations_memcached_pass)
nginx_all_items.update(nginx_upstream)
nginx_all_items.update(nginx_upstream_servers)
nginx_all_items.update(nginx_certificate)
nginx_all_items.update(nginx_dhparam)

##
# List of types
##

nginx_type_name = [k for k, v in nginx_all_items.items() if v == 'name']
nginx_type_str = [k for k, v in nginx_all_items.items() if v == 'str']
nginx_type_b64 = [k for k, v in nginx_all_items.items() if v == 'b64']
nginx_type_int = [k for k, v in nginx_all_items.items() if v == 'int']
nginx_type_float = [k for k, v in nginx_all_items.items() if v == 'float']
nginx_type_bool = [k for k, v in nginx_all_items.items() if v == 'bool']
nginx_type_list = [k for k, v in nginx_all_items.items() if v == 'list']
nginx_type_dict = [k for k, v in nginx_all_items.items() if v == 'dict']
nginx_type_lod = [k for k, v in nginx_all_items.items() if v == 'lod']


def ip_validation(data):
    try:

        # processing IP range
        if '-' in data:
            IPRange(data.split('-')[0], data.split('-')[1])
        # processing IPv4/IPv6/CIDR
        else:
            IPNetwork(data)

        return True

    except Exception as e:
        log.error('An error occurred while IP validation for "{}": {}'.format(data, e))
        return False


def address_processing(data):
    try:

        # ipv6
        if '[' in data:
            addr_list = [x for x in re.split(r'(\[.+])', data) if x]
            if len(addr_list) > 1:
                addr_list[1] = addr_list[1].lstrip(':')
        # ipv4
        else:
            addr_list = data.split(':')

        # host and port processing
        host = addr_list[0]
        if len(addr_list) > 1:
            port = int(addr_list[1])
        else:
            port = ''

        # result
        return host, port

    except Exception:
        return False, False


def ip_domain_validation(data):

    # IPv6 preprocessing
    data = data.lstrip('[').rstrip(']')

    # IP validation
    try:
        if IPAddress(data):
            return True
    except Exception:
        pass

    # Domain validation if item is not an IP
    try:

        # skip localhost
        if data == 'localhost':
            return True

        # domain validation
        if validators.domain(data):
            return True

    except Exception:
        pass

    # if not validated
    return False


def get_nginx_hash(table, waf_id):

    # init memcached
    memc_key = 'hash_' + table + '_' + waf_id

    try:

        # get data from the memcached
        with memc_lock:
            try:
                memc_bin = memc.get(memc_key)
            except Exception as e:
                memc_bin = None
                log.error('An error occurred while receiving data from the memcached: {}'.format(e))

        # data in the memcached
        if memc_bin:
            return memc_bin.decode('UTF-8')

        # data not in the memcached
        else:

            # set data into the memcached
            set_nginx_hash(table, waf_id)

            # get data from the memcached
            with memc_lock:
                try:
                    memc_bin = memc.get(memc_key)
                except Exception as e:
                    memc_bin = None
                    log.error('An error occurred while receiving data from the memcached: {}'.format(e))

            # return the result
            return memc_bin.decode('UTF-8') if memc_bin else None

    except Exception as e:
        log.error('An error occurred while extracting the hash for "{}" table: {}'.format(table, e))
        return False


def upstream_search(err_prefix, waf_id, upstream):
    try:

        ##
        # Search the upstream in the database
        ##

        # search upstream name
        query = """ SELECT COUNT(*) FROM nginx_upstream WHERE (waf_id = %s AND name = %s) """
        connect = init_db_connection()
        cursor = connect.cursor()
        cursor.execute(query, (waf_id, upstream,))
        res = cursor.fetchone()
        connect.close()

        # processing the result
        if not res or res[0] == 0:
            return False, '{}: upstream "{}" is not found'.format(err_prefix, upstream)
        else:
            return True, True

    except Exception as e:
        err_msg = 'An error occurred while searching upstream {}: {}'.format(upstream, e)
        log.info(err_msg)
        return False, err_msg


def set_nginx_hash(table, waf_id):

    # init
    memc_key = 'hash_' + table + '_' + waf_id

    try:

        ##
        # Extract data from the database
        ##

        # send a query
        query = """ SELECT * FROM {} WHERE waf_id = %s """.format(table)
        connect = init_db_connection()
        cursor = connect.cursor()
        cursor.execute(query, (waf_id,))
        res = cursor.fetchall()
        connect.close()

        # Processing and hashing data
        if res:

            # get data hash
            hdata = hashlib.sha256(str(res).encode('UTF-8')).hexdigest()

            # set into the memcached
            with memc_lock:
                try:
                    memc.set(memc_key, hdata)
                except Exception as e:
                    log.error('An error occurred while placing data into the memcached: {}'.format(e))

        else:

            # delete data from the memcached
            with memc_lock:
                try:
                    memc.delete(memc_key)
                except Exception as e:
                    log.error('An error occurred while deleting data from the memcached: {}'.format(e))

    except Exception as e:
        err_msg = 'An error occurred while updating the hash from data in table "{}": {}'.format(table, e)
        log.error(err_msg)


def nginx_format_processing(nginx_type, data):

    # init
    res = ''
    indent_1 = ' ' * 2
    indent_2 = indent_1 * 2

    try:

        ##
        # Processing data
        ##

        # global
        if nginx_type == 'global':

            # processing the dictionary
            for k, v in data.items():

                # value as a list
                if k in ['set_real_ip_from', 'limit_req_zone']:
                    for i in literal_eval(v):
                        # if not empty
                        if i:
                            res = res + str(k) + ' ' + str(i) + ';\n'
                    # skip updating the format below
                    continue

                # update the format
                res = res + str(k) + ' ' + str(v) + ';\n'

            # response
            return res

        # vhost
        elif nginx_type == 'vhost':

            # set header
            res = 'server {\n'

            # move listen to the start
            if 'listen' in data:

                # init
                dl = []
                is_http2 = False
                data['listen'] = literal_eval(data['listen'])

                # http2 new format fix
                for idx, item in enumerate(data['listen'].copy()):

                    # data normalization
                    item = [x.strip() for x in item.split(' ') if x]

                    # http2 processing
                    if 'http2' in item:
                        item.remove('http2')
                        is_http2 = True

                    # update the data
                    item = ' '.join(item)
                    dl.append(item)

                # update listen data
                data['listen'] = str(dl) if dl else str(data['listen'])

                # move listen to the start
                d = {'listen': data['listen']}
                d = {**d, **{'http2': 'on'}} if is_http2 else d
                del data['listen']
                data = {**d.copy(), **data.copy()}
                del dl, d

            # move locations to the end
            if 'locations' in data:
                locations = data['locations']
                del data['locations']
                data['locations'] = locations
                del locations

            # processing the dictionary
            for k, v in data.items():

                ##
                # Update value
                ##

                if k in ['listen']:

                    # convert to the list
                    v = literal_eval(v)

                    # processing the list
                    for item in v:
                        res = res + indent_1 + str(k) + ' ' + str(item) + ';\n'

                    # skip updating the format below
                    continue

                elif k in ['certificate_name']:

                    ##
                    # Convert certificate_name to certificate and certificate_key
                    ##

                    k1 = 'ssl_certificate'
                    v1 = 'nwaf/conf/nginx/ssl/certs/' + str(v) + '.crt'
                    res = res + indent_1 + str(k1) + ' ' + str(v1) + ';\n'

                    k1 = 'ssl_certificate_key'
                    v1 = 'nwaf/conf/nginx/ssl/private/' + str(v) + '.key'
                    res = res + indent_1 + str(k1) + ' ' + str(v1) + ';\n'

                    # skip updating the format below
                    continue

                elif k in ['dhparam_name']:

                    # convert dhparam_name to ssl_dhparam
                    k = 'ssl_dhparam'
                    v = 'nwaf/conf/nginx/ssl/dhparam/' + str(v) + '.pem'
                    res = res + indent_1 + str(k) + ' ' + str(v) + ';\n'

                    # skip updating the format below
                    continue

                elif k in ['add_header', 'error_page']:

                    # convert to the dictionary
                    v = literal_eval(v)

                    # add_header and error_page
                    for k1, v1 in v.items():
                        res = res + indent_1 + str(k) + ' ' + str(k1) + ' ' + str(v1) + ';\n'

                    # skip updating the format below
                    continue

                elif k in ['locations']:

                    # processing each dictionary in locations list
                    for locations in literal_eval(data['locations']):

                        # extract location and pass_type
                        res = res + indent_1 + 'location ' + str(locations['location']) + ' {\n'
                        del locations['location']

                        # processing locations items
                        for k1, v1 in locations.items():

                            # proxy_set_header/add_header/error_page
                            if k1 in ['add_header', 'error_page', 'proxy_set_header']:

                                # processing the dictionary
                                for k2, v2 in v1.items():
                                    res = res + indent_2 + str(k1) + ' ' + str(k2) + ' ' + str(v2) + ';\n'

                            # root/alias path processing
                            elif k1 in ['pass_type']:
                                if ('path' in locations) and (locations['path']):
                                    res = res + indent_2 + str(v1) + ' ' + str(locations['path']) + ';\n'

                            # skip path (it processing with root/alias)
                            elif k1 in ['path']:
                                continue

                            # other items processing
                            else:
                                res = res + indent_2 + str(k1) + ' ' + str(v1) + ';\n'

                        # update the format
                        res = res + indent_1 + '}\n'

                    # skip updating the format below
                    continue

                # update the format
                res = res + indent_1 + str(k) + ' ' + str(v) + ';\n'

            # set footer
            res = res + '}\n'

            # response
            return res

        # upstream
        elif nginx_type == 'upstream':

            # set header
            res = 'upstream ' + str(data['name']) + ' {\n'
            del data['name']

            # set balancing method
            if 'balancing_method' in data and data['balancing_method']:
                res = res + indent_1 + str(data['balancing_method']) + ';\n'
                del data['balancing_method']

            # processing the dictionary
            for item in literal_eval(data['servers']):

                # extract server
                server = str(item['server'])
                del item['server']

                # set sub-header
                server_params = indent_1 + 'server' + ' ' + server

                # processing server's params
                for k, v in item.items():

                    # bool parameters: backup/down/resolve etc.
                    if isinstance(v, bool):
                        if v:
                            server_params = server_params + ' ' + str(k)
                    # other parameters (weight etc.)
                    else:
                        server_params = server_params + ' ' + str(k) + '=' + str(v)

                # update the format
                res = res + server_params + ';\n'

            # set footer
            res = res + '}\n'

            # response
            return res

    except Exception as e:
        log.error('An error occurred while processing data in Nginx format: {}'.format(e))

    return res


def nginx_dhparam_validation(name, dhparam):
    try:

        # b64decode
        dhparam = base64.b64decode(dhparam)

        # file validation
        if load_pem_parameters(dhparam):
            return True, ''

    except Exception as e:
        err_msg = 'An error occurred while validation the DH parameters {}: {}'.format(name, e)
        return False, err_msg


def nginx_certificate_validation(name, crt, key):
    try:

        # b64decode
        crt = base64.b64decode(crt)
        key = base64.b64decode(key)

        # Convert key and cert to obj
        crt = OpenSSL.crypto.load_certificate(OpenSSL.crypto.FILETYPE_PEM, crt)
        key = OpenSSL.crypto.load_privatekey(OpenSSL.crypto.FILETYPE_PEM, key)

        ct = OpenSSL.SSL.Context(OpenSSL.SSL.TLSv1_METHOD)
        ct.use_certificate(crt)
        ct.use_privatekey(key)

        # checking the key matches the certificate
        ct.check_privatekey()
        return True, ''

    except Exception as e:
        err_msg = 'An error occurred while validation the certificate {}: {}'.format(name, e)
        return False, err_msg


def nginx_value_extend_validation(waf_id, nginx_type, data, k, v):
    try:

        # init
        err_prefix = 'An error occurred during extended validation ({}/{})'.format(k, v)

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        ##
        # Value validation
        ##

        # global
        if nginx_type == 'global':

            # client_max_body_size 50m
            if k in ['client_max_body_size']:

                # regex validation
                if not re.match(r'^\d+[mg]$', v.lower()):
                    close_db_connection(connect, cursor)
                    return False, '{}: parameter "{}" contains an incorrect value "{}"'.format(err_prefix, k, v)

            # resolver address ... [valid=time] [ipv4=on|off] [ipv6=on|off];
            elif k in ['resolver']:

                # convert to list
                v_list = v.split()

                # allowed parameters
                allowed_opt = ['valid', 'ipv4', 'ipv6']

                ##
                # Parameters validation
                ##

                # processing the parameters
                for param in v_list:

                    # additional parameter validation
                    if '=' in param:

                        # init
                        p = param.split('=')

                        # is allowed validation
                        if p[0] not in allowed_opt:
                            close_db_connection(connect, cursor)
                            return False, '{}: parameter "{}" is not allowed'.format(err_prefix, p[0])

                        # empty value is not allowed
                        if len(p) == 0:
                            close_db_connection(connect, cursor)
                            return False, '{}: empty value'.format(err_prefix)

                        # valid validation
                        if p[0] == 'valid':
                            if not re.match(r'^\d+[smhd]$', p[1].lower()):
                                close_db_connection(connect, cursor)
                                return False, '{}: parameter "{}" contains an incorrect value "{}"'.format(err_prefix, p[0], p[1])

                        # ipv4/ipv6 validation
                        elif p[0] in ['ipv4', 'ipv6']:
                            if p[1] not in ['on', 'off']:
                                close_db_connection(connect, cursor)
                                return False, '{}: parameter "{}" contains an incorrect value "{}"'.format(err_prefix, p[0], p[1])

                    # address validation
                    else:

                        # if not UNIX-domain sockets
                        # (UNIX-domain sockets is allowed and not validated)
                        if not param.startswith('unix:'):

                            # host and port | host in ipv6 and port (optional)
                            if ':' in param:

                                # extract host and port from the address
                                r = address_processing(param)
                                if not r[0]:
                                    close_db_connection(connect, cursor)
                                    return False, '{}: incorrect address "{}"'.format(err_prefix, param)
                                else:
                                    host = r[0]
                                    port = r[1]

                                # port validation
                                if port:
                                    if (not isinstance(port, int)) or (isinstance(port, bool)) or (1 > port > 65535):
                                        close_db_connection(connect, cursor)
                                        return False, '{}: incorrect port "{}"'.format(err_prefix, port)

                            # host or port only
                            else:
                                host = param

                            # host validation
                            if not ip_domain_validation(host):
                                close_db_connection(connect, cursor)
                                return False, '{}: incorrect address "{}"'.format(err_prefix, host)

            # resolver_timeout 30s;
            elif k in ['resolver_timeout']:

                # regex validation
                if not re.match(r'^\d+[smh]$', v.lower()):
                    close_db_connection(connect, cursor)
                    return False, '{}: parameter "{}" contains an incorrect value "{}"'.format(err_prefix, k, v)

            # X-Real-IP|X-Forwarded-For|proxy_protocol
            elif k in ['real_ip_header']:
                if v.lower() not in ['x-real-ip', 'x-forwarded-for', 'proxy_protocol']:
                    close_db_connection(connect, cursor)
                    return False, '{}: parameter "{}" contains an incorrect value "{}"'.format(err_prefix, k, v)

            # ip(6)|ip(6)/cidr
            elif k in ['set_real_ip_from']:
                for i in v:
                    if not ip_validation(i):
                        close_db_connection(connect, cursor)
                        return False, '{}: incorrect IP address'.format(err_prefix)

            # key zone=name:size rate=rate
            # $binary_remote_addr zone=my_zone:10m rate=1r/s
            elif k in ['limit_req_zone']:

                # processing the list
                for item in v:

                    # key validation
                    r = [x for x in item.split() if x.startswith('$')]
                    if len(r) != 1 or r[0] not in ['$binary_remote_addr', '$remote_addr']:
                        close_db_connection(connect, cursor)
                        return False, '{}: incorrect key data'.format(err_prefix)

                    # zone validation
                    sep = 'zone='
                    r = [x for x in item.split() if x.startswith(sep)]
                    if len(r) != 1:
                        close_db_connection(connect, cursor)
                        return False, '{}: incorrect zone data'.format(err_prefix)
                    if not re.match(r'^[\w\-.]+$', r[0].split(':')[0].split(sep)[1]):
                        close_db_connection(connect, cursor)
                        return False, '{}: incorrect zone name'.format(err_prefix)
                    if not re.match(r'^\d+[bkmg]$', r[0].split(':')[1].lower()):
                        close_db_connection(connect, cursor)
                        return False, '{}: incorrect zone size'.format(err_prefix)

                    # rate validation
                    sep = 'rate='
                    r = [x for x in item.split() if x.startswith(sep)]
                    if len(r) != 1 or not re.match(r'^\d+r/[smh]$', r[0].split('=')[1].lower()):
                        close_db_connection(connect, cursor)
                        return False, '{}: incorrect rate data'.format(err_prefix)

                ##
                # Search data in the database
                ##

                # set zone list
                zone_list = []
                db_zone_list = []
                _ = [zone_list.extend(x.split()) for x in v]
                zone_list = set([x.split(':')[0] for x in zone_list if x.startswith('zone=')])

                # search zone name
                query = """ SELECT locations FROM nginx_vhost WHERE waf_id = %s """
                cursor.execute(query, (waf_id,))
                res = cursor.fetchall()

                # processing the result
                for row in res:

                    # skip empty
                    if not row[0]:
                        continue

                    # convert row
                    row = literal_eval(row[0])

                    # search zone
                    for k, v in row[0].items():
                        if k in ['limit_req'] and v:
                            # append zone from the locations to db_zone_list
                            db_zone_list.append([x for x in v.split() if x.startswith('zone=')][0])

                # error if at least one zone from locations is missing from the zone list
                res = [x for x in db_zone_list if x not in zone_list]
                if res:
                    close_db_connection(connect, cursor)
                    return False, '{}: zone ({}) is used'.format(err_prefix, ', '.join(res))

        # upstream
        if nginx_type == 'upstream':

            # incompatible options
            balancing_method_vs_backup = ['hash', 'ip_hash', 'random']

            # hash|ip_hash|least_conn|random
            if k in ['balancing_method']:

                # method validation
                if v not in ['ip_hash', 'least_conn', 'random']:
                    close_db_connection(connect, cursor)
                    return False, '{}: parameter "{}" contains an incorrect value "{}"'.format(err_prefix, k, v)

                # backup cannot be used along with the hash|ip_hash|random
                if v in balancing_method_vs_backup:

                    # search in the request
                    if 'servers' in data:
                        if [x for x in data['servers'] if 'backup' in x and x['backup']]:
                            close_db_connection(connect, cursor)
                            err_msg = '{}: parameter "backup" cannot be used along with "{}" method'.format(err_prefix, v)
                            return False, err_msg

                    # search in the database
                    else:

                        # search balancing_method in the database
                        query = """ SELECT servers FROM nginx_upstream WHERE (waf_id = %s AND name = %s) """
                        cursor.execute(query, (waf_id, data['name'],))
                        res = cursor.fetchone()

                        # if balancing_method not in the result or balancing_method is hash/ip_hash/random
                        if res:
                            res = literal_eval(res[0])
                            if [x for x in res if 'backup' in x and x['backup']]:
                                close_db_connection(connect, cursor)
                                err_msg = '{}: parameter "backup" cannot be used along with "{}" method'.format(err_prefix, v)
                                return False, err_msg

            # server
            elif k in ['server']:

                # if not UNIX-domain sockets
                # (UNIX-domain sockets is allowed and not validated)
                if not v.startswith('unix:'):

                    # host and port | host in ipv6
                    if ':' in v:

                        # extract host and port from the address
                        r = address_processing(v)
                        if not r[0]:
                            close_db_connection(connect, cursor)
                            return False, '{}: incorrect address "{}"'.format(err_prefix, v)
                        else:
                            host = r[0]
                            port = r[1]

                        # port validation
                        if port:
                            if (not isinstance(port, int)) or (isinstance(port, bool)) or (1 > port > 65535):
                                close_db_connection(connect, cursor)
                                return False, '{}: incorrect port "{}"'.format(err_prefix, port)

                    # host only
                    else:
                        host = v

                    # host validation
                    if host != '[::]':
                        if not ip_domain_validation(host):
                            close_db_connection(connect, cursor)
                            return False, '{}: incorrect server address "{}"'.format(err_prefix, host)

            # int
            elif k in ['weight', 'max_fails']:

                # data normalization
                try:
                    v = literal_eval(v)
                except Exception:
                    pass

                # int validation
                if (not isinstance(v, int)) or (isinstance(v, bool)):
                    close_db_connection(connect, cursor)
                    return False, '{}: parameter "{}" contains an incorrect value "{}"'.format(err_prefix, k, v)

            # timeout
            elif k in ['fail_timeout']:

                # regex validation
                if not re.match(r'^\d+[smh]$', v.lower()):
                    close_db_connection(connect, cursor)
                    return False, '{}: parameter "{}" contains an incorrect value "{}"'.format(err_prefix, k, v)

            # bool
            elif k in ['backup', 'down']:

                # bool validation
                if not isinstance(v, bool):
                    return False, '{}: parameter "{}" contains an incorrect value "{}"'.format(err_prefix, k, v)

                # backup/down cannot be used without at least one active server
                if not [x for x in data['servers'] if 'backup' not in x and 'down' not in x]:
                    close_db_connection(connect, cursor)
                    err_msg = '{}: no servers in upstream'.format(err_prefix)
                    return False, err_msg

                # backup cannot be used along with the hash|ip_hash|random
                if k in ['backup'] and v:

                    # search in the request
                    if 'balancing_method' in data:
                        if data['balancing_method'] in balancing_method_vs_backup:
                            close_db_connection(connect, cursor)
                            err_msg = '{}: parameter "{}" cannot be used along with "{}" method'.format(err_prefix, k, data['balancing_method'])
                            return False, err_msg

                    # search in the database
                    else:

                        # search balancing_method in the database
                        query = """ SELECT balancing_method FROM nginx_upstream WHERE (waf_id = %s AND name = %s) """
                        cursor.execute(query, (waf_id, data['name'],))
                        res = cursor.fetchone()

                        # if balancing_method not in the result or balancing_method is hash/ip_hash/random
                        if res and (res[0] in balancing_method_vs_backup):
                            close_db_connection(connect, cursor)
                            err_msg = '{}: parameter "backup" cannot be used along with "{}" method'.format(err_prefix, res[0])
                            return False, err_msg

        # vhost
        elif nginx_type == 'vhost':

            # listen ["address[:port] [default_server] [ssl] [http2 | spdy] [proxy_protocol]", "..."]
            if k in ['listen']:

                # processing for each item
                # v = ['localhost:443 default_server ssl', '...']
                for item in v:

                    # split to list ()
                    # param_list = ['localhost:443', 'default_server', 'ssl']
                    param_list = item.split()

                    # list validation
                    if len(param_list) < 1:
                        close_db_connection(connect, cursor)
                        return False, '{}: address or/and port is not set'.format(err_prefix)

                    # mutual exclusion validation
                    if 'http2' in v and 'spdy' in param_list:
                        close_db_connection(connect, cursor)
                        return False, '{}: use of mutually exclusive parameters'.format(err_prefix)

                    # processing the list of the parameters
                    for idx, v1 in enumerate(param_list):

                        # Address validation
                        if idx == 0:

                            # init
                            host = ''

                            # if not UNIX-domain sockets
                            # (UNIX-domain sockets is allowed and not validated)
                            if not v1.startswith('unix:'):

                                # host and port | host in ipv6
                                if ':' in v1:

                                    # extract host and port from the address
                                    r = address_processing(v1)
                                    if not r[0]:
                                        close_db_connection(connect, cursor)
                                        return False, '{}: incorrect address "{}"'.format(err_prefix, v1)
                                    else:
                                        host = r[0]
                                        port = r[1]

                                # host or port only
                                else:

                                    # port only
                                    try:
                                        v1 = literal_eval(v1)
                                        if not isinstance(v1, int):
                                            v1 = str(v1)
                                    except Exception:
                                        pass

                                    if isinstance(v1, int) and (not isinstance(v1, bool)):
                                        port = v1
                                    else:
                                        port = host

                                # host validation
                                if host:
                                    if host not in ['*', '[::]', '0.0.0.0']:
                                        if not ip_domain_validation(host):
                                            close_db_connection(connect, cursor)
                                            return False, '{}: incorrect server address "{}"'.format(err_prefix, host)

                                # port validation
                                if port:
                                    if (not isinstance(port, int)) or (isinstance(port, bool)) or (1 > port > 65535):
                                        close_db_connection(connect, cursor)
                                        return False, '{}: incorrect port "{}"'.format(err_prefix, port)

                        # other parameters
                        else:

                            # ssl and certificate application
                            if v1 in ['ssl']:

                                # certificate_name is not set or null
                                if ('certificate_name' not in data) or (data['certificate_name'] is None):
                                    close_db_connection(connect, cursor)
                                    err_msg = 'Validation failed: missing "certificate_name" while "ssl" is used in "listen"'
                                    return False, err_msg

                            # non-support parameters
                            if v1 not in ['default_server', 'ssl', 'http2', 'spdy', 'proxy_protocol']:
                                close_db_connection(connect, cursor)
                                err_msg = 'Validation failed: parameter "{}" is not allowed"'.format(v1)
                                return False, err_msg

            # zone=name [burst=number] [nodelay | delay=number]
            elif k in ['limit_req']:

                # zone validation
                sep = 'zone='
                r = [x for x in v.split() if x.startswith(sep)]
                zone = r[0].split(sep)[1] if r else False
                if len(r) > 1 or not zone:
                    close_db_connection(connect, cursor)
                    return False, '{}: zone is not set'.format(k, v)

                # burst validation
                params = ['burst=', 'delay=']
                for param in params:
                    sep = param
                    r = [x for x in v.split() if x.startswith(sep)]
                    if r:
                        p = int(r[0].split(sep)[1]) if r else False
                        if len(r) > 1 or (not isinstance(p, int)) or (isinstance(p, bool)):
                            close_db_connection(connect, cursor)
                            return False, '{}'.format(err_prefix)

                # nodelay validation
                sep = 'nodelay'
                r = [x for x in v.split() if x.startswith(sep)]
                if r:
                    if len(r) > 1 or r[0] != sep:
                        close_db_connection(connect, cursor)
                        return False, '{}'.format(err_prefix)

                # mutual exclusion validation
                if [x for x in v.split() if x.startswith('nodelay')] and [x for x in v.split() if x.startswith('delay')]:
                    close_db_connection(connect, cursor)
                    return False, '{}: use of mutually exclusive parameters'.format(err_prefix)

                ##
                # Search data in the database
                ##

                # search zone name
                query = """ SELECT limit_req_zone FROM nginx_global WHERE waf_id = %s """
                cursor.execute(query, (waf_id,))
                res = cursor.fetchone()

                # processing the result
                if (not res) or (not res[0]) or (not 'zone=' + zone + ':' in res[0]):
                    close_db_connection(connect, cursor)
                    return False, '{}: limit_req_zone "{}" is not found'.format(err_prefix, zone)

            # TLSv1 TLSv1.1 TLSv1.2
            elif k in ['ssl_protocols']:

                # init
                v_list = v.split()
                allowed = ['tlsv1', 'tlsv1.1', 'tlsv1.2', 'tlsv1.3']

                # ssl protocols validation
                if [x for x in v_list if x.lower() not in allowed]:
                    close_db_connection(connect, cursor)
                    return False, '{}'.format(err_prefix)

            # 1.0|1.1
            elif k in ['proxy_http_version']:
                if v not in [1.0, 1.1]:
                    close_db_connection(connect, cursor)
                    return False, '{}: parameter "{}" contains an incorrect value "{}"'.format(err_prefix, k, v)

            # error_page {code: "[=[response]] uri", ...};
            elif k in ['error_page']:

                # processing the dictionary
                for ki, vi in v.items():

                    ##
                    # Code validation
                    ##

                    # split the key
                    kki = ki.split()

                    # processing the codes
                    for i in kki:

                        # extract code
                        try:
                            i = literal_eval(i)
                        except Exception:
                            close_db_connection(connect, cursor)
                            return False, '{}: incorrect code {}'.format(err_prefix, i)

                        if (not isinstance(i, int)) or (isinstance(i, bool)) or (200 > i > 599):
                            close_db_connection(connect, cursor)
                            return False, '{}: incorrect code {}'.format(err_prefix, i)

                    # split the value
                    vvi = vi.split()

                    # processing the values
                    for i in vvi:

                        # response code validation
                        if i.startswith('='):

                            # response code format validation
                            if i == '=':
                                close_db_connection(connect, cursor)
                                return False, '{}: incorrect response code {}'.format(err_prefix, i)

                            # extract response code
                            try:
                                i = literal_eval(i.split('=')[1])
                            except Exception:
                                close_db_connection(connect, cursor)
                                return False, '{}: incorrect response code {}'.format(err_prefix, i)

                            # validate the response code
                            if (not isinstance(i, int)) or (isinstance(i, bool)) or (200 > i > 599):
                                close_db_connection(connect, cursor)
                                return False, '{}: incorrect response code {}'.format(err_prefix, i)

                        # uri validation
                        elif i.startswith('@'):

                            # get upstream name
                            upstream = i.replace('@', '')

                            # search the upstream
                            res = upstream_search(err_prefix, waf_id, upstream)
                            if not res[0]:
                                close_db_connection(connect, cursor)
                                return res[0], res[1]

            # http://proxy_pass_upstream
            # memcached_pass_upstream
            elif k in ['proxy_pass', 'memcached_pass', 'nwaf_ws_proxy_pass']:

                # proxy_pass preprocessing
                if k in ['proxy_pass', 'nwaf_ws_proxy_pass']:

                    # should be starting with schema
                    if not re.match(r'^\w+://', v):
                        close_db_connection(connect, cursor)
                        return False, '{}: parameter "{}" contains an incorrect value "{}"'.format(err_prefix, k, v)

                    # remove the schema
                    v = v.split('://')
                    if len(v) == 1:
                        close_db_connection(connect, cursor)
                        return False, '{}: incorrect address'.format(err_prefix)
                    else:
                        v = v[1]

                    # if not UNIX-domain sockets
                    # (UNIX-domain sockets is allowed and not validated)
                    if not v.startswith('unix:'):

                        # remove the path
                        v = v.split('/')[0]

                        ##
                        # Address and port validation
                        ##

                        # server address
                        if (':' in v) or ('.' in v):

                            # extract host and port from the address
                            r = address_processing(v)
                            if not r[0]:
                                close_db_connection(connect, cursor)
                                return False, '{}: incorrect address "{}"'.format(err_prefix, v)
                            else:
                                host = r[0]
                                port = r[1]

                            # host validation
                            if host:
                                if not ip_domain_validation(host):
                                    close_db_connection(connect, cursor)
                                    return False, '{}: incorrect server address'.format(err_prefix)
                            else:
                                close_db_connection(connect, cursor)
                                return False, '{}: "host" is missing'.format(err_prefix)

                            # port validation
                            if port:
                                if (not isinstance(port, int)) or (isinstance(port, bool)) or (1 > port > 65535):
                                    close_db_connection(connect, cursor)
                                    return False, '{}: incorrect port "{}"'.format(err_prefix, port)

                        # upstream
                        else:

                            # search the upstream
                            res = upstream_search(err_prefix, waf_id, v)
                            if not res[0]:
                                close_db_connection(connect, cursor)
                                return res[0], res[1]

                # memcached_pass preprocessing
                elif k in ['memcached_pass']:

                    ##
                    # Address and port validation
                    ##

                    # if not UNIX-domain sockets
                    # (UNIX-domain sockets is allowed and not validated)
                    if not v.startswith('unix:'):

                        # server address (must contain a port)
                        if ':' in v:

                            # extract host and port from the address
                            r = address_processing(v)
                            if not r[0]:
                                close_db_connection(connect, cursor)
                                return False, '{}: incorrect address "{}"'.format(err_prefix, v)
                            else:
                                host = r[0]
                                port = r[1]

                            # host validation
                            if host:
                                if not ip_domain_validation(host):
                                    close_db_connection(connect, cursor)
                                    return False, '{}: incorrect server address "{}"'.format(err_prefix, host)
                            else:
                                close_db_connection(connect, cursor)
                                return False, '{}: host is missing'.format(err_prefix)

                            # port validation
                            if port:
                                if (not isinstance(port, int)) or (isinstance(port, bool)) or (1 > port > 65535):
                                    close_db_connection(connect, cursor)
                                    return False, '{}: incorrect port "{}"'.format(err_prefix, port)
                            else:
                                close_db_connection(connect, cursor)
                                return False, '{}: port is missing'.format(err_prefix)

                        # upstream
                        else:

                            # search the upstream
                            res = upstream_search(err_prefix, waf_id, v)
                            if not res[0]:
                                close_db_connection(connect, cursor)
                                return res[0], res[1]

            # pass_type validation
            elif k in ['pass_type']:
                if v not in ['root', 'alias', 'proxy_pass', 'memcached_pass', 'nwaf_ws_proxy_pass']:
                    close_db_connection(connect, cursor)
                    return False, '{}: parameter "{}" contains an incorrect value "{}"'.format(err_prefix, k, v)

            # parameters with on/off
            elif k in ['autoindex', 'proxy_intercept_errors']:
                if v not in ['on', 'off']:
                    close_db_connection(connect, cursor)
                    return False, '{}: parameter "{}" contains an incorrect value "{}"'.format(err_prefix, k, v)

            # parameters with time
            elif k in ['ssl_session_timeout', 'proxy_connect_timeout', 'proxy_send_timeout', 'proxy_read_timeout']:

                # regex validation
                if not re.match(r'^\d+[smh]$', v.lower()):
                    close_db_connection(connect, cursor)
                    return False, '{}: parameter "{}" contains an incorrect value "{}"'.format(err_prefix, k, v)

            # certificate/dhparam/upstream
            elif k in ['certificate_name', 'dhparam_name']:

                # is not null
                if v is not None:

                    # set table name
                    item = 'certificate' if k == 'certificate_name' else 'dhparam'
                    table = 'nginx_' + item

                    # search item's name
                    query = """ SELECT COUNT(*) FROM {} WHERE (waf_id = %s AND name = %s) """.format(table)
                    cursor.execute(query, (waf_id, v,))
                    res = cursor.fetchone()

                    # processing the result
                    if not res or res[0] < 1:
                        close_db_connection(connect, cursor)
                        return False, '{}: "{}"/"{}" is not found'.format(err_prefix, item, v)

                    # certificate_name without ssl
                    elif k in ['certificate_name']:

                        # - listen is set and ssl not in listen
                        # or
                        # - listen is not set and ssl not in listen in the database

                        if ('listen' in data) and (not [x for x in data['listen'] if 'ssl' in x.split()]):
                            close_db_connection(connect, cursor)
                            return False, 'Validation failed: missing "ssl" in "listen" while "certificate_name" is used'

                        # listen is not set or no ssl in listen
                        elif ('listen' not in data) or (not [x for x in data['listen'] if 'ssl' in x.split()]):

                            # search listen:ssl in the database
                            query = """ SELECT listen FROM nginx_vhost WHERE (waf_id = %s AND name = %s) """
                            cursor.execute(query, (waf_id, data['name'],))
                            res = cursor.fetchone()

                            # if ssl not in the result
                            if (not res) or (not [x for x in literal_eval(res[0]) if 'ssl' in x.split()]):
                                close_db_connection(connect, cursor)
                                return False, 'Validation failed: missing "ssl" in "listen" while "certificate_name" is used'

        # close database connection
        close_db_connection(connect, cursor)

        # finish processing
        return True, ''

    except Exception as e:

        if k is None:
            err_msg = 'An error occurred during extended validation "{}" value: {}'.format(v, e)
        else:
            err_msg = 'An error occurred during extended validation "{}" value of "{}" key: {}'.format(v, k, e)

        return False, err_msg


def nginx_value_basic_validation(k, v):
    try:

        # value of 'name' type
        if k in nginx_type_name:

            if k in nginx_name:
                if v is None:
                    return True, ''
                elif not re.match(r'^[\w\-.]+$', v):
                    return False, 'The value "{}" of parameter "{}" contains an invalid characters'.format(v, k)
                else:
                    return True, ''

            elif k in nginx_ne_name:
                if not re.match(r'^[\w\-.]+$', v):
                    return False, 'The value "{}" of parameter "{}" contains an invalid characters'.format(v, k)
                else:
                    return True, ''

        # value is a str
        elif k in nginx_type_str:
            if v is None:
                return True, ''
            elif not isinstance(v, str):
                return False, 'The value "{}" of parameter "{}" is not a string or not a null'.format(v, k)
            else:
                return True, ''

        # value is a b64
        elif k in nginx_type_b64:
            if v is None:
                return True, ''
            elif not is_base64(v):
                return False, 'The value "{}" of parameter "{}" is not in base64 or not a null'.format(v, k)
            else:
                return True, ''

        # value is an int
        elif k in nginx_type_int:
            if v is None:
                return True, ''
            elif (not isinstance(v, int)) or (isinstance(v, bool)):
                return False, 'The value "{}" of parameter "{}" is not an integer or not a null'.format(v, k)
            else:
                return True, ''

        # value is a float
        elif k in nginx_type_float:
            if v is None:
                return True, ''
            elif not isinstance(v, float):
                return False, 'The value "{}" of parameter "{}" is not a float or not a null'.format(v, k)
            else:
                return True, ''

        # value is a bool
        elif k in nginx_type_bool:
            if not isinstance(v, bool):
                return False, 'The value "{}" of parameter "{}" is not a boolean'.format(v, k)
            else:
                return True, ''

        # value is a list
        elif k in nginx_type_list:
            if v is None:
                return True, ''
            else:
                if not isinstance(v, list):
                    return False, 'The value "{}" of parameter "{}" is not a list or not a null'.format(v, k)
                else:
                    return True, ''

        # value is a dict
        elif k in nginx_type_dict:
            if v is None:
                return True, ''
            else:
                if not isinstance(v, dict):
                    return False, 'The value "{}" of parameter "{}" is not a dictionary or not a null'.format(v, k)
                else:
                    return True, ''

        # value is a list of dict
        elif k in nginx_type_lod:
            if v is None:
                return True, ''
            else:
                for i in v:
                    if not isinstance(i, dict):
                        return False, 'The value "{}" of parameter "{}" is not a list of dictionaries or not a null'.format(v, k)
                return True, ''

    except Exception as e:

        if k is None:
            err_msg = 'An error occurred during validation "{}" value: {}'.format(v, e)
        else:
            err_msg = 'An error occurred during validation "{}" value of "{}" key: {}'.format(v, k, e)

        return False, err_msg

    # response
    err_msg = '({}/{}): parameter "{}" is not allowed'.format(k, v, k)
    return False, err_msg


def nginx_validation(waf_id, nginx_type, data, k, v):
    try:

        ##
        # Basic validation
        ##

        r = nginx_value_basic_validation(k, v)
        if not r[0]:
            return r[0], r[1]

        ##
        # Extended validation
        ##

        # skip None (None processing in basic validation)
        if v is None:
            return True, ''

        # list of dictionary
        if k in nginx_type_lod:

            # locations preprocessing
            if k in ['locations']:

                # find duplicates locations
                if len([x['location'] for x in v if 'location' in x]) != len(set([x['location'] for x in v if 'location' in x])):
                    err_msg = 'An error occurred while processing the locations: duplication of locations'.format(data['name'])
                    return False, '{}'.format(err_msg)

                # memcached_pass only with set
                if [x for x in v if 'memcached_pass' in x and 'set' not in x]:
                    err_msg = (
                        'An error occurred while processing the locations: "memcached_pass" can only be used with "set"'
                        .format(data['name'])
                    )
                    return False, '{}'.format(err_msg)

            # param processing
            for item in v:

                # processing the dict
                for k1, v1 in item.items():

                    # basic validation
                    r = nginx_value_basic_validation(k1, v1)
                    if not r[0]:
                        return r[0], r[1]

                    # extended validation
                    r = nginx_value_extend_validation(waf_id, nginx_type, data, k1, v1)
                    if not r[0]:
                        return r[0], r[1]

        # other types (e.g., str/int/list/dict)
        else:
            r = nginx_value_extend_validation(waf_id, nginx_type, data, k, v)
            if not r[0]:
                return r[0], r[1]

        ##
        # Certificate validation
        # (certificate and private key should be in data)
        ##

        if (nginx_type == 'certificate') and (k == 'certificate'):

            # if crt and key in data
            if data['certificate'] and data['certificate_key']:
                r = nginx_certificate_validation(data['name'], data['certificate'], data['certificate_key'])
                if not r[0]:
                    return r[0], r[1]
            else:
                err_msg = 'An error occurred while updating the certificate {}: certificate and private key are missing'.format(data['name'])
                return False, err_msg

        ##
        # DH parameters validation
        ##

        if (nginx_type == 'dhparam') and (k == 'dhparam'):
            r = nginx_dhparam_validation(data['name'], v)
            if not r[0]:
                return r[0], r[1]

    except Exception as e:
        err_msg = 'An error occurred during Nginx validation: {}/{}: {}'.format(k, v, e)
        log.error(err_msg)
        return False, err_msg

    return True, ''


def get_nginx(url, request, nginx_type, get_nginx_with_active, get_nginx_with_format, waf_id):
    try:

        # init
        resp_in_nginx_format = False

        # set table
        table = 'nginx_' + nginx_type

        # set response format
        if url in get_nginx_with_format:
            try:
                args = request.query_params.get('format').lower()
                if args == 'nginx':
                    resp_in_nginx_format = True
            except Exception:
                pass

        # init ORDER BY
        if url != get_nginx_global_uri:
            order_by = 'ORDER BY name'
        else:
            order_by = ''

        # set the query
        if url in get_nginx_with_active and resp_in_nginx_format:
            # if table with 'active' column and response in Nginx format
            query = """ SELECT to_jsonb(array_agg({} {})) FROM {} WHERE (waf_id = %s AND active IS TRUE) """.format(table, order_by, table)
        else:
            query = """ SELECT to_jsonb(array_agg({} {})) FROM {} WHERE (waf_id = %s) """.format(table, order_by, table)

        # get data from the database
        connect = init_db_connection()
        cursor = connect.cursor()
        cursor.execute(query, (waf_id,))
        res = cursor.fetchall()
        close_db_connection(connect, cursor)

        # processing result
        if not res or not res[0][0]:

            if resp_in_nginx_format:
                response = ''
            else:
                response = {'status': 'none'}
                response = json.dumps(response)

            # response
            return response

        else:
            try:

                # init
                response = ''
                data_list = res[0][0]

                # Nginx format
                if resp_in_nginx_format:

                    # data processing
                    for idx, data in enumerate(data_list.copy()):

                        # vhost processing
                        if nginx_type == 'vhost' and 'name' in data:
                            del data['name']

                        # remove unused columns
                        for k in ['active', 'waf_id']:
                            if k in data:
                                del data[k]

                        # remove empty/None values
                        data = {k: v for k, v in data.items() if v not in [None, '']}

                        # update data list
                        data_list[idx] = data

                    # convert data to Nginx format
                    for data in data_list:
                        response = response + nginx_format_processing(nginx_type, data)

                # JSON format
                else:

                    # init
                    res = {}

                    # add status
                    for idx, data in enumerate(data_list.copy()):

                        # remove unused columns
                        if 'waf_id' in data:
                            del data['waf_id']

                        # data processing
                        for k, v in data.items():

                            # dhparam/private key masking for non-secure URL
                            if url in [get_nginx_dhparam_uri, get_nginx_certificate_uri]:
                                if k in ['dhparam', 'certificate_key']:
                                    try:
                                        data[k] = base64.b64encode(
                                            base64.b64decode(v.encode('UTF-8'))[:40]
                                            + b'*****'
                                            + base64.b64decode(v.encode('UTF-8'))[-40:]
                                        ).decode('UTF-8')
                                    except Exception as e:
                                        log.error('An error occurred while masking the {} in {}: {}'.format(k, url, e))
                                        data[k] = ''

                            # list / list of dictionary  normalization
                            if k in nginx_type_list + nginx_type_dict + nginx_type_lod:
                                if v is not None:
                                    data[k] = literal_eval(v)

                        # update data list
                        data_list[idx] = data

                    # update result
                    res['status'] = 'success'
                    res['configuration'] = data_list

                    # response
                    response = json.dumps(res)

                # response
                return response

            except Exception as e:
                log.error('An error occurred while processing result from table "{}": {}'.format(table, e))

    except Exception as e:
        log.error('An error occurred in get_nginx: {}'.format(e))


def set_nginx(url, data, nginx_type, cols, agent_ip, waf_id):
    try:

        # set the table
        table = 'nginx_' + nginx_type

        ##
        # Check allowed command
        ##

        # init
        cmd = ''
        name = ''

        # search for allowed command in JSON
        for i in nginx_cmd:
            if i in data:
                cmd = i.lower()
                break

        # processing result
        if not cmd:

            # logging
            log.error('Required key missing in JSON from {} to {}'.format(agent_ip, url))

            # response
            response = {'status': 'fail', 'description': 'Required key missing'}
            return response

        ##
        # Processing JSON
        ##

        # convert all keys to lowercase
        data[cmd] = {k.lower(): v for k, v in data[cmd].items()}

        # required key check
        if table in ['nginx_vhost', 'nginx_upstream', 'nginx_certificate', 'nginx_dhparam']:
            req_k = 'name'
            if req_k not in data[cmd]:
                err_msg = 'The required key "{}" is missing'.format(req_k)
                log.error(err_msg)
                response = {'status': 'fail', 'description': err_msg}
                return response
            else:
                name = data[cmd]['name']

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # processing command
        if cmd == 'set':
            try:

                # init
                rdata = {}

                # processing data
                for k in cols:

                    if k in data[cmd]:

                        # init
                        v = data[cmd][k]

                        # value validations
                        r = nginx_validation(waf_id, nginx_type, data[cmd], k, v)
                        if not r[0]:
                            close_db_connection(connect, cursor)
                            response = {'status': 'fail', 'description': r[1]}
                            return response

                        # del vars
                        del r

                        # convert list/dictionary to string
                        if isinstance(v, list) or isinstance(v, dict):
                            v = str(v)

                        # update value lists
                        rdata[k] = v

                # check JSON parameters
                if not rdata:
                    close_db_connection(connect, cursor)
                    response = {'status': 'fail', 'description': 'Parameters are missing in JSON'}
                    return response
                else:
                    # add WAF ID
                    rdata['waf_id'] = waf_id

                # init
                k = list(rdata.keys())
                v = list(rdata.values())
                del rdata

                # set conflict columns list
                key_columns = ['waf_id'] if table == 'nginx_global' else ['waf_id', 'name']

                # insert the data
                query = """ INSERT INTO {} ({}) VALUES %s ON CONFLICT ({}) DO UPDATE SET ({}) = %s """.format(
                    table, ','.join(k), ','.join(key_columns), ','.join(k)
                )
                cursor.execute(query, (tuple(v), tuple(v),))

                # update the hash
                set_nginx_hash(table, waf_id)

                # close database connection
                close_db_connection(connect, cursor)

                # response
                response = {'status': 'success'}
                return response

            except Exception as e:
                log.error('An error occurred while set the Nginx settings: {}'.format(e))
                close_db_connection(connect, cursor)
                response = {'status': 'fail', 'description': str(e)}
                return response

        elif cmd == 'del':
            try:

                if table not in ['nginx_vhost', 'nginx_upstream', 'nginx_certificate', 'nginx_dhparam']:
                    err_msg = 'The command "{}" is not applicable to {}'.format(cmd, url)
                    log.error(err_msg)
                    close_db_connection(connect, cursor)
                    response = {'status': 'fail', 'description': err_msg}
                    return response

                # checking that the entity is not used
                if nginx_type == 'vhost':

                    # delete vhost
                    query = """ DELETE FROM {} WHERE (name = %s) AND (waf_id = %s) """.format(table)
                    cursor.execute(query, (name, waf_id,))

                    # update the hash
                    set_nginx_hash(table, waf_id)

                    # close database connection
                    close_db_connection(connect, cursor)

                    # response
                    response = {'status': 'success'}
                    return response

                elif nginx_type in ['certificate', 'dhparam']:

                    # search entity in nginx_vhost
                    column = 'certificate_name' if nginx_type == 'certificate' else 'dhparam_name'
                    query = """ SELECT COUNT(*) from nginx_vhost WHERE (waf_id = %s AND {} = %s) """.format(column)
                    cursor.execute(query, (waf_id, name,))
                    res = cursor.fetchone()

                    # processing result
                    if not res or res[0] < 1:
                        query = """ DELETE FROM {} WHERE (name = %s) AND (waf_id = %s) """.format(table)
                        cursor.execute(query, (name, waf_id,))
                    else:
                        err_msg = 'The {} "{}" in use'.format(nginx_type, name)
                        response = {'status': 'fail', 'description': err_msg}
                        return response

                    # update the hash
                    set_nginx_hash(table, waf_id)

                    # close database connection
                    close_db_connection(connect, cursor)

                    # response
                    response = {'status': 'success'}
                    return response

                elif nginx_type == 'upstream':

                    # search upstream name in the database
                    query = """ SELECT error_page, locations FROM nginx_vhost WHERE (waf_id = %s) """
                    cursor.execute(query, (waf_id,))
                    res = cursor.fetchall()

                    # processing result
                    for row in res:

                        # skip empty
                        if not row:
                            continue

                        # search the upstream in error_page
                        if row[0]:

                            # processing the values
                            for v in literal_eval(row[0]).values():

                                # skip empty value
                                if not v:
                                    continue

                                # strip the value
                                v = v.strip()

                                # if value contains the upstream
                                if v.endswith('@' + name):
                                    close_db_connection(connect, cursor)
                                    err_msg = 'Upstream "{}" is in use'.format(name)
                                    response = {'status': 'fail', 'description': err_msg}
                                    return response

                        # search the upstream in locations
                        if row[1]:

                            # processing for each location
                            for location in literal_eval(row[1]):

                                # skip empty location
                                if not location:
                                    continue

                                # search upstream in proxy_pass/nwaf_ws_proxy_pass
                                for k in ['proxy_pass', 'nwaf_ws_proxy_pass']:
                                    if k in location and location[k].strip().endswith('://' + name):
                                        close_db_connection(connect, cursor)
                                        err_msg = 'Upstream "{}" is in use'.format(name)
                                        response = {'status': 'fail', 'description': err_msg}
                                        return response

                                # search upstream in memcached_pass
                                k = 'memcached_pass'
                                if k in location and location[k].strip() == name:
                                    close_db_connection(connect, cursor)
                                    err_msg = 'Upstream "{}" is in use'.format(name)
                                    response = {'status': 'fail', 'description': err_msg}
                                    return response

                                # search the upstream in error_page
                                k = 'error_page'
                                if k in location:

                                    # processing the values
                                    for v in location[k].values():

                                        # skip empty value
                                        if not v:
                                            continue

                                        # strip the value
                                        v = v.strip()

                                        # if value contains the upstream
                                        if v.endswith('@' + name):
                                            close_db_connection(connect, cursor)
                                            err_msg = 'Upstream "{}" is in use'.format(name)
                                            response = {'status': 'fail', 'description': err_msg}
                                            return response

                    # delete the entity
                    query = """ DELETE FROM {} WHERE (name = %s) AND (waf_id = %s)""".format(table)
                    cursor.execute(query, (name, waf_id,))

                    # update the hash
                    set_nginx_hash(table, waf_id)

                    # close database connection
                    close_db_connection(connect, cursor)

                    # response
                    response = {'status': 'success'}
                    return response

            except Exception as e:
                log.error('An error occurred while deleting "{}" entity: {}'.format(nginx_type, e))
                close_db_connection(connect, cursor)
                response = {'status': 'fail', 'description': str(e)}
                return response

        else:
            close_db_connection(connect, cursor)
            err_msg = 'Invalid command received: {}'.format(cmd)
            response = {'status': 'fail', 'description': err_msg}
            return response

    except Exception as e:
        log.error('An error occurred in set_nginx: {}'.format(e))
