#!/usr/bin/env python3

"""
Nemesida WAF API
Copyright (c) 2017-2025 Pentestit LLC - All Rights Reserved
Author: Romanov R.
"""

import base64
import hashlib
import json
import jsonref
import re

from openapi_spec_validator import validate
from psycopg2.extras import RealDictCursor

from core import idna_normalization
from core import init_db_connection, close_db_connection
from core import memc
from core import memc_lock

from logger import log


def schema_name_after_change(old_name, new_name, waf_id):

    # update the hash
    if not set_openapi_hash('openapi_schemas', waf_id):
        return False

    # init database connection
    connect = init_db_connection()
    cursor = connect.cursor(cursor_factory=RealDictCursor)

    ##
    # Update the URL
    ##

    try:

        # search the data associated with the old schema name
        query = """ SELECT CONVERT_FROM(url,'UTF8') AS url FROM openapi_url WHERE (waf_id = %s) """
        cursor.execute(query, (waf_id,))
        rows = cursor.fetchone()

        # data processing
        if rows:

            try:
                url = json.loads(rows.get('url'))
            except json.JSONDecodeError:
                url = rows.strip('[]').split(',')

            url = [i.replace(old_name, new_name, 1) for i in url]

            try:
                url = json.dumps(url)
            except json.JSONDecodeError:
                url = str(url)

            query = """
                UPDATE
                    openapi_url
                SET
                    url = %s,
                    timestamp = CURRENT_TIMESTAMP
                WHERE
                    (waf_id = %s)
            """
            cursor.execute(query, (url, waf_id,))

            # update the hash
            if not set_openapi_hash('openapi_url', waf_id):
                close_db_connection(connect, cursor)
                return False

    except Exception as e:
        log.error('An error occurred while updating OpenAPI URL (WAF ID: {}): {}'.format(waf_id, e))

    ##
    # Update the paths
    ##

    try:

        # delete the data associated with the new schema name to avoid update conflicts.
        query = """ DELETE FROM openapi_schema_paths WHERE (schema_name = %s) AND (waf_id = %s) """
        cursor.execute(query, (new_name, waf_id,))

        # search the data associated with the old schema name
        query = """ SELECT id, url, schema_name FROM openapi_schema_paths WHERE (schema_name = %s) AND (waf_id = %s) """
        cursor.execute(query, (old_name, waf_id,))
        rows = cursor.fetchall()

        # data processing
        for i in rows:
            url = i.get('url', '').replace(old_name, new_name, 1)
            query = """
                UPDATE
                    openapi_schema_paths
                SET
                    url = %s,
                    schema_name = %s,
                    timestamp = CURRENT_TIMESTAMP
                WHERE
                    (id = %s)
            """
            cursor.execute(query, (url, new_name, i.get('id'),))

    except Exception as e:
        log.error('An error occurred while updating OpenAPI paths for schema {} (WAF ID: {}): {}'.format(old_name, waf_id, e))

    ##
    # Update the exclude paths
    ##

    try:

        # search the data associated with the old schema name
        query = """ SELECT CONVERT_FROM(exclude_paths,'UTF8') AS exclude_paths FROM openapi_exclude_paths WHERE (waf_id = %s) """
        cursor.execute(query, (waf_id,))
        rows = cursor.fetchone()

        # data processing
        if rows:

            # extract the data
            data = rows.get('exclude_paths')
            data = json.loads(data) if data else {}

            # data processing
            for idx, item in enumerate(data):

                # rename the schema
                for k, v in item.items():
                    if (k == 'schema_name') and (v == old_name):
                        item[k] = new_name

                # update the data
                data[idx] = item

                # del vars
                del item

            # convert to JSON
            data = json.dumps(data)

            query = """ UPDATE openapi_exclude_paths SET exclude_paths = %s, timestamp = CURRENT_TIMESTAMP WHERE (waf_id = %s) """
            cursor.execute(query, (data, waf_id,))

            # del vars
            del data

    except Exception as e:
        log.error('An error occurred while updating OpenAPI exclude paths (WAF ID: {}): {}'.format(waf_id, e))

    # close database connection
    close_db_connection(connect, cursor)

    # update the hash
    return set_openapi_hash('openapi_exclude_paths', waf_id)


def oa_validation_rwo(data, status, status_fail):
    try:

        ##
        # OpenAPI readOnly/writeOnly simultaneous use validation
        ##

        # dict processing
        if isinstance(data, (dict, jsonref.JsonRef)):

            # if readOnly and writeOnly in data
            if 'readOnly' in data and 'writeOnly' in data:

                # if they are both True
                if data['readOnly'] and data['writeOnly']:
                    status = status_fail
                    return data, status, status_fail

            # processing the data values
            for k, v in data.copy().items():

                # dict processing
                if isinstance(v, (dict, jsonref.JsonRef)):
                    data[k], status, status_fail = oa_validation_rwo(v, status, status_fail)

                # list processing
                elif isinstance(v, list):

                    # reset the value
                    data[k] = []

                    # update the value and the status
                    for x in v:
                        if not isinstance(x, str):
                            r = oa_validation_rwo(v, status, status_fail)
                            data[k].append(r[0])
                            status = status_fail if r[1] == status_fail else status

                # other type processing
                else:
                    data[k] = v

        # finish processing
        else:
            return data, status, status_fail

    except Exception as e:
        log.error('An error occurred while OpenAPI schema validation (readOnly/writeOnly): {}'.format(e))

    return data, status, status_fail


def set_openapi_hash(table, waf_id):

    # init database connection
    connect = init_db_connection()
    cursor = connect.cursor()

    if table == 'openapi_schemas':

        field = 'schema'
        memc_key = 'hash_schemas_' + waf_id
        query = """
            SELECT
                timestamp, CONVERT_FROM(schema,'UTF8') AS schema
            FROM
                openapi_schemas
            WHERE
                waf_id = %s
            ORDER BY
                timestamp
            DESC
        """

    elif table == 'openapi_url':

        memc_key = 'hash_url_' + waf_id
        field = 'url'
        query = """
            SELECT
                timestamp, CONVERT_FROM(url,'UTF8') AS url
            FROM
                openapi_url
            WHERE
                waf_id = %s
        """

    elif table == 'openapi_exclude_paths':

        memc_key = 'hash_exclude_path_' + waf_id
        field = 'exclude_paths'
        query = """
            SELECT
                timestamp, CONVERT_FROM(exclude_paths,'UTF8') AS exclude_paths
            FROM
                openapi_exclude_paths
            WHERE
                waf_id = %s
        """

    else:
        log.error('An error occurred while generate hash: table {} is not valid'.format(table))
        close_db_connection(connect, cursor)
        return False

    cursor.execute(query, (waf_id,))
    rows = cursor.fetchall()

    h = waf_id
    if field == 'schema':
        if rows:
            timestamp = rows[0][0]
            h += str(timestamp)

        for i in rows:
            h += str(i[1])
    else:
        if rows:
            timestamp = rows[0][0]
            h += str(timestamp)
            try:
                rows = json.loads(rows[0][1])
            except json.JSONDecodeError:
                rows = rows[0][1].strip('[]').split(',')

        for i in rows:
            h += str(i)

    if rows:

        # get data hash
        hash_elms = hashlib.sha256(h.encode('UTF-8')).hexdigest()

        # update data in the memcached
        with memc_lock:
            try:
                memc.set(memc_key, hash_elms)
            except Exception as e:
                log.error('An error occurred while placing data into the memcached: {}'.format(e))
    else:

        # delete data from the memcached
        with memc_lock:
            try:
                memc.delete(memc_key)
            except Exception as e:
                log.error('An error occurred while deleting data from the memcached: {}'.format(e))

    # close database connection
    close_db_connection(connect, cursor)

    # response
    return True


def get_openapi_schema(waf_id):
    try:

        # search data
        query = """ SELECT name AS schema_name, convert_from(schema,'UTF8') AS schema FROM openapi_schemas WHERE (waf_id = %s) """
        connect = init_db_connection()
        cursor = connect.cursor(cursor_factory=RealDictCursor)
        cursor.execute(query, (waf_id,))
        rows = cursor.fetchall()
        close_db_connection(connect, cursor)

        # response
        return {'schemas': rows}

    except Exception as e:
        log.error('An error occurred in set_openapi_exclude_paths: {}'.format(e))


def set_openapi_schema(data, agent_ip, waf_id):
    try:

        # init
        if data.get('set'):
            cmd = 'set'
            schema_data = data.get('set')
        elif data.get('del'):
            cmd = 'del'
            schema_data = data.get('del')
        else:
            return {'status': 'fail', 'description': 'Incorrect action'}

        # extract the schema name
        schema_name = schema_data.get('schema_name')

        # schema name processing
        if not schema_name:
            return {'status': 'fail', 'description': 'The OpenAPI name of schema is not set'}

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # create/update the schema
        if cmd == 'set':

            # extract old schema name
            old_name = schema_data.get('old_name')

            # extract the schema
            schema = schema_data.get('schema')

            # if old name and the schema are not set
            if not old_name and not schema:
                close_db_connection(connect, cursor)
                return {'status': 'fail', 'description': 'Invalid OpenAPI schema'}

            # schema validation
            if schema:
                try:

                    # load the schema
                    s = json.loads(
                        base64.b64decode(schema).decode('UTF-8')
                    )

                    # schema validation: basic
                    try:
                        validate(s)
                    except Exception as e:
                        close_db_connection(connect, cursor)
                        return {'status': 'fail', 'description': 'Schema validation failed ({})'.format(e)}

                    # schema validation: readOnly w/ writeOnly
                    r = oa_validation_rwo(s.copy(), '', 'fail')
                    if r[1] == 'fail':
                        close_db_connection(connect, cursor)
                        return {'status': 'fail', 'description': 'Schema validation failed (readOnly and writeOnly cannot be used at the same time)'}

                except Exception:
                    close_db_connection(connect, cursor)
                    return {'status': 'fail', 'description': 'OpenAPI schema is not in JSON format'}

            # search for a schema's name
            query = """ SELECT id FROM openapi_schemas WHERE (waf_id = %s AND name = %s) """

            # search for a schema name
            cursor.execute(query, (waf_id, schema_name))
            rows = cursor.fetchone()

            # search for old schema name
            if old_name:

                # if new schema name already exists
                if rows:
                    close_db_connection(connect, cursor)
                    return {'status': 'fail', 'description': 'The OpenAPI schema "{}" already exists'.format(schema_name)}

                # if old schema name does not exist
                cursor.execute(query, (waf_id, old_name))
                if not cursor.fetchone():
                    close_db_connection(connect, cursor)
                    return {'status': 'fail', 'description': 'The OpenAPI schema "{}" does not exist'.format(old_name)}

            # if the schema exists or old name is set
            if rows or old_name:

                # if old name is set
                if old_name:

                    if schema:
                        query = """
                            UPDATE
                                openapi_schemas
                            SET
                                schema = %s,
                                timestamp = CURRENT_TIMESTAMP,
                                name = %s
                            WHERE
                                (waf_id = %s AND name = %s)
                        """
                        cursor.execute(query, (schema, schema_name, waf_id, old_name))
                    else:
                        query = """
                            UPDATE
                                openapi_schemas
                            SET
                                timestamp = CURRENT_TIMESTAMP,
                                name = %s
                            WHERE
                                (waf_id = %s AND name = %s)
                        """
                        cursor.execute(query, (schema_name, waf_id, old_name))

                    # update schema's data
                    if not schema_name_after_change(old_name, schema_name, waf_id):
                        log.error(
                            'The OpenAPI schema "{}" received from client {} has not been updated'
                            .format(old_name, agent_ip)
                        )
                        close_db_connection(connect, cursor)
                        return {'status': 'fail', 'description': 'The OpenAPI schema has not been updated'}
                    else:
                        log.info(
                            'The OpenAPI schema "{}" received from client {} has been updated'
                            .format(old_name, agent_ip)
                        )

                # if old name is not set
                else:

                    query = """
                        UPDATE
                            openapi_schemas
                        SET
                            schema = %s,
                            timestamp = CURRENT_TIMESTAMP
                        WHERE
                            (waf_id = %s AND name = %s)
                    """
                    cursor.execute(query, (schema, waf_id, schema_name))

                    # update the hash
                    if not set_openapi_hash('openapi_schemas', waf_id):
                        log.error(
                            'The OpenAPI schema "{}" received from client {} has not been updated'
                            .format(schema_name, agent_ip)
                        )
                        close_db_connection(connect, cursor)
                        return {'status': 'fail'}
                    else:
                        log.info(
                            'The OpenAPI schema "{}" received from client {} has been updated'
                            .format(schema_name, agent_ip)
                        )

            # if the schema does not exist
            else:

                # insert the data
                query = """ INSERT INTO openapi_schemas (schema, waf_id, name) VALUES (%s, %s, %s) RETURNING id """
                cursor.execute(query, (schema, waf_id, schema_name))

                # update the hash
                if not set_openapi_hash('openapi_schemas', waf_id):
                    log.error(
                        'The OpenAPI schema "{}" received from client {} has not been created'
                        .format(schema_name, agent_ip)
                    )
                    close_db_connection(connect, cursor)
                    return {'status': 'fail'}
                else:
                    log.info(
                        'The OpenAPI schema with name "{}" received from client {} has been created'
                        .format(schema_name, agent_ip)
                    )

            # response
            close_db_connection(connect, cursor)
            return {'status': 'success'}

        # delete the schema
        elif cmd == 'del':

            # delete data
            query = """ DELETE FROM openapi_schemas WHERE (name = %s) AND (waf_id = %s)"""
            cursor.execute(query, (schema_name, waf_id,))

            # update the hash
            if not set_openapi_hash('openapi_schemas', waf_id):
                close_db_connection(connect, cursor)
                return {'status': 'fail'}

            # response
            close_db_connection(connect, cursor)
            return {'status': 'success'}

        # close database connection
        close_db_connection(connect, cursor)

        # response
        return {'status': 'fail'}

    except Exception as e:
        log.error('An error occurred in set_openapi_schema: {}'.format(e))


def set_openapi_schema_paths(set_path, del_path, waf_id):
    try:

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # update/insert the data
        if set_path:

            # extract the request data
            schema_name = set_path.get('schema_name')
            schema_url = set_path.get('url')
            schema = set_path.get('schema')

            # data validation
            if schema_name is None:
                schema_name = del_path.get('name')
            if not schema_name:
                close_db_connection(connect, cursor)
                return {'status': 'fail', 'description': 'An empty schema name'}

            # search data
            query = """ SELECT id FROM openapi_schema_paths WHERE (schema_name = %s AND url = %s AND waf_id = %s) """
            cursor.execute(query, (schema_name, schema_url, waf_id))
            rows = cursor.fetchone()

            # update data
            if rows:

                query = """
                    UPDATE
                        openapi_schema_paths
                    SET
                        schema_name = %s,
                        url = %s,
                        schema = %s,
                        timestamp = CURRENT_TIMESTAMP
                    WHERE
                        (schema_name = %s AND url = %s AND waf_id = %s)
                """
                cursor.execute(query, (schema_name, schema_url, schema, schema_name, schema_url, waf_id))

            # insert data
            else:

                query = """
                    INSERT INTO openapi_schema_paths (
                        schema_name, schema, url, waf_id
                    ) VALUES (
                        %s, %s, %s, %s
                    )
                """
                cursor.execute(query, (schema_name, schema, schema_url, waf_id))

            # response
            close_db_connection(connect, cursor)
            return {'status': 'success'}

        # delete the data
        elif del_path:

            # extract the request data
            schema_name = del_path.get('schema_name')
            schema_url = del_path.get('url')

            # data validation
            if schema_name is None:
                schema_name = del_path.get('name')
            if not schema_name:
                return {'status': 'fail', 'description': 'An empty schema name'}

            query = """ DELETE FROM openapi_schema_paths WHERE (url = %s) AND (schema_name = %s) AND (waf_id = %s) """
            cursor.execute(query, (schema_url, schema_name, waf_id))

            # response
            close_db_connection(connect, cursor)
            return {'status': 'success'}

        # close database connection
        close_db_connection(connect, cursor)

        # response
        return {'status': 'fail'}

    except Exception as e:
        log.error('An error occurred in set_openapi_schema_paths: {}'.format(e))


def get_openapi_url(waf_id):
    try:

        # search the data
        query = """ SELECT CONVERT_FROM(url,'UTF8') AS urls FROM openapi_url WHERE (waf_id = %s) """
        connect = init_db_connection()
        cursor = connect.cursor()
        cursor.execute(query, (waf_id,))
        rows = cursor.fetchone()
        close_db_connection(connect, cursor)

        # result processing
        if rows:
            try:

                # extract paths list
                rows = json.loads(rows[0])

                # replace any to str
                for idx, row in enumerate(rows):
                    rows[idx] = re.sub(r'{\w+}', '{str}', row)

            except json.JSONDecodeError:
                rows = []

        # response
        return {'url': rows}

    except Exception as e:
        log.error('An error occurred in get_openapi_url: {}'.format(e))


def set_openapi_url(data, waf_id):
    try:

        # load data as list
        schema_url = data.get('url')

        ##
        # IDNA vhost normalization
        ##

        # init
        response = []

        # processing for each schema in schema's list
        for i in schema_url:

            # extract vhost/path
            vhost = i.split('/', 1)[0]
            path = '/' + i.split('/', 1)[1]

            # non-latin vhost normalization
            vhost = idna_normalization(vhost)

            # update list
            response.append(vhost + path)

        # update schema_url
        schema_url = json.dumps(response.copy())

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # processing data
        if schema_url:

            if not schema_url.strip('[]'):

                query = """ DELETE FROM openapi_url WHERE (waf_id = %s) """
                cursor.execute(query, (waf_id,))

                # delete data from the memcached
                memc_key = 'hash_url_' + waf_id
                with memc_lock:
                    try:
                        memc.delete(memc_key)
                    except Exception as e:
                        log.error('An error occurred while deleting data from the memcached: {}'.format(e))

            else:

                query = """ SELECT id FROM openapi_url WHERE (waf_id = %s) """
                cursor.execute(query, (waf_id,))
                rows = cursor.fetchone()

                if rows:

                    # update data
                    query = """ UPDATE openapi_url SET url = %s, timestamp = CURRENT_TIMESTAMP WHERE (waf_id = %s) """
                    cursor.execute(query, (schema_url, waf_id))

                    # update the hash
                    if not set_openapi_hash('openapi_url', waf_id):
                        close_db_connection(connect, cursor)
                        return {'status': 'fail'}

                else:

                    # insert the data
                    query = """ INSERT INTO openapi_url (url, waf_id) VALUES (%s, %s) RETURNING id """
                    cursor.execute(query, (schema_url, waf_id))

                    # update the hash
                    if not set_openapi_hash('openapi_url', waf_id):
                        close_db_connection(connect, cursor)
                        return {'status': 'fail'}

            # response
            close_db_connection(connect, cursor)
            return {'status': 'success'}

        # close database connection
        close_db_connection(connect, cursor)

        # response
        return {'status': 'fail'}

    except Exception as e:
        log.error('An error occurred in set_openapi_url: {}'.format(e))


def get_openapi_schema_paths(waf_id):
    try:

        # init
        result = []

        # receive the data from the database
        query = """ SELECT schema_name, url, CONVERT_FROM(schema,'UTF8') AS schema FROM openapi_schema_paths WHERE (waf_id = %s) """
        connect = init_db_connection()
        cursor = connect.cursor()
        cursor.execute(query, (waf_id,))
        res = cursor.fetchall()
        close_db_connection(connect, cursor)

        # data processing
        for row in res:

            # skip empty
            if not row:
                continue

            # update the result
            result.append({
                'schema_name': row[0],
                'url': row[1],
                'schema': row[2]
            })

        # response
        return {'paths': result}

    except Exception as e:
        log.error('An error occurred in get_openapi_schema_paths: {}'.format(e))


def get_openapi_exclude_paths(waf_id):
    try:

        # search data in the database
        query = """ SELECT CONVERT_FROM(exclude_paths,'UTF8') AS exclude_paths FROM openapi_exclude_paths WHERE (waf_id = %s) """
        connect = init_db_connection()
        cursor = connect.cursor()
        cursor.execute(query, (waf_id,))
        rows = cursor.fetchone()
        close_db_connection(connect, cursor)

        # result processing
        if rows:
            try:
                rows = json.loads(rows[0])
            except json.JSONDecodeError:
                rows = rows

        # close database connection
        close_db_connection(connect, cursor)

        # response
        return {'exclude_paths': rows}

    except Exception as e:
        log.error('An error occurred in get_openapi_exclude_paths: {}'.format(e))


def set_openapi_exclude_paths(data, waf_id):
    try:

        # data validation
        if (data is None) or (not isinstance(data, dict)):
            return {'status': 'fail', 'description': 'Incorrect data format'}

        # extract set data
        data = data.get('set')

        # none/missing set data
        if (data is None) or (not isinstance(data, dict)):
            return {'status': 'fail', 'description': 'Incorrect data format'}

        # extract exclude_paths data
        data = data.get('exclude_paths')

        # data validation
        if data:

            # init
            required = ['schema_name', 'url']
        
            # delete extra fields
            for idx, x in enumerate(data):
                data[idx] = {k: v for k, v in x.items() if k in required}

            # data consistent validation
            for item in data:

                # required processing
                for x in required:

                    # required keys
                    if x not in item:
                        return {'status': 'fail', 'description': 'Key "{}" is missing'.format(x)}
                    
                    # value validation: schema_name
                    if (x == 'schema_name') and (not isinstance(item.get(x), str)):
                        return {'status': 'fail', 'description': 'Value of "{}" contains incorrect data'.format(x)}
                    
                    # value validation: url
                    if (x == 'url') and (item.get(x)) and (not isinstance(item.get(x), str)):
                        return {'status': 'fail', 'description': 'Value of "{}" contains incorrect data'.format(x)}

        # empty data
        if not data:
            data = {}

        # data normalization
        data = json.dumps(data)

        # Update data in the database
        query = """
            INSERT INTO openapi_exclude_paths (
                exclude_paths, waf_id
            ) VALUES (
                %s, %s
            ) ON CONFLICT (
                waf_id
            ) DO UPDATE SET
                exclude_paths = %s,
                timestamp = CURRENT_TIMESTAMP
            """
        connect = init_db_connection()
        cursor = connect.cursor()
        cursor.execute(query, (data, waf_id, data,))
        close_db_connection(connect, cursor)

        # update the hash
        if not set_openapi_hash('openapi_exclude_paths', waf_id):
            return {'status': 'fail', 'description': 'An error occurred while updating data in the database'}
        else:
            return {'status': 'success'}

    except Exception as e:
        log.error('An error occurred in set_openapi_exclude_paths: {}'.format(e))
