#!/usr/bin/env python3

"""
Nemesida WAF API
Copyright (c) 2017-2025 Pentestit LLC - All Rights Reserved
Author: Romanov R.
"""

import base64
import hashlib
import json
import re
import validators

from ast import literal_eval
from datetime import datetime
from netaddr import IPNetwork, IPRange
from urllib.parse import urlparse, urljoin

from core import domain_validation
from core import get_auth
from core import get_log_key
from core import idna_normalization
from core import init_db_connection, close_db_connection
from core import k2lower
from logger import log

##
# MLC settings keys
##

# set forbidden columns by tarif_id
mlc_be_cols = ['main__ai_extra', 'openapi__vhosts_list']

# set columns
mlc_settings_cols = [
    'main__ai_extra',
    'ddos__enable',
    'ddos__wl_ip',
    'ddos__wl_url',
    'ddos__interval',
    'ddos__latest_only',
    'ddos__send_possible',
    'brute__enable',
    'brute__brute_detect',
    'brute__flood_detect',
    'brute__wl_ip',
    'brute__wl_host',
    'brute__interval',
    'brute__max_val',
    'brute__latest_only',
    'brute__send_possible',
    'openapi__vhosts_list'
]
mlc_settings_cols_v2 = [
    'brute__bf_detect'
]

# set types
mlc_settings_type_domain = ['brute__wl_host', 'openapi__vhosts_list']
mlc_settings_type_ip = ['ddos__wl_ip', 'brute__wl_ip']
mlc_settings_type_url = ['ddos__wl_url', 'brute__brute_detect', 'brute__flood_detect']
mlc_settings_type_int = ['ddos__interval', 'brute__interval', 'brute__max_val']
mlc_settings_type_lod = ['brute__bf_detect']
mlc_settings_type_bool = [
    'main__ai_extra', 'ddos__enable', 'ddos__latest_only', 'ddos__send_possible',
    'brute__enable', 'brute__latest_only', 'brute__send_possible'
]

mlc_settings_idna = ['ddos__wl_url', 'brute__wl_host', 'brute__brute_detect', 'brute__flood_detect', 'openapi__vhosts_list']

##
# Dyn settings keys
##

##
# Set types
##

dyn_settings_type_int = [
    'nwaf_antibot_mem'
]

dyn_settings_type_limit = [
    'nwaf_limit'
]

dyn_settings_type_zone = [
    'nwaf_b64_decode_disable'
]

dyn_settings_type_ip = [
    'nwaf_ip_wl', 'nwaf_ip_lm',
    'nwaf_openapi_ip_wl',
    'nwaf_openapi_ip_lm',
    'nwaf_antibot_ip_wl',
    'nwaf_av_ip_lm',
    'nwaf_av_ip_wl'
]

dyn_settings_type_domain_ip = [
    'nwaf_log_request'
]

dyn_settings_type_url = [
    'nwaf_body_exclude',
    'nwaf_body_bin_exclude',
    'nwaf_url_wl',
    'nwaf_openapi_url_wl',
    'nwaf_openapi_url_lm',
    'nwaf_rfc_violation_wl',
    'nwaf_rfc_violation_lm',
    'nwaf_antibot_url',
    'nwaf_antibot_url_wl',
    'nwaf_av_url_lm',
    'nwaf_av_url_wl'
]

dyn_settings_type_domain = [
    'nwaf_ddos_detect_host_lm',
    'nwaf_host_lm', 'nwaf_host_wl',
    'nwaf_ai_extra_host_wl',
    'nwaf_ai_extra_host_lm',
    'nwaf_bf_detect_host_lm',
    'nwaf_geoip_mla_disable',
    'nwaf_mla_host_lm',
    'nwaf_put_body_exclude',
    'nwaf_post_body_exclude',
    'nwaf_rmq_host_exclude'
]

dyn_settings_inda = [
    'nwaf_limit',
    'nwaf_ip_lm',
    'nwaf_ip_wl',
    'nwaf_url_wl',
    'nwaf_host_lm',
    'nwaf_host_wl',
    'nwaf_mla_host_lm',
    'nwaf_body_exclude',
    'nwaf_openapi_ip_wl',
    'nwaf_openapi_ip_lm',
    'nwaf_openapi_url_wl',
    'nwaf_openapi_url_lm',
    'nwaf_ai_extra_host_lm',
    'nwaf_ai_extra_host_wl',
    'nwaf_body_bin_exclude',
    'nwaf_post_body_exclude',
    'nwaf_put_body_exclude',
    'nwaf_rfc_violation_lm',
    'nwaf_rfc_violation_wl',
    'nwaf_rmq_host_exclude',
    'nwaf_bf_detect_host_lm',
    'nwaf_geoip_mla_disable',
    'nwaf_b64_decode_disable',
    'nwaf_ddos_detect_host_lm',
    'nwaf_log_request',
    'nwaf_antibot_url',
    'nwaf_antibot_url_wl',
    'nwaf_antibot_ip_wl',
    'nwaf_av_ip_lm',
    'nwaf_av_ip_wl',
    'nwaf_av_url_lm',
    'nwaf_av_url_wl'
]

dyn_settings_type_cc = [
    'nwaf_antibot_cc_wl'
]

# set columns
dyn_settings_cols = \
    ['wl', 'rl'] + \
    dyn_settings_type_int + \
    dyn_settings_type_limit + \
    dyn_settings_type_zone + \
    dyn_settings_type_url + \
    dyn_settings_type_domain + \
    dyn_settings_type_ip + \
    dyn_settings_type_domain_ip + \
    dyn_settings_type_cc

##
# Dyn WL keys
##

# set columns
dyn_wl_cols = [
    'id',
    'domain',
    'mz',
    'extension',
    'rl_id'
]

##
# Dyn BL keys
##

# set columns
dyn_bl_cols = ['bl']

##
# Dyn ERL keys
##

# set types
dyn_erl_type_boolean = ['api', 'lm', 'noban', 'no_cookie', 'no_ua', 'no_referer', 'block_tor', 'block_proxy']
dyn_erl_type_data = ['ip', 'country', 'domain', 'url', 'args', 'body', 'cookie', 'ua', 'referer', 'other_headers']
dyn_erl_type_data_v2_cols = ['url_x', 'args_x', 'body_x', 'cookie_x', 'ua_x', 'referer_x', 'other_headers_x', 'zone_max_size']
dyn_erl_type_data_v2 = dyn_erl_type_data + dyn_erl_type_data_v2_cols
dyn_erl_type_required = dyn_erl_type_data + ['no_cookie', 'no_ua', 'no_referer', 'block_tor', 'block_proxy']
dyn_erl_type_required_v2 = dyn_erl_type_required + dyn_erl_type_data_v2_cols

# set columns
dyn_erl_cols = dyn_erl_type_boolean + dyn_erl_type_data
dyn_erl_cols_v2 = dyn_erl_cols + dyn_erl_type_data_v2_cols

##
# DynSyncBan settings
##

dyn_sync_ban_dict = ['ban_data']
dyn_sync_ban_int = ['timestamp']
dyn_sync_ban_cols = dyn_sync_ban_dict + dyn_sync_ban_int
dyn_sync_ban_memc_ttl = 60

##
# NWS target keys
##

nws_target_cols = [
    'active', 'target', 'scan_proxy', 'auth_params', 'exclude_modules',
]

nws_target_ap_cols = [
    'active', 'url', 'username_field', 'username_value', 'password_field', 'password_value',
]

##
# Set list of allowed command
##

settings_cmd = ['set', 'del']  # (Dyn, MLC)
dyn_wl_erl_cmd = ['add', 'upd', 'del']  # Dyn WL/ERL/BL


def base64_decode(agent_ip, k, v, s):

    # convert to binary
    try:

        # validation
        status = base64.b64encode(base64.b64decode(s.encode('UTF-8'))).decode('UTF-8', 'ignore') == s

        # check status
        if not status:
            return False

        # return decoded
        return base64.b64decode(s).decode('UTF-8')

    except Exception:
        log.error(
            'An error occurred while B64 decoding value from {}: {}/{} ({})'
            .format(agent_ip, k, v, s)
        )
        return False


def dyn_ip_domain_validation(item_list):
    try:

        # init
        is_ip_exist = False

        # domain validation
        for item in item_list:

            # domain validation
            if item.startswith('domain='):

                # extract values
                domain = item.split('domain=')[1]

                # domain validation
                if not domain_validation(domain):
                    return False

            # IP validation
            else:

                # IP range
                if '-' in item:
                    try:
                        IPRange(item.split('-')[0], item.split('-')[1])
                        is_ip_exist = True
                    except Exception:
                        return False

                # IP/CIDR
                else:
                    try:
                        IPNetwork(item)
                        is_ip_exist = True
                    except Exception:
                        return False

        # IP does not exist
        if not is_ip_exist:
            return False

    except Exception as e:
        log.error('An error occurred while processing data in dyn_ip_domain_validation: {}'.format(e))
        return False

    return True


def dyn_domain_ip_validation(item_list):
    try:

        # init
        is_domain_exist = False

        # domain validation
        for item in item_list:

            # domain validation
            if item.startswith('domain='):

                # extract values
                domain = item.split('domain=')[1]

                # domain validation
                if not domain_validation(domain):
                    return False
                else:
                    is_domain_exist = True

            # IP validation
            else:

                # IP range
                if '-' in item:
                    try:
                        IPRange(item.split('-')[0], item.split('-')[1])
                    except Exception:
                        return False

                # IP/CIDR
                else:
                    try:
                        IPNetwork(item)
                    except Exception:
                        return False

        # domain does not exist
        if not is_domain_exist:
            return False

    except Exception as e:
        log.error('An error occurred while processing data in dyn_domain_ip_validation: {}'.format(e))
        return False

    return True


def dyn_api_validation(item_list):
    try:

        # domain validation
        for item in item_list:

            # api validation
            if (item.startswith('api=')) and (item not in ['api=yes', 'api=no']):
                return False

    except Exception as e:
        log.error('An error occurred while processing data in dyn_api_validation: {}'.format(e))
        return False

    return True


def get_hash(agent_ip, url, t, waf_id):
    try:

        ##
        # Get ts
        ##

        query = """ SELECT modify FROM {} WHERE waf_id = %s """.format(t)
        connect = init_db_connection()
        cursor = connect.cursor()
        cursor.execute(query, (waf_id,))
        res = cursor.fetchone()
        connect.close()

        # convert to timestamp
        if res:
            res = res[0].timestamp()
        else:
            res = 0

        # processing result
        if not res:
            response = str({'status': 'none'})
            return response, 200

        else:
            res = hashlib.md5(str(res).encode('UTF-8')).hexdigest()
            response = {'status': 'success', 'hash': str(res)}
            return str(response), 200

    except Exception as e:
        log.error('An error occurred in get_hash in {} from {}: {}'.format(agent_ip, url, e))

    # default response
    return None, 503


def set_hash(agent_ip, url, t, waf_id):
    try:

        ##
        # Update ts
        ##

        query = """ INSERT INTO {} (modify, waf_id) VALUES (%s, %s) ON CONFLICT (waf_id) DO UPDATE SET modify = %s """.format(t)
        connect = init_db_connection()
        cursor = connect.cursor()
        cursor.execute(query, (str(datetime.now()), waf_id, str(datetime.now()),))
        connect.close()

    except Exception as e:
        log.error('An error occurred in set_hash in {} from {}: {}'.format(agent_ip, url, e))


def mlc_value_validation(agent_ip, k, v):
    try:

        # skip empty
        if v is None:
            return True

        # active validation
        if k == 'active':
            return True if isinstance(v, bool) else False

        # remove spaces before and after
        if isinstance(v, str):
            v = v.strip()
            if not v:
                return True

        if k in mlc_settings_type_ip:

            # split to list
            v = v.split()

            # processing IPv4/IPv6/CIDR
            for item in v:

                # IP range
                if '-' in item:
                    # processing IP range
                    IPRange(item.split('-')[0], item.split('-')[1])
                else:
                    IPNetwork(item)

        elif k in mlc_settings_type_domain:

            # list processing
            for item in v:

                # processing '*'
                if item == '*':

                    # '*' is not allowed for this option
                    if k == 'brute__wl_host':
                        return False
                    else:
                        continue

                # domain validation
                if not domain_validation(item):
                    return False

        elif k in mlc_settings_type_url:

            # list processing
            for item in v:

                # '*' is not allowed for this type
                if item == '*':
                    return False

                # value validation
                if not settings_type_url_validation(item):
                    return False

        elif k in mlc_settings_type_int:

            # value normalization and validation
            if (not isinstance(v, int)) or (isinstance(v, bool)):
                return False

            # small value validation
            if v < 1:
                return False

            # interval
            if k in ['ddos__interval', 'brute__interval']:
                if v > 180:
                    return False

            # max_val
            elif k in ['brute__max_val']:
                if v > 900:
                    return False

            # other values
            else:
                if v > 999999:
                    return False

        elif k in mlc_settings_type_bool:
            if not isinstance(v, bool):
                return False

        elif k in mlc_settings_type_lod:

            # processing the items
            for item in v:

                # convert keys to lover
                item = {str(k).lower(): v for k, v in item.items()}

                # required keys processing
                if not item.get('url'):
                    return False

                # simultaneous use validation or missing both
                if item.get('brute') == item.get('flood'):
                    return False

                # processing the items
                for k1, v1, in item.items():

                    # key normalization
                    k1 = str(k1).lower()

                    # URL validation
                    if k1 == 'url':
                        if not settings_type_url_validation(v1):
                            return False

                    # bool validation
                    elif k1 in ['brute', 'flood']:
                        if not isinstance(v1, bool):
                            return False

                    # zone validation
                    elif k1 == 'zone':

                        # skip empty
                        if not v1:
                            continue

                        # type validation
                        if not isinstance(v1, list):
                            return False

                        # data validation
                        if any(
                            x for x in v1 if
                            (not x) or (not isinstance(x, str)) or (not x.lower() in ['args', 'body'])
                        ):
                            return False

                    # condition validation
                    elif k1 == 'condition':

                        # skip empty
                        if not v1:
                            continue

                        # type validation
                        if not isinstance(v1, dict):
                            return False

                        # data  processing
                        for k2, v2 in v1.items():

                            # non-empty key validation
                            if not k2:
                                return False

                            # key normalization
                            k2 = str(k2).lower()

                            # zone validation
                            if k2 not in ['args', 'body']:
                                return False

                            # skip empty value
                            if not v2:
                                continue

                            # value type validation
                            if not isinstance(v2, list):
                                return False

                            # list value type validation
                            if any(
                                [x for x in v2 if not isinstance(x, str)]
                            ):
                                return False

    except Exception as e:
        log.error('An error occurred while validating MLC key/value from {}: {}/{}: {}'.format(agent_ip, k, v, e))
        return False

    return True


def dyn_value_validation(agent_ip, k, v):
    try:

        # skip empty
        if v is None:
            return True

        # active validation
        if k == 'active':
            return True if isinstance(v, bool) else False

        # int validation
        if k in dyn_settings_type_int:

            # type validation
            if (not isinstance(v, int)) or (isinstance(v, bool)):
                return False

            # nwaf_antibot_mem validation
            if (k == 'nwaf_antibot_mem') and (v < 100):
                return False

            # default response
            return True

        # remove spaces before and after
        v = v.strip()

        # skip empty
        if not v:
            return True

        # split values to list
        vl = v.split(', ')

        # zone validation
        if k in dyn_settings_type_zone:

            # list processing
            for item in vl:

                for v1 in item.split():

                    # zone validation
                    if v1.startswith('z:'):

                        # extract the value
                        val = re.match(r'^z:([\w|]+)+$', v1)

                        # check value
                        if not val:
                            return False

                        # value validation
                        if not settings_type_zone_validation(val[1].split('|')):
                            return False

                    # url validation
                    else:

                        if not settings_type_url_validation(v1):
                            return False

        # URL validation
        elif k in dyn_settings_type_url:
            for item in vl:
                if not settings_type_url_validation(item):
                    return False

        # IP validation
        elif k in dyn_settings_type_ip:

            # processing items
            for item in vl:

                # IP/domain validation
                if not dyn_ip_domain_validation([x for x in item.split()]):
                    return False

                # check duplicates in value
                if len([x for x in item.split() if x.strip().startswith('domain=')]) > 1:
                    return False
                if len([x for x in item.split() if not x.strip().startswith('domain=')]) > 1:
                    return False

        # domain validation
        elif k in dyn_settings_type_domain:

            # list processing
            for item in vl:

                # domain validation
                if not domain_validation(item):
                    return False

                # check duplicates in value
                if len([x for x in item.split() if x.strip().startswith('domain=')]) > 1:
                    return False
                if len([x for x in item.split() if not x.strip().startswith('domain=')]) > 1:
                    return False

        # domain with IP validation
        elif k in dyn_settings_type_domain_ip:

            # list processing
            for item in vl:

                # check required
                if 'domain=' not in item:
                    return False

                # remove '*.' or '.' for future domain validation
                item = re.sub(r'^[*]?\.', '', item)

                # IP/domain validation
                if not dyn_domain_ip_validation(item.split()):
                    return False

                # check duplicates in value
                if len([x for x in item.split() if x.strip().startswith('domain=')]) > 1:
                    return False
                if len([x for x in item.split() if not x.strip().startswith('domain=')]) > 1:
                    return False

        # limit validation
        elif k in dyn_settings_type_limit:

            """
            rate=5r/m (required)
            block_time=300 (required)
            domain=example.com (optional)
            """

            # init
            opts = ['domain=', 'rate=', 'block_time=']

            # don't allow to use a space before/after '=' symbol or '=' in the end
            if re.search(r'\s=|=(\s|$)', str(vl)):
                return False

            # don't allow to use multiple settings without domain or 'domain=*'
            if len([x for x in vl if 'domain=' not in x or 'domain=* ' in x]) > 1:
                return False

            # don't allow to use multiple settings with a same domain
            dl = [re.sub(r'.*(domain=\S*).*', r'\1', x) for x in vl if 'domain=' in x]
            if len(dl) != len(set(dl)):
                return False

            # list processing
            for item in vl:

                # split item to list
                vl1 = item.split()

                # check required parameters
                if not any(x for x in vl1 if x.startswith('rate=')) or not any(x for x in vl1 if x.startswith('block_time=')):
                    return False

                # check duplicates in value
                for opt in opts:
                    if len([x for x in vl1 if x.strip().startswith(opt)]) > 1:
                        return False

                # list processing
                for v1 in vl1:

                    # name validation
                    if not v1.startswith(tuple(opts)):
                        return False

                    # domain validation
                    if v1.startswith('domain'):
                        return domain_validation(v1.split('domain=')[1])

                    # rate validation
                    if v1.startswith('rate'):

                        # extract value
                        rate_data = v1.split('rate=')[1]

                        # extract rate and intensity
                        rate = rate_data.split('r/')[0]
                        rate_intensity = rate_data.split('r/')[1]

                        # rate rate intensity validation
                        if int(rate) < 0:
                            return False

                        # rate validation
                        if rate_intensity not in ['s', 'm', 'h']:
                            return False

                    # block_time validation
                    if v1.startswith('block_time'):

                        # value validation
                        if int(v1.split('block_time=')[1]) < 0:
                            return False

        # cc validation
        elif k in dyn_settings_type_cc:

            # list processing
            for item in vl:

                # value validation
                if (not isinstance(item, str)) or (len(item) != 2):
                    return False

    except Exception as e:
        log.error('An error occurred while validating Dyn key/value from {}: {}/{}: {}'.format(agent_ip, k, v, e))
        return False

    return True


def settings_type_url_validation(item):
    try:

        # allow '*'
        if item == '*':
            return True

        # schema is now allowed
        if '://' in item:
            return False

        # replace '*' to example.com for future URL validation
        item = re.sub(r'^\*/', 'example.com/', item)

        # remove '*.' or '.' for future URL validation
        item = re.sub(r'^[*]?\.', '', item)

        # path normalization (remove '\\*' in path)
        item = urlparse(item).netloc + urlparse(item).path.replace('\\*', '')

        # add the schema for validation
        item = 'http://' + item

        # URL
        return False if not validators.url(item) else True

    except Exception as e:
        log.error('An error occurred while validating data {}: {}'.format(item, e))
        return False


def settings_type_zone_validation(item_list):
    try:

        list_zone = ['url', 'args', 'headers', 'body']

        # list processing
        for item in item_list:

            # check value
            if item.lower() not in list_zone:
                return False

        return True

    except Exception as e:
        log.error('An error occurred while processing data in settings_type_zone_validation: {}'.format(e))
        return False


def dyn_wl_value_validation(agent_ip, k, v):
    try:

        ##
        # RL/WL/LM validation
        ##

        # active validation
        if k == 'active':
            return True if isinstance(v, bool) else False

        # domain validation
        elif k == 'domain':

            # allow empty
            if not v:
                return True

            # init
            domain = v

            # domain validation
            return domain_validation(domain)

        # rule id for WL/LM validation
        if k == 'rl_id':
            if v != '*' and not re.search(r'^\d+$', v):
                return False

        # zone validation
        if k == 'mz':

            # split zone to list
            z_list = v.split('|')

            # check zone
            for item in z_list:

                # only allowed symbols
                if not re.match(r'^[\w-]+$', item, re.I):
                    return False

        # extension validation
        if k == 'extension':

            # split extension to list
            ext_list = v.split('|$')

            # check zone and extension
            for item in ext_list:

                # skip empty
                if not item:
                    continue

                # add "$" to extension if it removed while split
                if not item.startswith('$'):
                    item = '$' + item

                # only allowed symbols
                if not re.match(r'^\$[\w-]+(_x)?:', item, re.I):
                    return False

    except Exception as e:
        log.error('An error occurred while validating Dyn key/value from {}: {}/{}: {}'.format(agent_ip, k, v, e))
        return False

    return True


def dyn_bl_value_validation(agent_ip, k, v):
    try:

        # active validation
        if k == 'active':
            if not isinstance(v, bool):
                return False

        # BL validation
        elif k == 'bl':

            # type validation
            if not isinstance(v, (list, type(None))):
                return False

            # API validation
            if v and any(x for x in v if not dyn_api_validation(x.split())):
                return False

            # IP/domain validation
            if v and any(x for x in v if not dyn_ip_domain_validation(
                    [x for x in x.split() if not x.startswith('api=')]
                )
            ):
                return False

    except Exception as e:
        log.error(
            'An error occurred while validating BL key/value from {}: {}/{}: {}'
            .format(agent_ip, k, v, e)
        )
        return False

    return True


def dyn_erl_value_validation(agent_ip, k, v):
    try:

        # Bool validation
        if k in dyn_erl_type_boolean:
            if not isinstance(v, bool):
                return False

        # IP validation
        elif k == 'ip':

            # split values to list
            # and remove empty items
            vl = list(filter(None, re.split(r'[!&|]', v)))

            # list processing
            for item in vl:

                # IP range
                if '-' in item:
                    IPRange(item.split('-')[0], item.split('-')[1])
                # IP/CIDR
                else:
                    IPNetwork(item)

        # Country validation
        elif k == 'country':

            # split values to list
            # and remove empty items
            vl = list(filter(None, re.split(r'[!&|]', v)))

            # list processing
            for item in vl:

                # base64 decode
                try:

                    # decode
                    item = base64_decode(agent_ip, k, v, item)

                    # decode validation
                    if item is False:
                        return False

                except Exception as e:
                    log.error(
                        'An error occurred while B64 decoding ERL value from {}: {}/{} ({}): {}'
                        .format(agent_ip, k, v, item, e)
                    )
                    return False

                if not re.match(r'^[a-z]{2}$', item, re.I):
                    return False

        # Domain validation
        elif k == 'domain':

            # split values to list
            # and remove empty items
            vl = list(filter(None, re.split(r'[!&|]', v)))

            # list processing
            for item in vl:

                # base64 decode
                try:

                    # decode
                    item = base64_decode(agent_ip, k, v, item)

                    # decode validation
                    if item is False:
                        return False

                except Exception as e:
                    log.error(
                        'An error occurred while B64 decoding ERL value from {}: {}/{} ({}): {}'
                        .format(agent_ip, k, v, item, e)
                    )
                    return False

                # processing '*'
                if item == '*':
                    continue

                # remove '*.' or '.' for future domain validation
                item = re.sub(r'^[*]?\.', '', item)

                # processing localhost
                if item == 'localhost':
                    continue

                # check value
                if not validators.domain(item):
                    return False

        # URL validation
        elif k in ['url']:

            # split values to list
            # and remove empty items
            vl = list(filter(None, re.split(r'[!&|]', v)))

            # list processing
            for item in vl:

                # base64 decode
                try:

                    # decode
                    item = base64_decode(agent_ip, k, v, item)

                    # decode validation
                    if item is False:
                        return False

                except Exception as e:
                    log.error(
                        'An error occurred while B64 decoding ERL value from {}: {}/{} ({}): {}'
                        .format(agent_ip, k, v, item, e)
                    )
                    return False

                """
                URL must be without domain and schema (path only).
                Therefore, we specifically add parameters for verification.
                """

                # add schema and domain
                item = 'http://example.com/' + item

                # URL
                if not validators.url(item):
                    return False

        # other zone data validation, except other headers
        elif k in ['args', 'body', 'cookie', 'ua', 'referer', 'url_x', 'args_x', 'body_x', 'cookie_x', 'ua_x', 'referer_x']:

            # split values to list
            # and remove empty items
            vl = list(filter(None, re.split(r'[!&|]', v)))

            # list processing
            for item in vl:

                # base64 decode
                try:

                    # decode
                    item = base64_decode(agent_ip, k, v, item)

                    # decode validation
                    if item is False:
                        return False

                except Exception as e:
                    log.error(
                        'An error occurred while B64 decoding ERL value from {}: {}/{} ({}): {}'
                        .format(agent_ip, k, v, item, e)
                    )
                    return False

        # other headers validation
        elif k in ['other_headers', 'other_headers_x']:

            # list processing
            for item in v:

                # list processing
                for v1 in item.split():

                    # split to k/v
                    v1_k = v1.split(':')[0].strip()
                    v1_v = v1.split(':')[1].strip()

                    # remove '!!' from the start of key
                    v1_k = re.sub(r'^!!', '', v1_k)

                    # init
                    s = v1_k

                    ##
                    # Base64 decode: key
                    ##

                    try:

                        # decode
                        s = base64_decode(agent_ip, k, v, s)

                        # decode validation
                        if s is False:
                            return False

                    except Exception as e:
                        log.error(
                            'An error occurred while B64 decoding ERL value from {}: {}/{} ({}): {}'
                            .format(agent_ip, k, v, s, e)
                        )
                        return False

                    ##
                    # Base64 decode: value
                    ##

                    # split values with "!", "&" and "|"
                    vl2 = list(filter(None, re.split(r'[!&|]', v1_v)))

                    # list processing
                    for v2 in vl2:

                        # init
                        s = v2

                        # base64 decode
                        try:

                            # decode
                            s = base64_decode(agent_ip, k, v, s)

                            # decode validation
                            if s is False:
                                return False

                        except Exception as e:
                            log.error(
                                'An error occurred while B64 decoding ERL value from {}: {}/{} ({}): {}'
                                .format(agent_ip, k, v, s, e)
                            )
                            return False

        # zone size validation
        elif k in ['zone_max_size']:

            # init
            zl =  ['url', 'args', 'body', 'cookie', 'ua', 'referer', 'other_headers']

            # convert all keys to lowercase
            v = {k1.lower(): v1 for k1, v1 in v.items()}

            # key validation
            if any(x for x in v.keys() if x not in zl):
                return False

            # value validation
            if any(x for x in v.values() if (isinstance(x, bool)) or (not isinstance(x, int)) or (x < 0)):
                return False
                                
    except Exception as e:
        log.error(
            'An error occurred while validating ERL key/value from {}: {}/{}: {}'
            .format(agent_ip, k, v, e)
        )
        return False

    return True


def dyn_sync_ban_value_validation(agent_ip, k, v):
    try:

        # dict validation
        if k in dyn_sync_ban_dict:

            # processing empty values
            if not isinstance(v, list):
                return False

            # processing non-empty values
            for item in v:
                if not isinstance(item, dict):
                    return False

        # int validation
        elif k in dyn_sync_ban_int:

            if isinstance(v, str):
                if not v.isdigit():
                    return False
            else:
                return False

    except Exception as e:
        log.error('An error occurred while validating DynSynBan key/value from {}: {}/{}: {}'.format(agent_ip, k, v, e))
        return False

    return True


def nws_target_value_validation(agent_ip, k, v):
    try:

        # active validation
        if k == 'active':
            return True if isinstance(v, bool) else False

        # remove spaces before and after
        if isinstance(v, str):
            v = v.strip()
            if not v:
                return True

        # target validation
        if k == 'target':

            # type validation
            if not isinstance(v, (str, type(None))):
                return False

            # URL validation
            if not validators.url(v):
                return False

            # extract host and port
            s = urlparse(v).netloc.split(':')
            host = s[0]
            port = s[1] if len(s) > 1 else '80'

            # domain validation
            if not validators.domain(host):
                return False

            # port validation
            if (not port.isdigit()) or (1 > int(port) > 65535):
                return False

            # no path validation
            if urlparse(v).path:
                return False

        # Scan proxy validation
        elif k == 'scan_proxy':

            # URL validation
            if not validators.url(v):
                return False

        # Auth. params validation
        elif k == 'auth_params':

            # dict validation
            for i in v.keys():

                if i == 'active':
                    if not isinstance(v[i], bool):
                        return False

                elif i == 'url':

                    # str validation
                    if not isinstance(v[i], str):
                        return False

                    # if path
                    if v[i].startswith('/'):

                        # join with default domain for path validation
                        if not urljoin('http://example.com', v[i]).endswith(v[i]):
                            return False

                    # if not path
                    else:

                        # URL validation
                        if not validators.url(v[i]):
                            return False

                else:

                    # str validation
                    if not isinstance(v[i], str):
                        return False

        # Exclude modules list validation
        elif k == 'exclude_modules':

            # list validation
            if not isinstance(v, list):
                return False

    except Exception as e:
        log.error('An error occurred while validating NWS target key/value from {}: {}/{}: {}'.format(agent_ip, k, v, e))
        return False

    return True


def dyn_settings_idna_normalization(k, v, agent_ip):
    try:

        # init
        result = []

        # skip non-domain items
        if k not in dyn_settings_inda:
            return v

        # with 'domain='
        if k in ['nwaf_limit', 'nwaf_openapi_ip_lm', 'nwaf_openapi_ip_wl', 'nwaf_ip_wl', 'nwaf_ip_lm', 'nwaf_log_request', 'nwaf_antibot_ip_wl']:

            # domain IDNA normalization
            for v1 in v.split(','):

                # init
                r = []
                v1 = v1.strip()
                splt = 'domain='

                # skip empty
                if not v1:
                    continue

                # item processing
                for v2 in v1.split():

                    # init
                    v2 = v2.strip()

                    # skip empty
                    if not v2:
                        continue

                    if v2.startswith(splt):
                        s = v2.split(splt)
                        if len(s) > 1:
                            vhost = idna_normalization(s[1].strip())
                            r.append(splt + vhost)
                        else:
                            r.append(v2)
                    else:
                        r.append(v2)

                # update the result
                result.append(' '.join(r))

        # domain/URL
        else:

            for v1 in v.split(','):

                # skip empty
                if not v1:
                    continue

                # init
                r = []
                splt = '/'
                s = v1.split(splt, 1)

                # idna normalization
                if len(s) > 1:
                    vhost = idna_normalization(s[0])
                    path = splt + s[1].lower()
                    r.append(vhost + path)
                else:
                    vhost = idna_normalization(s[0])
                    r.append(vhost)

                # update the result
                result.append(' '.join(r))

        # result normalization
        result = ', '.join(result)

        # return the result
        return result

    except Exception as e:
        log.error('An error occurred in dyn_settings_idna_normalization for key/value from {}: {}/{}: {}'.format(agent_ip, k, v, e))
        return v


def dyn_bl_idna_normalization(v, agent_ip):
    try:

        # init
        result = []

        # idna normalization
        for v1 in v:

            # init
            r = []

            # skip empty
            if not v1:
                continue

            # item processing
            for v2 in v1.split():

                # init
                splt = 'domain='

                # skip empty
                if not v2:
                    continue

                # idna normalization
                if v2.startswith(splt):
                    s = v2.split(splt)
                    if len(s) > 1:
                        vhost = idna_normalization(s[1].strip())
                        r.append(splt + vhost)
                    else:
                        r.append(v2)
                else:
                    r.append(v2)

            # update the result
            result.append(' '.join(r))

        # return the result
        return result

    except Exception as e:
        log.error('An error occurred in dyn_bl_idna_normalization for key/value from {}: bl/{}: {}'.format(agent_ip, v, e))
        return v


def mlc_settings_idna_normalization(k, v, agent_ip):
    try:

        ##
        # Mix format columns
        ##

        # idna normalization
        if k in mlc_settings_type_lod:

            # list processing
            for idx, item in enumerate(v):

                # dict processing
                for k1, v1 in item.items():

                    # URL processing
                    if k1 == 'url':

                        # skip empty
                        if not v1:
                            continue

                        # init
                        splt = '/'
                        s = v1.split(splt, 1)

                        # idna normalization
                        if len(s) > 1:
                            v[idx][k1] = idna_normalization(s[0]) + splt + s[1].lower()
                        else:
                            v[idx][k1] = idna_normalization(s[0])

            # return the result
            return v

        ##
        # Single format columns
        ##

        # skip non-domain items
        if k not in mlc_settings_idna:
            return v

        # init
        result = []

        # idna normalization
        for v1 in v:

            # skip empty
            if not v1:
                continue

            # init
            splt = '/'
            s = v1.split(splt, 1)

            # idna normalization
            if len(s) > 1:
                result.append(idna_normalization(s[0]) + splt + s[1].lower())
            else:
                result.append(idna_normalization(s[0]))

        # return the result
        return result

    except Exception as e:
        log.error('An error occurred in mlc_settings_idna_normalization for key/value from {}: {}/{}: {}'.format(agent_ip, k, v, e))
        return v


def get_mlc_settings(request, agent_ip, waf_id):
    try:

        # init
        resp_in_json = False

        # extract output format
        try:
            args = request.query_params.get('format').lower()
            if args == 'json':
                resp_in_json = True
        except Exception:
            pass

        # search in the database
        query = """ SELECT to_jsonb(array_agg(mlc_settings)) FROM mlc_settings WHERE (waf_id = %s AND active IS TRUE) """
        connect = init_db_connection()
        cursor = connect.cursor()
        cursor.execute(query, (waf_id,))
        res = cursor.fetchone()[0]
        close_db_connection(connect, cursor)

        # processing result
        if not res:

            # init
            response = {'status': 'none'}

            # format result
            if resp_in_json:
                response = json.dumps(response)
            else:
                response = str(response)

            # response
            return response

        else:

            # extract first item from list
            # (only 1 item must be received from the database)
            response = res[0]
            del res

            # none if no any main cols
            if not [x for x in response.keys() if x not in ['active']]:
                response = {'status': 'none'}
                return response

            # set status
            response['status'] = 'success'

            # del columns
            del response['active']
            del response['modify']
            del response['waf_id']

            # exclude v2 columns
            response = {k: v for k, v in response.items() if k not in mlc_settings_cols_v2}

            # delete keys with None/empty value
            response = {k: v for k, v in response.items() if v not in [None, '', '[]']}

            # v1 data normalization
            for k, v in response.copy().items():
                if k in ['status']:
                    continue
                if k in ['ddos__wl_ip', 'brute__wl_ip']:
                    continue
                elif isinstance(v, str):
                    response[k] = literal_eval(v)

            # idna normalization
            for k, v in response.copy().items():
                response[k] = mlc_settings_idna_normalization(k, v, agent_ip)

            # v1 data normalization
            for k, v in response.copy().items():
                if isinstance(v, list):
                    response[k] = str(v).replace('\'', '"')

            # format result
            if resp_in_json:
                response = json.dumps(response)
            else:
                response = str(response)

            # response
            return response

    except Exception as e:
        log.error('An error occurred in get_mlc_settings: {}'.format(e))


def get_mlc_settings_v2(agent_ip, waf_id):
    try:

        ##
        # Extract data from the database
        ##

        # search in the database
        query = """ SELECT to_jsonb(array_agg(mlc_settings)) FROM mlc_settings WHERE (waf_id = %s) """
        connect = init_db_connection()
        cursor = connect.cursor()
        cursor.execute(query, (waf_id,))
        res = cursor.fetchone()[0]
        close_db_connection(connect, cursor)

        # processing result
        if not res:

            # response
            response = {'status': 'none'}
            return response

        else:

            # extract first item from list
            # (only 1 item must be received from the database)
            response = res[0]
            del res

            # none if no any main cols
            if not [x for x in response.keys() if x not in ['active']]:
                response = {'status': 'none'}
                return response

            # set status
            response['status'] = 'success'

            # del columns
            del response['waf_id']

            # delete keys with None/empty value
            response = {k: v for k, v in response.items() if v not in [None, '', '[]']}

            # modify normalization
            response['modify'] = int(datetime.fromisoformat(response['modify'].split('.')[0]).strftime('%s'))

            # v2 data normalization
            k = 'brute__bf_detect'
            if response.get(k):
                try:
                    response[k] = json.loads(response[k])  # new data format
                except Exception:
                    try:
                        response[k] = literal_eval(response[k])  # prev. data format
                    except Exception as e:
                        response[k] = None
                        log.error(
                            'An error occurred while processing {} in get_mlc_settings_v2: {}, set it again'
                            .format(k, e)
                        )

            # v1 to v2 converter
            for k, v in response.copy().items():
                if k in ['status']:
                    continue
                elif k in ['brute__bf_detect']:
                    continue
                elif k in ['ddos__wl_ip', 'brute__wl_ip'] and response.get(k):
                    response[k] = v.split()
                elif isinstance(v, str):
                    response[k] = literal_eval(v)

            # idna normalization
            for k, v in response.copy().items():
                if response.get(k):
                    response[k] = mlc_settings_idna_normalization(k, v, agent_ip)

            # brute__bf_detect: set zone if it does not exist or null/empty
            for k, v in response.copy().items():
                if k in ['brute__bf_detect'] and v:
                    for idx, i in enumerate(v):
                        if not i.get('zone'):
                            response[k][idx]['zone'] = ['args', 'body']

            # response
            return response

    except Exception as e:
        log.error('An error occurred in get_mlc_settings_v2: {}'.format(e))


def set_mlc_settings(url, data, key, agent_ip, waf_id):
    try:

        ##
        # Check allowed command
        ##

        # init
        cmd = ''

        # search for allowed command in JSON
        for i in settings_cmd:
            if i in data:
                cmd = i.lower()
                break

        # processing result
        if not cmd:

            # logging
            log.error('Required key missing in JSON from {} to {}'.format(agent_ip, url))

            # response
            response = {'status': 'fail', 'description': 'Required key missing'}
            return response

        ##
        # Processing JSON
        ##

        # create extended columns
        cols_ext = mlc_settings_cols + ['active']

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # processing command
        if cmd == 'set':

            # init
            k_list = []
            v_list = []

            # convert all keys to lowercase
            data[cmd] = {k.lower(): v for k, v in data[cmd].items()}

            # get PlanID
            _, tarif_id, _, _ = get_auth(key, agent_ip)

            ##
            # Create row if it does not exist
            ##

            try:

                # insert the data
                query = """ INSERT INTO mlc_settings (waf_id) VALUES (%s) ON CONFLICT DO NOTHING """
                cursor.execute(query, (waf_id,))

            except Exception as e:

                # response
                close_db_connection(connect, cursor)
                response = {'status': 'fail', 'description': str(e)}
                return response

            ##
            # Processing request data
            ##

            for k in cols_ext:

                # processing data
                if k in data[cmd]:

                    # processing items that not allowed for Light
                    if not tarif_id and k in mlc_be_cols:
                        close_db_connection(connect, cursor)
                        err_msg = 'Parameter "{}" is not available for your plan'.format(k)
                        response = {'status': 'fail', 'description': err_msg}
                        return response

                    # value validations
                    if not mlc_value_validation(agent_ip, k, data[cmd][k]):
                        close_db_connection(connect, cursor)
                        response = {'status': 'fail', 'description': 'Incorrect value for {}'.format(k)}
                        return response

                    # replace quotes in list
                    if isinstance(data[cmd][k], list):
                        v = str(data[cmd][k]).replace("'", '"')
                    else:
                        v = data[cmd][k]

                    # convert to str and replace newline/multiple space to space
                    if isinstance(v, str):
                        v = re.sub(r'\s+', ' ', v)

                    # update key/value lists
                    k_list.append(k)
                    v_list.append(v)

            ##
            # Save data to the database
            ##

            # check JSON parameters
            if not k_list:
                close_db_connection(connect, cursor)
                response = {'status': 'fail', 'description': 'Parameters are missing in JSON'}
                return response

            try:

                try:

                    # prepare SQL data
                    k_list.extend(['modify'])
                    v_list.extend([str(datetime.now())])

                    # update data
                    query = """ UPDATE mlc_settings SET ({}) = %s WHERE (waf_id = %s) """.format(','.join(k_list))
                    cursor.execute(query, (tuple(v_list), waf_id,))

                except Exception as e:

                    # logging
                    log.info('An error occurred while processing data for key "{}" in mlc_settings: {}'.format(get_log_key(key), e))

                    # response
                    close_db_connection(connect, cursor)
                    response = {'status': 'fail', 'description': str(e)}
                    return response

                # close database connection
                close_db_connection(connect, cursor)

                # response
                response = {'status': 'success'}
                return response

            except Exception as e:

                # response
                response = {'status': 'fail', 'description': str(e)}
                return response

        # processing command
        elif cmd == 'del':

            ##
            # Delete settings from the database
            ##

            try:

                query = """ DELETE FROM mlc_settings WHERE (waf_id = %s) """
                cursor.execute(query, (waf_id,))

                # close database connection
                close_db_connection(connect, cursor)

                # response
                response = {'status': 'success'}
                return response

            except Exception as e:

                # response
                close_db_connection(connect, cursor)
                response = {'status': 'fail', 'description': str(e)}
                return response

    except Exception as e:
        log.error('An error occurred in set_mlc_settings: {}'.format(e))


def set_mlc_settings_v2(url, data, key, agent_ip, waf_id):
    try:

        ##
        # Check allowed command
        ##

        # init
        cmd = ''

        # search for allowed command in JSON
        for i in settings_cmd:
            if i in data:
                cmd = i.lower()
                break

        # processing result
        if not cmd:

            # logging
            log.error('Required key missing in JSON from {} to {}'.format(agent_ip, url))

            # response
            response = {'status': 'fail', 'description': 'Required key missing'}
            return response

        ##
        # Processing JSON
        ##

        # create extended columns
        cols_ext = mlc_settings_cols_v2 + mlc_settings_cols + ['active']

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # processing command
        if cmd == 'set':

            # init
            k_list = []
            v_list = []

            # convert all keys to lowercase
            data[cmd] = {k.lower(): v for k, v in data[cmd].items()}

            # get PlanID
            _, tarif_id, _, _ = get_auth(key, agent_ip)

            ##
            # Create row if it does not exist
            ##

            try:

                # insert the data
                query = """ INSERT INTO mlc_settings (waf_id) VALUES (%s) ON CONFLICT DO NOTHING """
                cursor.execute(query, (waf_id,))

            except Exception as e:

                # response
                close_db_connection(connect, cursor)
                response = {'status': 'fail', 'description': str(e)}
                return response

            ##
            # Processing request data
            ##

            for k in cols_ext:

                # processing data
                if k in data[cmd]:

                    # processing items that not allowed for Light
                    if not tarif_id and k in mlc_be_cols:

                        # response
                        close_db_connection(connect, cursor)
                        err_msg = 'Parameter "{}" is not available for your plan'.format(k)
                        response = {'status': 'fail', 'description': err_msg}
                        return response

                    # v2 data normalization
                    if k in ['brute__bf_detect']:

                        # convert to lowercase
                        data[cmd][k] = k2lower(data[cmd][k])

                    # v2 to v1 converter
                    if k in ['ddos__wl_ip', 'brute__wl_ip']:

                        # value pre-validations
                        if not isinstance(data[cmd][k], list):
                            close_db_connection(connect, cursor)
                            response = {'status': 'fail', 'description': 'Incorrect value for {}'.format(k)}
                            return response

                        # data normalization to v1 format
                        data[cmd][k] = ' '.join(data[cmd][k])

                    # value validations
                    if not mlc_value_validation(agent_ip, k, data[cmd][k]):
                        close_db_connection(connect, cursor)
                        response = {'status': 'fail', 'description': 'Incorrect value for {}'.format(k)}
                        return response

                    # v2 value normalization
                    if k in ['brute__bf_detect']:
                        v = json.dumps(data[cmd][k])

                    # non-v2 value normalization
                    else:

                        # replace quotes in list
                        if isinstance(data[cmd][k], list):
                            v = str(data[cmd][k]).replace("'", '"')
                        else:
                            v = data[cmd][k]

                        # replace newline/multiple space to space for str
                        if isinstance(v, str):
                            v = re.sub(r'\s+', ' ', v)

                    # update key/value lists
                    k_list.append(k)
                    v_list.append(v)

            # del vars
            del data

            ##
            # Save data to the database
            ##

            # check JSON parameters
            if not k_list:
                close_db_connection(connect, cursor)
                response = {'status': 'fail', 'description': 'Parameters are missing in JSON'}
                return response

            try:

                try:

                    # prepare SQL data
                    k_list.extend(['modify'])
                    v_list.extend([str(datetime.now())])

                    # update data
                    query = """ UPDATE mlc_settings SET ({}) = %s WHERE (waf_id = %s) """.format(','.join(k_list))
                    cursor.execute(query, (tuple(v_list), waf_id,))

                except Exception as e:

                    # logging
                    log.info('An error occurred while processing data for key "{}" in mlc_settings: {}'.format(get_log_key(key), e))

                    # response
                    close_db_connection(connect, cursor)
                    response = {'status': 'fail', 'description': str(e)}
                    return response

                # close database connection
                close_db_connection(connect, cursor)

                # response
                response = {'status': 'success'}
                return response

            except Exception as e:

                # response
                response = {'status': 'fail', 'description': str(e)}
                return response

        # processing command
        elif cmd == 'del':

            ##
            # Delete settings from the database
            ##

            try:

                query = """ DELETE FROM mlc_settings WHERE (waf_id = %s) """
                cursor.execute(query, (waf_id,))

                # close database connection
                close_db_connection(connect, cursor)

                # response
                response = {'status': 'success'}
                return response

            except Exception as e:

                # response
                close_db_connection(connect, cursor)
                response = {'status': 'fail', 'description': str(e)}
                return response

    except Exception as e:
        log.error('An error occurred in set_mlc_settings_v2: {}'.format(e))


def get_dyn_settings(request, agent_ip, waf_id):
    try:

        # init
        resp_in_json = False

        # extract output format
        try:
            args = request.query_params.get('format').lower()
            if args == 'json':
                resp_in_json = True
        except Exception:
            pass

        ##
        # Extract data from the database
        ##

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # search in the database
        query = """ SELECT vhosts_list FROM vhosts_list WHERE (waf_id = %s) """
        cursor.execute(query, (waf_id,))
        res = cursor.fetchone()
        r = [] if not res else literal_eval(res[0])
        r = r + ['www.' + x for x in r if (not x.startswith('*')) and (not x.startswith('.')) and (not x.startswith('www.'))]
        nwaf_mla_host = ', '.join(r)

        # search in the database
        query = """ SELECT to_jsonb(array_agg(dyn_settings)) FROM dyn_settings WHERE (waf_id = %s AND active IS TRUE) """
        cursor.execute(query, (waf_id,))
        res = cursor.fetchone()[0]

        # close database connection
        close_db_connection(connect, cursor)

        # processing result
        if (not res) and (not nwaf_mla_host):

            # init
            response = {'status': 'none'}

            # format result
            if resp_in_json:
                response = json.dumps(response)
            else:
                response = str(response)

            # response
            return response

        else:

            # extract first item from list
            # (only 1 item must be received from the database)
            response = res[0] if res else {}
            del res

            # none if no any main cols
            if (not [x for x in response.keys() if x not in ['active']]) and (not nwaf_mla_host):
                response = {'status': 'none'}
                return json.dumps(response)

            # set status
            response['status'] = 'success'

            # add columns
            response['nwaf_mla_host'] = nwaf_mla_host

            # del columns
            for k in ['active', 'modify', 'waf_id']:
                if k in response:
                    del response[k]

            # delete keys with None/empty value
            response = {k: v for k, v in response.items() if v not in [None, '']}

            # punycode normalization
            for k, v in response.copy().items():
                response[k] = dyn_settings_idna_normalization(k, v, agent_ip)

            # format result
            if resp_in_json:
                response = json.dumps(response)
            else:
                response = str(response)

            # response
            return response

    except Exception as e:
        log.error('An error occurred in get_dyn_settings: {}'.format(e))


def get_dyn_settings_v2(agent_ip, waf_id):
    try:

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # search in the database
        query = """ SELECT vhosts_list FROM vhosts_list WHERE (waf_id = %s) """
        cursor.execute(query, (waf_id,))
        res = cursor.fetchone()
        r = [] if not res else literal_eval(res[0])
        r = r + ['www.' + x for x in r if (not x.startswith('*')) and (not x.startswith('.')) and (not x.startswith('www.'))]
        nwaf_mla_host = ', '.join(r)

        # search in the database
        query = """ SELECT to_jsonb(array_agg(dyn_settings)) FROM dyn_settings WHERE (waf_id = %s) """
        cursor.execute(query, (waf_id,))
        res = cursor.fetchone()[0]

        # close database connection
        close_db_connection(connect, cursor)

        # processing result
        if (not res) and (not nwaf_mla_host):

            # response
            response = {'status': 'none'}
            return response

        else:

            # extract first item from list
            # (only 1 item must be received from the database)
            response = res[0] if res else {}
            del res

            # none if no any main cols
            if (not [x for x in response.keys() if x not in ['active']]) and (not nwaf_mla_host):
                response = {'status': 'none'}
                return response

            # set status
            response['status'] = 'success'

            # add columns
            response['nwaf_mla_host'] = nwaf_mla_host

            # del columns
            for k in ['waf_id']:
                if k in response:
                    del response[k]

            # delete keys with None/empty value
            response = {k: v for k, v in response.items() if v not in [None, '']}

            # punycode normalization
            for k, v in response.copy().items():
                response[k] = dyn_settings_idna_normalization(k, v, agent_ip)

            # modify normalization
            k = 'modify'
            if k in response:
                response[k] = int(datetime.fromisoformat(response[k].split('.')[0]).strftime('%s'))

            # v1 to v2 converter
            for k, v in response.copy().items():
                if k in ['status', 'modify']:
                    continue
                elif isinstance(v, str):
                    response[k] = [x.strip() for x in v.split(',')]

            # response
            return response

    except Exception as e:
        log.error('An error occurred in get_dyn_settings_v2: {}'.format(e))


def set_dyn_settings(url, data, key, agent_ip, waf_id):
    try:

        ##
        # Check allowed command
        ##

        # init
        cmd = ''

        # search for allowed command in JSON
        for i in settings_cmd:
            if i in data:
                cmd = i.lower()
                break

        # processing result
        if not cmd:

            # logging
            log.error('Required key missing in JSON from {} to {}'.format(agent_ip, url))

            # response
            response = {'status': 'fail', 'description': 'Required key missing'}
            return response

        ##
        # Processing JSON
        ##

        # create extended columns
        cols_ext = dyn_settings_cols + ['active']

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # processing command
        if cmd == 'set':

            # init
            k_list = []
            v_list = []

            # convert all keys to lowercase
            data[cmd] = {k.lower(): v for k, v in data[cmd].items()}

            ##
            # Create row if it does not exist
            ##

            try:

                # insert the data
                query = """ INSERT INTO dyn_settings (waf_id) VALUES (%s) ON CONFLICT DO NOTHING """
                cursor.execute(query, (waf_id,))

            except Exception as e:

                # response
                close_db_connection(connect, cursor)
                response = {'status': 'fail', 'description': str(e)}
                return response

            ##
            # Processing request data
            ##

            for k in cols_ext:

                # processing data
                if k in data[cmd]:

                    # convert to str and replace newline/multiple space to space
                    v = data[cmd][k]
                    if not isinstance(v, (bool, type(None))):
                        v = re.sub(r'\s+', ' ', str(v))

                    # data preprocessing according to the type of settings
                    if 'nwaf_b64_decode_disable' == k and re.match(r'\sz:$', v):
                        v = v.replace('z:', 'z:URL|ARGS|HEADERS|BODY')

                    # value validations
                    if not dyn_value_validation(agent_ip, k, v):
                        close_db_connection(connect, cursor)
                        response = {'status': 'fail', 'description': 'Incorrect value for {}'.format(k)}
                        return response

                    # update key/value lists
                    k_list.append(k)
                    v_list.append(v)

            ##
            # Save data to the database
            ##

            # check JSON parameters
            if not k_list:
                close_db_connection(connect, cursor)
                response = {'status': 'fail', 'description': 'Parameters are missing in JSON'}
                return response

            try:

                try:

                    # prepare SQL data
                    k_list.extend(['modify'])
                    v_list.extend([str(datetime.now())])

                    # update data
                    query = """ UPDATE dyn_settings SET ({}) = %s WHERE (waf_id = %s) """.format(','.join(k_list))
                    cursor.execute(query, (tuple(v_list), waf_id,))

                except Exception as e:

                    # logging
                    log.info('An error occurred while processing data for key "{}" in dyn_settings: {}'.format(get_log_key(key), e))

                    # response
                    close_db_connection(connect, cursor)
                    response = {'status': 'fail', 'description': str(e)}
                    return response

                # close database connection
                close_db_connection(connect, cursor)

                # response
                response = {'status': 'success'}
                return response

            except Exception as e:

                # response
                response = {'status': 'fail', 'description': str(e)}
                return response

        # processing command
        elif cmd == 'del':

            ##
            # Delete settings from the database
            ##

            try:

                try:

                    query = """ DELETE FROM dyn_settings WHERE (waf_id = %s) """
                    cursor.execute(query, (waf_id,))

                except Exception as e:

                    # logging
                    log.info('An error occurred while processing data for key "{}" in dyn_settings: {}'.format(get_log_key(key), e))

                    # response
                    close_db_connection(connect, cursor)
                    response = {'status': 'fail', 'description': str(e)}
                    return response

                # close database connection
                close_db_connection(connect, cursor)

                # response
                response = {'status': 'success'}
                return response

            except Exception as e:

                # response
                response = {'status': 'fail', 'description': str(e)}
                return response

    except Exception as e:
        log.error('An error occurred in set_dyn_settings: {}'.format(e))


def set_dyn_settings_v2(url, data, key, agent_ip, waf_id):
    try:

        ##
        # Check allowed command
        ##

        # init
        cmd = ''

        # search for allowed command in JSON
        for i in settings_cmd:
            if i in data:
                cmd = i.lower()
                break

        # processing result
        if not cmd:

            # logging
            log.error('Required key missing in JSON from {} to {}'.format(agent_ip, url))

            # response
            response = {'status': 'fail', 'description': 'Required key missing'}
            return response

        ##
        # Processing JSON
        ##

        # create extended columns
        cols_ext = dyn_settings_cols + ['active']

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # processing command
        if cmd == 'set':

            # init
            k_list = []
            v_list = []

            # convert all keys to lowercase
            data[cmd] = {k.lower(): v for k, v in data[cmd].items()}

            ##
            # Create row if it does not exist
            ##

            try:

                # insert the data
                query = """ INSERT INTO dyn_settings (waf_id) VALUES (%s) ON CONFLICT DO NOTHING """
                cursor.execute(query, (waf_id,))

            except Exception as e:

                # response
                close_db_connection(connect, cursor)
                response = {'status': 'fail', 'description': str(e)}
                return response

            ##
            # Processing request data
            ##

            for k in cols_ext:

                # processing data
                if k in data[cmd]:

                    # v2 to v1 converter
                    if k in dyn_settings_cols:

                        # value pre-validations
                        if isinstance(data[cmd][k], str):
                            close_db_connection(connect, cursor)
                            response = {'status': 'fail', 'description': 'Incorrect value for {}'.format(k)}
                            return response

                        # data normalization to v1 format
                        if (k not in dyn_settings_type_int) and (isinstance(data[cmd][k], (list, dict, tuple))):
                            data[cmd][k] = ', '.join(data[cmd][k])

                    # convert to str and replace newline/multiple space to space
                    v = data[cmd][k]
                    if not isinstance(v, (int, bool, type(None))):
                        v = re.sub(r'\s+', ' ', str(v))

                    # data preprocessing according to the type of settings
                    if k == 'nwaf_b64_decode_disable':
                        if re.match(r'\sz:$', str(v)):
                            v = str(v).replace('z:', 'z:URL|ARGS|HEADERS|BODY')

                    # value validations
                    if not dyn_value_validation(agent_ip, k, v):
                        close_db_connection(connect, cursor)
                        response = {'status': 'fail', 'description': 'Incorrect value for {}'.format(k)}
                        return response

                    # update key/value lists
                    k_list.append(k)
                    v_list.append(v)

            ##
            # Save data to the database
            ##

            # check JSON parameters
            if not k_list:
                close_db_connection(connect, cursor)
                response = {'status': 'fail', 'description': 'Parameters are missing in JSON'}
                return response

            try:

                try:

                    # prepare SQL data
                    k_list.extend(['modify'])
                    v_list.extend([str(datetime.now())])

                    # update data
                    query = """ UPDATE dyn_settings SET ({}) = %s WHERE (waf_id = %s) """.format(','.join(k_list))
                    cursor.execute(query, (tuple(v_list), waf_id,))

                except Exception as e:

                    # logging
                    log.info('An error occurred while processing data for key "{}" in dyn_settings: {}'.format(get_log_key(key), e))

                    # response
                    close_db_connection(connect, cursor)
                    response = {'status': 'fail', 'description': str(e)}
                    return response

                # close database connection
                close_db_connection(connect, cursor)

                # response
                response = {'status': 'success'}
                return response

            except Exception as e:

                # response
                response = {'status': 'fail', 'description': str(e)}
                return response

        # processing command
        elif cmd == 'del':

            ##
            # Delete settings from the database
            ##

            try:

                try:

                    query = """ DELETE FROM dyn_settings WHERE (waf_id = %s) """
                    cursor.execute(query, (waf_id,))

                except Exception as e:

                    # logging
                    log.info('An error occurred while processing data for key "{}" in dyn_settings: {}'.format(get_log_key(key), e))

                    # response
                    close_db_connection(connect, cursor)
                    response = {'status': 'fail', 'description': str(e)}
                    return response

                # close database connection
                close_db_connection(connect, cursor)

                # response
                response = {'status': 'success'}
                return response

            except Exception as e:

                # response
                response = {'status': 'fail', 'description': str(e)}
                return response

    except Exception as e:
        log.error('An error occurred in set_dyn_settings_v2: {}'.format(e))


def get_dyn_erl(request, waf_id):
    try:

        # init
        get_extended = False
        resp_in_json = False

        # extract output format
        try:
            args = request.query_params.get('extended').lower()
            if args == 'yes':
                get_extended = True
        except Exception:
            pass

        # extract output format
        try:
            args = request.query_params.get('format').lower()
            if args == 'json':
                resp_in_json = True
        except Exception:
            pass

        ##
        # Extract data from the database
        ##

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # search in the database
        query = """ SELECT to_jsonb(array_agg(dyn_erl)) from dyn_erl WHERE (waf_id = %s AND active IS TRUE) """
        if get_extended or resp_in_json:
            query = """ SELECT to_jsonb(array_agg(dyn_erl)) from dyn_erl WHERE (waf_id = %s) """
        cursor.execute(query, (waf_id,))
        res = cursor.fetchall()

        # close database connection
        close_db_connection(connect, cursor)

        ##
        # Response in JSON format
        ##

        # processing result
        if not res or not res[0][0]:

            # init
            response = {'status': 'none'}

            # format result
            if resp_in_json:
                response = json.dumps(response)
            else:
                response = str(response)

            # response
            return response

        else:

            # init
            response = {'status': 'success'}

            # response in JSON
            if resp_in_json:

                # init
                r = []

                # processing columns
                for row in res[0][0]:

                    # delete v2 columns
                    for k in dyn_erl_type_data_v2_cols:
                        del row[k]

                    # add rid key
                    row['rid'] = row['id']

                    # idna normalization
                    if row['domain']:
                        dl = []
                        for x in re.split('([!|&])', row['domain']):
                            if not x:
                                continue
                            elif x in ['!', '|', '&']:
                                dl.append(x)
                            else:
                                dl.append(
                                    base64.b64encode(
                                        idna_normalization(
                                            base64.b64decode(
                                                x.encode('UTF-8')
                                            ).decode('UTF-8')
                                        ).encode('UTF-8')
                                    ).decode('UTF-8')
                                )
                        row['domain'] = ''.join(dl)
                        del dl

                    # remove unused
                    del row['waf_id']

                    # update the list
                    r.append(row)

                # update response
                response['erl'] = r
                del r

                # convert to JSON
                response = json.dumps(response)

            else:

                # init
                response = str(response)

                # processing each row
                for row in res[0][0]:
                    try:

                        # idna normalization
                        if row['domain']:
                            dl = []
                            for x in re.split('([!|&])', row['domain']):
                                if not x:
                                    continue
                                elif x in ['!', '|', '&']:
                                    dl.append(x)
                                else:
                                    dl.append(
                                        base64.b64encode(
                                            idna_normalization(
                                                base64.b64decode(
                                                    x.encode('UTF-8')
                                                ).decode('UTF-8')
                                            ).encode('UTF-8')
                                        ).decode('UTF-8')
                                    )
                            row['domain'] = ''.join(dl)
                            del dl

                        # delete unused columns
                        for k in row.copy().keys():

                            # if is get_extended
                            if get_extended:
                                if k == 'active':
                                    # don't delete key
                                    continue

                            # other delete conditions
                            if k != 'id' and k not in dyn_erl_cols:
                                del row[k]

                        # update empty columns
                        row = {k: ('' if v is None else v) for k, v in row.items()}

                        ##
                        # update other_headers
                        ##

                        # init
                        other_headers = []

                        # update other_headers
                        if not row['other_headers']:
                            row['other_headers'] = other_headers
                        else:
                            row['other_headers'] = row['other_headers'].split(', ')

                        if not get_extended:

                            # update values for dyn
                            row['id'] = str(row['id'])
                            row['lm'] = str(row['lm']).lower()
                            row['api'] = str(row['api']).lower()
                            row['noban'] = str(row['noban']).lower()
                            row['no_cookie'] = str(row['no_cookie']).lower()
                            row['no_ua'] = str(row['no_ua']).lower()
                            row['no_referer'] = str(row['no_referer']).lower()
                            row['block_tor'] = str(row['block_tor']).lower()
                            row['block_proxy'] = str(row['block_proxy']).lower()

                        # update result
                        response = response + '\n' + json.dumps(row)

                    except Exception as e:
                        log.error('An error occurred while processing result from table "dyn_erl": {}'.format(e))

            # response
            return response

    except Exception as e:
        log.error('An error occurred in get_dyn_erl: {}'.format(e))


def get_dyn_erl_v2(waf_id):
    try:

        ##
        # Extract data from the database
        ##

        # search in the database
        query = """ SELECT to_jsonb(array_agg(dyn_erl)) from dyn_erl WHERE (waf_id = %s) """
        connect = init_db_connection()
        cursor = connect.cursor()
        cursor.execute(query, (waf_id,))
        res = cursor.fetchall()
        close_db_connection(connect, cursor)

        ##
        # Response in JSON format
        ##

        # processing result
        if not res or not res[0][0]:

            # init
            response = {'status': 'none'}

            # response
            return response

        else:

            # init
            response = {'status': 'success', 'erl': []}

            # processing columns
            for row in res[0][0]:

                # init
                zsz = ['zone_max_size']
                ohl = ['other_headers', 'other_headers_x']

                # remove unused
                del row['waf_id']

                # modify normalization
                row['modify'] = int(datetime.fromisoformat(row['modify'].split('.')[0]).strftime('%s'))

                # domain: idna normalization and convert to list
                if row['domain']:
                    dl = []
                    for x in re.split('([!|&])', row['domain']):
                        if not x:
                            continue
                        elif x in ['!', '|', '&']:
                            dl.append(x)
                        else:
                            dl.append(
                                base64.b64encode(
                                    idna_normalization(
                                        base64.b64decode(
                                            x.encode('UTF-8')
                                        ).decode('UTF-8')
                                    ).encode('UTF-8')
                                ).decode('UTF-8')
                            )
                    row['domain'] = ''.join(dl)
                    del dl

                # convert value to list for other zone
                for k in ohl:
                    row[k] = [x.strip() for x in row[k].split(',')] if row[k] else []

                # zone size normalization
                for k in zsz:
                    row[k] = json.loads(row[k]) if row.get(k) else {}

                # convert none to str
                for k in dyn_erl_type_data_v2:

                    # skip processed
                    if k not in ohl + zsz:
                        row[k] = '' if not row[k] else row[k]

                # update the list
                response['erl'].append(row)

            # response
            return response

    except Exception as e:
        log.error('An error occurred in get_dyn_erl_v2: {}'.format(e))


def set_dyn_erl(url, data, agent_ip, waf_id):
    try:

        ##
        # Check allowed command
        ##

        # init
        cmd = ''

        # search for allowed command in JSON
        for i in dyn_wl_erl_cmd:
            if i in data:
                cmd = i.lower()
                break

        # processing result
        if not cmd:

            # logging
            log.error('Required key missing in JSON from {} to {}'.format(agent_ip, url))

            # response
            response = {'status': 'fail', 'description': 'Required key missing'}
            return response

        ##
        # Processing JSON
        ##

        # init
        t = 'dyn_erl_ts'
        k_list = []
        v_list = []

        # create extended columns
        cols_ext = dyn_erl_cols + ['active']

        # processing command
        if cmd == 'add' or cmd == 'upd':

            # convert all keys to lowercase
            data[cmd] = {k.lower(): v for k, v in data[cmd].items()}

            ##
            # Checking data compatibility
            ##

            # no_cookie and cookie in data
            if 'no_cookie' in data[cmd] and 'cookie' in data[cmd]:

                # cannot be used no_cookie:true with non-empty cookie
                if data[cmd]['no_cookie'] and len(data[cmd]['cookie']) > 0:
                    err_msg = 'Cannot be used at the same time: "no_cookie" with true and non-empty data in "cookie"'
                    response = {'status': 'fail', 'description': err_msg}
                    return response

            # no_ua and ua in data
            if 'no_ua' in data[cmd] and 'ua' in data[cmd]:

                # cannot be used no_ua:true with non-empty ua
                if data[cmd]['no_ua'] and len(data[cmd]['ua']) > 0:
                    err_msg = 'Cannot be used at the same time: "no_ua" with true and non-empty data in "ua"'
                    response = {'status': 'fail', 'description': err_msg}
                    return response

            # no_referer and referer in data
            if 'no_referer' in data[cmd] and 'referer' in data[cmd]:

                # cannot be used no_referer:true with non-empty referer
                if data[cmd]['no_referer'] and len(data[cmd]['referer']) > 0:
                    err_msg = 'Cannot be used at the same time: "no_referer" with true and non-empty data in "referer"'
                    response = {'status': 'fail', 'description': err_msg}
                    return response

            # 'add' only
            if cmd == 'add':

                # checking for at least 1 required non-empty key
                if not {k.lower(): v for k, v in data[cmd].items() if k in dyn_erl_type_required and v}:
                    err_msg = 'At least 1 required key is missing: {}'.format(', '.join(dyn_erl_type_required))
                    response = {'status': 'fail', 'description': err_msg}
                    return response

            # processing data
            for k in cols_ext:
                try:

                    if k in data[cmd]:

                        # init
                        v = data[cmd][k]

                        # using & with | or multiple use
                        for pattern in ['|&', '&|', '||', '&&']:
                            if pattern in str(v):
                                err_msg = 'Unknown operator "{}" for '.format(pattern, k)
                                response = {'status': 'fail', 'description': err_msg}
                                return response

                        # other_headers
                        if k == 'other_headers':

                            # type validation
                            if (not isinstance(v, list)) or any(x for x in v if (not x) or (not isinstance(x, str))):
                                err_msg = 'Incorrect value or value data for {}'.format(k)
                                response = {'status': 'fail', 'description': err_msg}
                                return response

                            # processing list
                            for item in v:

                                # init
                                splt = ':'

                                # type validation
                                if splt not in item:
                                    err_msg = 'Incorrect value data for {}: {}'.format(k, item)
                                    response = {'status': 'fail', 'description': err_msg}
                                    return response

                                # format validation
                                if len(item.split(splt)) != 2:
                                    err_msg = 'Incorrect value data for {}: {}'.format(k, item)
                                    response = {'status': 'fail', 'description': err_msg}
                                    return response

                                # split to k/v and extract value
                                ik = item.split(splt)[0].strip()
                                iv = item.split(splt)[1].strip()

                                ##
                                #  Key validation
                                ##

                                # remove !! from the start before processing
                                if ik.startswith('!!'):
                                    ik = ik[2:]

                                # processing key
                                for pattern in ['!', '&', '|']:
                                    if pattern in ik:
                                        err_msg = 'The headers name cannot contain the operator "{}" in {}'.format(pattern, ik)
                                        response = {'status': 'fail', 'description': err_msg}
                                        return response

                                ##
                                # Value validation
                                ##

                                # validation for special characters at the beginning
                                for pattern in ['&!', '|!']:
                                    if iv.startswith(pattern):
                                        err_msg = 'The headers value cannot start with the operator "{}" in {}'.format(pattern, iv)
                                        response = {'status': 'fail', 'description': err_msg}
                                        return response

                                # validation for special characters at the end
                                for pattern in ['!', '&', '|']:
                                    if iv.endswith(pattern):
                                        err_msg = 'The headers value cannot end with the operator "{}" in {}'.format(pattern, iv)
                                        response = {'status': 'fail', 'description': err_msg}
                                        return response

                        # other non-boolean keys
                        elif (k != 'other_headers') and (k in dyn_erl_type_data):

                            # validation for special characters at the end
                            for pattern in ['!', '&', '|']:
                                if v.endswith(pattern):
                                    err_msg = 'The value cannot end with the operator "{}" in {}'.format(pattern, v)
                                    response = {'status': 'fail', 'description': err_msg}
                                    return response

                            # remove & or | from the beginning
                            for pattern in ['&', '|']:
                                if v.startswith(pattern):
                                    v = v[1:]

                        # value validations
                        if not dyn_erl_value_validation(agent_ip, k, v):
                            response = {'status': 'fail', 'description': 'Incorrect value for {}'.format(k)}
                            return response

                        # other_headers normalization
                        if (k == 'other_headers') and (isinstance(data[cmd][k], list)):
                            v = ', '.join(data[cmd][k])

                        # replace newline/multiple space to space
                        if not isinstance(v, (bool, type(None))):
                            v = re.sub(r'\s+', ' ', str(v))

                        # update key/value lists
                        k_list.append(k)
                        v_list.append(v)

                except Exception as e:
                    log.error('An error occurred in {} while processing data: {}'.format(url, e))
                    return

            # check JSON parameters
            if not k_list:
                response = {'status': 'fail', 'description': 'Parameters are missing in JSON'}
                return response

        ##
        # Processing the command
        ##

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        if cmd == 'add':
            try:

                # prepare SQL data
                k_list.extend(['waf_id', 'modify'])
                v_list.extend([waf_id, str(datetime.now())])

                # insert the data
                query = """ INSERT INTO dyn_erl ({}) VALUES %s RETURNING id """.format(','.join(k_list))
                cursor.execute(query, (tuple(v_list),))
                res = cursor.fetchone()

                # close database connection
                close_db_connection(connect, cursor)

                # update ts
                set_hash(agent_ip, url, t, waf_id)

                # set response
                if res:
                    # if ID was returned
                    response = {'status': 'success', 'rid': res[0]}
                else:
                    # if ID was not returned
                    response = {'status': 'fail', 'description': 'ID was not returned during the data saving process'}

                # response
                return response

            except Exception as e:

                # set error message
                err_msg = str(e)

                # processing err_msg
                if 'duplicate key value' in err_msg:
                    err_msg = 'Duplicate key value'

                # response
                close_db_connection(connect, cursor)
                response = {'status': 'fail', 'description': err_msg}
                return response

        elif cmd == 'upd':
            try:

                """
                Rules are displayed and could be edited for other members with the same WAF ID.
                """

                ##
                # Processing request data
                ##

                # extract value
                param = 'rid'
                if data[cmd].get(param):
                    try:
                        rid = int(data[cmd][param])
                    except Exception:
                        close_db_connection(connect, cursor)
                        response = {'status': 'fail', 'description': 'Parameter "rid" is not an integer'}
                        return response

                # if no ID in the request
                else:

                    # response
                    close_db_connection(connect, cursor)
                    err_msg = 'No ID received'
                    response = {'status': 'fail', 'description': err_msg}
                    return response

                ##
                # Search for rule ID (to update the hash in the future)
                ##

                query = """ SELECT id from dyn_erl WHERE (waf_id = %s AND id = %s) """
                cursor.execute(query, (waf_id, rid,))
                res = cursor.fetchone()

                # check result
                if not res:

                    """
                    Row is not found, return response
                    """

                    # response
                    close_db_connection(connect, cursor)
                    err_msg = 'ID {} not found in database'.format(rid)
                    response = {'status': 'fail', 'description': err_msg}
                    return response

                ##
                # Checking data compatibility:
                # at least 1 key from dyn_erl_type_data will be in the rule
                ##

                # if at least 1 key from dyn_erl_type_data in the data[cmd]
                if [i for i in dyn_erl_type_data if i in data[cmd].keys()]:

                    query = """ SELECT {} from dyn_erl WHERE (waf_id = %s AND id = %s) """.format(','.join(dyn_erl_type_data))
                    cursor.execute(query, (waf_id, rid,))
                    res = cursor.fetchone()

                    # check result
                    if res:

                        # init
                        res_dict = {}

                        # create PG response dict
                        for i, k in enumerate(dyn_erl_type_data):
                            res_dict[k] = res[i]

                        # remove unused columns
                        for k in res_dict.copy().keys():
                            if k not in dyn_erl_type_data:
                                res_dict.pop(k)

                        # merge result from the database with data from the request
                        for k in res_dict.copy().keys():
                            if k in data[cmd]:
                                res_dict[k] = data[cmd][k]

                        # delete keys with None/empty value
                        res_dict = {k: v for k, v in res_dict.items() if v not in [None, '', '[]']}

                        # If res_dict do not contain at least 1 non-empty key
                        if not len(res_dict):
                            close_db_connection(connect, cursor)
                            err_msg = 'After updating, the rule must contain at least 1 required key: {}'.format(', '.join(dyn_erl_type_data))
                            response = {'status': 'fail', 'description': err_msg}
                            return response

                    else:

                        """
                        Row is not found, return response
                        """

                        # response
                        close_db_connection(connect, cursor)
                        err_msg = 'ID {} not found in database'.format(rid)
                        response = {'status': 'fail', 'description': err_msg}
                        return response

                ##
                # Update data
                ##

                # prepare SQL data
                k_list.extend(['modify'])
                v_list.extend([str(datetime.now())])

                # insert the data
                query = """ UPDATE dyn_erl SET ({}) = %s WHERE (waf_id = %s AND id = %s) """.format(','.join(k_list))
                cursor.execute(query, (tuple(v_list), waf_id, rid,))

                # update ts
                set_hash(agent_ip, url, t, waf_id)

                # close database connection
                close_db_connection(connect, cursor)

                # response
                response = {'status': 'success'}
                return response

            except Exception as e:

                # response
                response = {'status': 'fail', 'description': str(e)}
                return response

        elif cmd == 'del':
            try:

                # convert all keys to lowercase
                data[cmd] = {k.lower(): v for k, v in data[cmd].items()}

                # extract value
                param = 'rid'
                if data[cmd].get(param):
                    try:
                        rid = int(data[cmd][param])
                    except Exception:
                        close_db_connection(connect, cursor)
                        response = {'status': 'fail', 'description': 'Parameter "rid" is not an integer'}
                        return response

                # if no ID in the request
                else:

                    # response
                    close_db_connection(connect, cursor)
                    err_msg = 'No ID received'
                    response = {'status': 'fail', 'description': err_msg}
                    return response

                try:

                    query = """ SELECT id from dyn_erl WHERE (waf_id = %s AND id = %s) """
                    cursor.execute(query, (waf_id, rid,))
                    res = cursor.fetchone()

                    # check result
                    if not res:

                        """
                        Row is not found, return response
                        """

                        # response
                        close_db_connection(connect, cursor)
                        err_msg = 'ID {} not found in database'.format(rid)
                        response = {'status': 'fail', 'description': err_msg}
                        return response

                    # search data
                    query = """ DELETE from dyn_erl WHERE (id = %s) AND (waf_id = %s) """
                    cursor.execute(query, (rid, waf_id,))

                    # update ts
                    set_hash(agent_ip, url, t, waf_id)

                    # close database connection
                    close_db_connection(connect, cursor)

                    # response
                    response = {'status': 'success'}
                    return response

                except Exception as e:

                    # response
                    close_db_connection(connect, cursor)
                    response = {'status': 'fail', 'description': str(e)}
                    return response

            except Exception as e:

                # response
                close_db_connection(connect, cursor)
                response = {'status': 'fail', 'description': str(e)}
                return response

    except Exception as e:
        log.error('An error occurred in set_dyn_erl: {}'.format(e))


def set_dyn_erl_v2(url, data, agent_ip, waf_id):
    try:

        ##
        # Check allowed command
        ##

        # init
        cmd = ''

        # search for allowed command in JSON
        for i in dyn_wl_erl_cmd:
            if i in data:
                cmd = i.lower()
                break

        # processing result
        if not cmd:

            # logging
            log.error('Required key missing in JSON from {} to {}'.format(agent_ip, url))

            # response
            response = {'status': 'fail', 'description': 'Required key missing'}
            return response

        ##
        # Processing JSON
        ##

        # init
        t = 'dyn_erl_ts'
        k_list = []
        v_list = []

        # create extended columns
        cols_ext = dyn_erl_cols_v2 + ['active']

        # processing command
        if cmd == 'add' or cmd == 'upd':

            # convert all keys to lowercase
            data[cmd] = {k.lower(): v for k, v in data[cmd].items()}

            ##
            # Checking data compatibility
            ##

            # no_cookie and cookie in data
            if 'no_cookie' in data[cmd] and 'cookie' in data[cmd]:

                # cannot be used no_cookie:true with non-empty cookie
                if data[cmd]['no_cookie'] and len(data[cmd]['cookie']) > 0:
                    err_msg = 'Cannot be used at the same time: "no_cookie" with true and non-empty data in "cookie"'
                    response = {'status': 'fail', 'description': err_msg}
                    return response

            # no_ua and ua in data
            if 'no_ua' in data[cmd] and 'ua' in data[cmd]:

                # cannot be used no_ua:true with non-empty ua
                if data[cmd]['no_ua'] and len(data[cmd]['ua']) > 0:
                    err_msg = 'Cannot be used at the same time: "no_ua" with true and non-empty data in "ua"'
                    response = {'status': 'fail', 'description': err_msg}
                    return response

            # no_referer and referer in data
            if 'no_referer' in data[cmd] and 'referer' in data[cmd]:

                # cannot be used no_referer:true with non-empty referer
                if data[cmd]['no_referer'] and len(data[cmd]['referer']) > 0:
                    err_msg = 'Cannot be used at the same time: "no_referer" with true and non-empty data in "referer"'
                    response = {'status': 'fail', 'description': err_msg}
                    return response

            # 'add' only
            if cmd == 'add':

                # checking for at least 1 required non-empty key
                if not {k.lower(): v for k, v in data[cmd].items() if k in dyn_erl_type_required_v2 and v}:
                    err_msg = 'At least 1 required key is missing: {}'.format(', '.join(dyn_erl_type_required_v2))
                    response = {'status': 'fail', 'description': err_msg}
                    return response

            # processing data
            for k in cols_ext:
                try:

                    if k in data[cmd]:

                        # init
                        v = data[cmd][k]
                        zsz = ['zone_max_size']
                        ohl = ['other_headers', 'other_headers_x']

                        # all non-boolean column type validation (except specific)
                        if k in dyn_erl_type_data_v2 and (k not in ohl+zsz):

                            # type validation
                            if (v is not None) and (not isinstance(v, str)):
                                err_msg = 'Incorrect value or value data for {}'.format(k)
                                response = {'status': 'fail', 'description': err_msg}
                                return response

                            # using & with | or multiple use
                            for pattern in ['|&', '&|', '||', '&&']:
                                if pattern in str(v):
                                    err_msg = 'Unknown operator "{}" for '.format(pattern, k)
                                    response = {'status': 'fail', 'description': err_msg}
                                    return response

                            # validation for special characters at the end
                            for pattern in ['!', '&', '|']:
                                if v.endswith(pattern):
                                    err_msg = 'The value cannot end with the operator "{}" in {}'.format(pattern, v)
                                    response = {'status': 'fail', 'description': err_msg}
                                    return response

                        # other headers validation
                        if k in ohl:

                            # type validation
                            if (not isinstance(v, list)) or any(x for x in v if (not x) or (not isinstance(x, str))):
                                err_msg = 'Incorrect value or value data for {}'.format(k)
                                response = {'status': 'fail', 'description': err_msg}
                                return response

                            # items type validation
                            if any(x for x in v if (not x) or (not isinstance(x, str))):
                                err_msg = 'Incorrect value or value data for {}'.format(k)
                                response = {'status': 'fail', 'description': err_msg}
                                return response

                            # processing list
                            for item in v:

                                # init
                                splt = ':'

                                # type validation
                                if splt not in item:
                                    err_msg = 'Incorrect value data for {}: {}'.format(k, item)
                                    response = {'status': 'fail', 'description': err_msg}
                                    return response

                                # format validation
                                if len(item.split(splt)) != 2:
                                    err_msg = 'Incorrect value data for {}: {}'.format(k, item)
                                    response = {'status': 'fail', 'description': err_msg}
                                    return response

                                # split to k/v and extract value
                                ik = item.split(splt)[0].strip()
                                iv = item.split(splt)[1].strip()

                                ##
                                # Key validation
                                ##

                                if ik.startswith('!!'):
                                    ik = ik[2:]

                                # processing key
                                for pattern in ['!', '&', '|']:
                                    if pattern in ik:
                                        err_msg = 'The headers name cannot contain the operator "{}" in {}'.format(pattern, ik)
                                        response = {'status': 'fail', 'description': err_msg}
                                        return response

                                ##
                                # Value validation
                                ##

                                # validation for special characters at the beginning
                                for pattern in ['&!', '|!']:
                                    if iv.startswith(pattern):
                                        err_msg = 'The headers value cannot start with the operator "{}" in {}'.format(pattern, iv)
                                        response = {'status': 'fail', 'description': err_msg}
                                        return response

                                # validation for special characters at the end
                                for pattern in ['!', '&', '|']:
                                    if iv.endswith(pattern):
                                        err_msg = 'The headers value cannot end with the operator "{}" in {}'.format(pattern, iv)
                                        response = {'status': 'fail', 'description': err_msg}
                                        return response

                        # country/domain normalization
                        if k in ['country', 'domain']:

                            # convert to list
                            vl = re.split(r'([!|&]+)', v)

                            # encode the items
                            for idx, item in enumerate(vl):

                                # skip seperator
                                if any(x for x in ['!', '|', '&'] if x in item):
                                    continue

                                # encode the item
                                vl[idx] = base64.b64encode(item.encode('UTF-8')).decode('UTF-8')

                            # update the value
                            v = ''.join(vl)

                            # del vars
                            del vl

                        # zone size normalization
                        if k in zsz:

                            # type validation
                            if (v is not None) and (not isinstance(v, dict)):
                                err_msg = 'Incorrect value data for {}'.format(k)
                                response = {'status': 'fail', 'description': err_msg}
                                return response

                            # empty value processing
                            v = {} if not v else v


                        # other data normalization
                        if (k not in ohl + zsz) and (k in dyn_erl_type_data_v2) and (v is not None):

                            # convert to str
                            v = ''.join(v)

                            # remove & or | from the beginning
                            for pattern in ['&', '|']:
                                if v.startswith(pattern):
                                    v = v[1:]

                        # value validations
                        if not dyn_erl_value_validation(agent_ip, k, v):
                            response = {'status': 'fail', 'description': 'Incorrect value for {}'.format(k)}
                            return response

                        # other headers normalization
                        if (k in ohl) and (isinstance(v, list)):
                            v = ', '.join(v)

                        # zone size normalization
                        if k in zsz:
                            v = json.dumps(v)

                        # replace newline/multiple space to space
                        if not isinstance(v, (bool, type(None))):
                            v = re.sub(r'\s+', ' ', str(v))

                        # update key/value lists
                        k_list.append(k)
                        v_list.append(v)

                except Exception as e:
                    log.error('An error occurred in {} while processing data: {}'.format(url, e))
                    return

            # check JSON parameters
            if not k_list:
                response = {'status': 'fail', 'description': 'Parameters are missing in JSON'}
                return response

        ##
        # Processing the command
        ##

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        if cmd == 'add':
            try:

                # prepare SQL data
                k_list.extend(['waf_id', 'modify'])
                v_list.extend([waf_id, str(datetime.now())])

                # insert the data
                query = """ INSERT INTO dyn_erl ({}) VALUES %s RETURNING id """.format(','.join(k_list))
                cursor.execute(query, (tuple(v_list),))
                res = cursor.fetchone()

                # update ts
                set_hash(agent_ip, url, t, waf_id)

                # set response
                if res:
                    # if ID was returned
                    response = {'status': 'success', 'id': res[0]}
                else:
                    # if ID was not returned
                    response = {'status': 'fail', 'description': 'ID was not returned during the data saving process'}

                # close database connection
                close_db_connection(connect, cursor)

                # response
                return response

            except Exception as e:

                # set error message
                err_msg = str(e)

                # processing err_msg
                if 'duplicate key value' in err_msg:
                    err_msg = 'Duplicate key value'

                # response
                close_db_connection(connect, cursor)
                response = {'status': 'fail', 'description': err_msg}
                return response

        elif cmd == 'upd':
            try:

                """
                Rules are displayed and could be edited for other members with the same WAF ID.
                """

                ##
                # Processing request data
                ##

                # extract value
                param = 'id'
                if data[cmd].get(param):
                    try:
                        rid = int(data[cmd][param])
                    except Exception:
                        close_db_connection(connect, cursor)
                        response = {'status': 'fail', 'description': 'Parameter "id" is not an integer'}
                        return response

                # if no ID in the request
                else:

                    # response
                    close_db_connection(connect, cursor)
                    err_msg = 'No ID received'
                    response = {'status': 'fail', 'description': err_msg}
                    return response

                ##
                # Search for rule ID (to update the hash in the future)
                ##

                query = """ SELECT id from dyn_erl WHERE (waf_id = %s AND id = %s) """
                cursor.execute(query, (waf_id, rid,))
                res = cursor.fetchone()

                # check result
                if not res:

                    """
                    Row is not found, return response
                    """

                    # response
                    close_db_connection(connect, cursor)
                    err_msg = 'ID {} not found in database'.format(rid)
                    response = {'status': 'fail', 'description': err_msg}
                    return response

                ##
                # Checking data compatibility:
                # at least 1 key from dyn_erl_type_data_v2 will be in the rule
                ##

                # if at least 1 key from dyn_erl_type_data_v2 in the data[cmd]
                if [i for i in dyn_erl_type_data_v2 if i in data[cmd].keys()]:

                    query = """ SELECT {} from dyn_erl WHERE (waf_id = %s AND id = %s) """.format(','.join(dyn_erl_type_data_v2))
                    cursor.execute(query, (waf_id, rid,))
                    res = cursor.fetchone()

                    # check result
                    if res:

                        # init
                        res_dict = {}

                        # create PG response dict
                        for i, k in enumerate(dyn_erl_type_data_v2):
                            res_dict[k] = res[i]

                        # remove unused columns
                        for k in res_dict.copy().keys():
                            if k not in dyn_erl_type_data_v2:
                                res_dict.pop(k)

                        # merge result from the database with data from the request
                        for k in res_dict.copy().keys():
                            if k in data[cmd]:
                                res_dict[k] = data[cmd][k]

                        # delete keys with None/empty value
                        res_dict = {k: v for k, v in res_dict.items() if v not in [None, '', '[]']}

                        # If res_dict do not contain at least 1 non-empty key
                        if not len(res_dict):
                            close_db_connection(connect, cursor)
                            err_msg = 'After updating, the rule must contain at least 1 required key: {}'.format(', '.join(dyn_erl_type_data_v2))
                            response = {'status': 'fail', 'description': err_msg}
                            return response

                    else:

                        """
                        Row is not found, return response
                        """

                        # response
                        close_db_connection(connect, cursor)
                        err_msg = 'ID {} not found in database'.format(rid)
                        response = {'status': 'fail', 'description': err_msg}
                        return response

                ##
                # Update data
                ##

                # prepare SQL data
                k_list.extend(['modify'])
                v_list.extend([str(datetime.now())])

                # insert the data
                query = """ UPDATE dyn_erl SET ({}) = %s WHERE (waf_id = %s AND id = %s) """.format(','.join(k_list))
                cursor.execute(query, (tuple(v_list), waf_id, rid,))

                # update ts
                set_hash(agent_ip, url, t, waf_id)

                # close database connection
                close_db_connection(connect, cursor)

                # response
                response = {'status': 'success'}
                return response

            except Exception as e:

                # response
                close_db_connection(connect, cursor)
                response = {'status': 'fail', 'description': str(e)}
                return response

        elif cmd == 'del':
            try:

                # convert all keys to lowercase
                data[cmd] = {k.lower(): v for k, v in data[cmd].items()}

                # extract value
                param = 'id'
                if data[cmd].get(param):
                    try:
                        rid = int(data[cmd][param])
                    except Exception:
                        close_db_connection(connect, cursor)
                        response = {'status': 'fail', 'description': 'Parameter "id" is not an integer'}
                        return response

                # if no ID in the request
                else:

                    # response
                    close_db_connection(connect, cursor)
                    err_msg = 'No ID received'
                    response = {'status': 'fail', 'description': err_msg}
                    return response

                try:

                    # search in the database
                    query = """ SELECT id from dyn_erl WHERE (id = %s) AND (waf_id = %s) """
                    cursor.execute(query, (rid, waf_id,))
                    res = cursor.fetchone()

                    # check result
                    if not res:

                        """
                        Row is not found, return response
                        """

                        # response
                        close_db_connection(connect, cursor)
                        err_msg = 'ID {} not found in database'.format(rid)
                        response = {'status': 'fail', 'description': err_msg}
                        return response

                    # search data
                    query = """ DELETE from dyn_erl WHERE (id = %s) AND (waf_id = %s) """
                    cursor.execute(query, (rid, waf_id,))

                    # update ts
                    set_hash(agent_ip, url, t, waf_id)

                    # close database connection
                    close_db_connection(connect, cursor)

                    # response
                    response = {'status': 'success'}
                    return response

                except Exception as e:

                    # response
                    close_db_connection(connect, cursor)
                    response = {'status': 'fail', 'description': str(e)}
                    return response

            except Exception as e:

                # response
                close_db_connection(connect, cursor)
                response = {'status': 'fail', 'description': str(e)}
                return response

    except Exception as e:
        log.error('An error occurred in set_dyn_erl_v2: {}'.format(e))


def get_dyn_wl(request, waf_id):
    try:

        # init
        get_extended = False
        resp_in_json = False

        # extract output format
        try:
            args = request.query_params.get('extended').lower()
            if args == 'yes':
                get_extended = True
        except Exception:
            pass

        # extract output format
        try:
            args = request.query_params.get('format').lower()
            if args == 'json':
                resp_in_json = True
        except Exception:
            pass

        ##
        # Extract data from the database
        ##

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # search in the database
        query = """ SELECT to_jsonb(array_agg(dyn_wl)) FROM dyn_wl WHERE (waf_id = %s AND active IS TRUE) """
        if get_extended or resp_in_json:
            query = """ SELECT to_jsonb(array_agg(dyn_wl)) FROM dyn_wl WHERE (waf_id = %s) """
        cursor.execute(query, (waf_id,))
        res = cursor.fetchall()

        # close database connection
        close_db_connection(connect, cursor)

        # processing result
        if not res or not res[0][0]:

            # init
            response = {'status': 'none'}

            # format result
            if resp_in_json:
                response = json.dumps(response)
            else:
                response = str(response)

            # response
            return response

        else:

            # init
            response = {'status': 'success'}

            # response in JSON
            if resp_in_json:

                # init
                r = []

                # processing columns
                for i in res[0][0]:

                    # rename column
                    i['rid'] = i['id']
                    del i['id']

                    # remove unused
                    del i['waf_id']

                    # idna normalization
                    i['domain'] = idna_normalization(i['domain'])

                    # update the list
                    r.append(i)

                # update response
                response['wl'] = r

                # convert to JSON
                response = json.dumps(response)

            else:

                # init
                response = str(response)

                # processing each row
                for row in res[0][0]:
                    try:

                        # init
                        r = 'WL'

                        # ID processing
                        r = r + ' ' + 'ID:' + str(row['rl_id'])

                        # Domain processing
                        if row['domain']:
                            r = r + ' domain=' + idna_normalization(str(row['domain']))

                        ##
                        # Z processing
                        ##

                        if not row['mz']:
                            row['mz'] = ''
                        if not row['extension']:
                            row['extension'] = ''

                        mz_ext = list(filter(None, (row['mz'].split('|') + row['extension'].split('|'))))
                        r = r + ' "' + 'Z:' + '|'.join(mz_ext) + '"'

                        # update result
                        if r:

                            if get_extended:
                                r = 'RID: ' + str(row['id']) + ' Active: ' + str(row['active']) + ' ' + r

                            response = response + '\n' + r

                    except Exception as e:
                        log.error('An error occurred while processing result from table "dyn_wl": {}'.format(e))

            # response
            return response

    except Exception as e:
        log.error('An error occurred in get_dyn_wl: {}'.format(e))


def get_dyn_wl_v2(waf_id):
    try:

        ##
        # Extract data from the database
        ##

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # search in the database
        query = """ SELECT to_jsonb(array_agg(dyn_wl)) FROM dyn_wl WHERE (waf_id = %s) """
        cursor.execute(query, (waf_id,))
        res = cursor.fetchall()

        # close database connection
        close_db_connection(connect, cursor)

        # processing result
        if not res or not res[0][0]:

            # init
            response = {'status': 'none'}

            # response
            return response

        else:

            # init
            response = {'status': 'success', 'wl': []}

            # processing columns
            for row in res[0][0]:

                # remove unused
                del row['waf_id']

                # idna normalization
                row['domain'] = idna_normalization(row['domain'])

                # modify normalization
                row['modify'] = int(datetime.fromisoformat(row['modify'].split('.')[0]).strftime('%s'))

                # update the list
                response['wl'].append(row)

            # response
            return response

    except Exception as e:
        log.error('An error occurred in get_dyn_wl_v2: {}'.format(e))


def set_dyn_wl(url, data, agent_ip, waf_id):
    try:

        ##
        # Check allowed command
        ##

        # init
        cmd = ''

        # search for allowed command in JSON
        for i in dyn_wl_erl_cmd:
            if i in data:
                cmd = i.lower()
                break

        # processing result
        if not cmd:

            # logging
            log.error('Required key missing in JSON from {} to {}'.format(agent_ip, url))

            # response
            response = {'status': 'fail', 'description': 'Required key missing'}
            return response

        ##
        # Processing JSON
        ##

        # init
        t = 'dyn_wl_ts'

        # create extended columns
        cols_ext = dyn_wl_cols + ['active']

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # processing command
        if cmd == 'add':

            # init
            k_list = []
            v_list = []

            # convert all keys to lowercase
            data[cmd] = {k.lower(): v for k, v in data[cmd].items()}

            # extract rl_id
            param = 'rl_id'
            if param not in data[cmd]:
                close_db_connection(connect, cursor)
                response = {'status': 'fail', 'description': 'No rl_id was received'}
                return response

            # processing data
            for k in cols_ext:

                if k in data[cmd]:

                    # convert to str and replace newline/multiple space to space
                    v = data[cmd][k]
                    if not isinstance(v, (bool, type(None))):
                        v = re.sub(r'\s+', ' ', str(v))

                    # value validations
                    if not dyn_wl_value_validation(agent_ip, k, v):
                        close_db_connection(connect, cursor)
                        response = {'status': 'fail', 'description': 'Incorrect value for {}'.format(k)}
                        return response

                    # update key/value lists
                    k_list.append(k)
                    v_list.append(v)

            # check JSON parameters
            if not k_list:
                close_db_connection(connect, cursor)
                response = {'status': 'fail', 'description': 'Parameters are missing in JSON'}
                return response

            ##
            # Create row
            ##

            try:

                # init
                k_list.extend(['waf_id', 'modify'])
                v_list.extend([waf_id, str(datetime.now())])

                # insert the data
                query = """ INSERT INTO dyn_wl ({}) VALUES %s RETURNING id """.format(','.join(k_list))
                cursor.execute(query, (tuple(v_list),))
                res = cursor.fetchone()

                # close database connection
                close_db_connection(connect, cursor)

                # update ts
                set_hash(agent_ip, url, t, waf_id)

                # set response
                if res:
                    # if ID was returned
                    response = {'status': 'success', 'rid': res[0]}
                else:
                    # if ID was not returned
                    response = {'status': 'fail', 'description': 'ID was not returned during the data saving process'}

                # response
                return response

            except Exception as e:

                # close database connection
                close_db_connection(connect, cursor)

                # set error message
                err_msg = str(e)

                # processing err_msg
                if 'duplicate key value' in err_msg:
                    err_msg = 'Duplicate key value'

                # response
                response = {'status': 'fail', 'description': err_msg}
                return response

        if cmd == 'upd':

            # init
            k_list = []
            v_list = []

            # convert all keys/values to lowercase
            data[cmd] = {k.lower(): v for k, v in data[cmd].items()}

            # extract value
            param = 'rid'
            if data[cmd].get(param):
                try:
                    rid = int(data[cmd][param])
                except Exception:
                    response = {'status': 'fail', 'description': 'Parameter "rid" is not an integer'}
                    return response

            # if no ID in the request
            else:
                close_db_connection(connect, cursor)
                err_msg = 'No ID received'
                response = {'status': 'fail', 'description': err_msg}
                return response

            ##
            # Extract RL ID
            ##

            query = """ SELECT id FROM dyn_wl WHERE (waf_id = %s AND id = %s) """
            cursor.execute(query, (waf_id, rid,))
            res = cursor.fetchone()

            # set type
            if not res:

                close_db_connection(connect, cursor)
                err_msg = 'ID {} not found in database'.format(rid)
                response = {'status': 'fail', 'description': err_msg}
                return response

            # processing data
            for k in cols_ext:

                if k in data[cmd]:

                    # convert to str and replace newline/multiple space to space
                    v = data[cmd][k]
                    if not isinstance(v, (bool, type(None))):
                        v = re.sub(r'\s+', ' ', str(v))

                    # value validations
                    if not dyn_wl_value_validation(agent_ip, k, v):
                        close_db_connection(connect, cursor)
                        response = {'status': 'fail', 'description': 'Incorrect value for {}'.format(k)}
                        return response

                    # update key/value lists
                    k_list.append(k)
                    v_list.append(v)

            # check JSON parameters
            if not k_list:
                close_db_connection(connect, cursor)
                response = {'status': 'fail', 'description': 'Parameters are missing in JSON'}
                return response

            ##
            # Update data in the database
            ##

            try:

                ##
                # Processing request data
                ##

                query = """ SELECT waf_id FROM dyn_wl WHERE (waf_id = %s AND id = %s) """
                cursor.execute(query, (waf_id, rid,))
                res = cursor.fetchone()

                # check result
                if not res:

                    """
                    Row is not found, return response
                    """

                    # close database connection
                    close_db_connection(connect, cursor)

                    # response
                    err_msg = 'ID {} not found in database'.format(rid)
                    response = {'status': 'fail', 'description': err_msg}
                    return response

                # init
                k_list.append('modify')
                v_list.append(str(datetime.now()))

                # update data
                query = """ UPDATE dyn_wl SET ({}) = %s WHERE (waf_id = %s AND id = %s) """.format(','.join(k_list))
                cursor.execute(query, (tuple(v_list), waf_id, rid,))

                # update ts
                set_hash(agent_ip, url, t, waf_id)

                # close database connection
                close_db_connection(connect, cursor)

                # response
                response = {'status': 'success'}
                return response

            except Exception as e:
                close_db_connection(connect, cursor)
                response = {'status': 'fail', 'description': str(e)}
                return response

        # processing command
        elif cmd == 'del':

            # convert all keys to lowercase
            data[cmd] = {k.lower(): v for k, v in data[cmd].items()}

            # extract value
            param = 'rid'
            if data[cmd].get(param):
                try:
                    rid = int(data[cmd][param])
                except Exception:
                    close_db_connection(connect, cursor)
                    response = {'status': 'fail', 'description': 'Parameter "rid" is not an integer'}
                    return response

            # if no ID in the request
            else:
                close_db_connection(connect, cursor)
                err_msg = 'No ID received'
                response = {'status': 'fail', 'description': err_msg}
                return response

            try:

                # search data
                query = """ SELECT waf_id FROM dyn_wl WHERE (waf_id = %s AND id = %s) """
                cursor.execute(query, (waf_id, rid,))
                res = cursor.fetchone()

                # check result
                if not res:

                    """
                    Row is not found, return response
                    """

                    # close database connection
                    close_db_connection(connect, cursor)

                    # response
                    err_msg = 'ID {} not found in database'.format(rid)
                    response = {'status': 'fail', 'description': err_msg}
                    return response

                query = """ DELETE FROM dyn_wl WHERE (id = %s) AND (waf_id = %s) """
                cursor.execute(query, (rid, waf_id,))

                # update ts
                set_hash(agent_ip, url, t, waf_id)

                # close database connection
                close_db_connection(connect, cursor)

                # response
                response = {'status': 'success'}
                return response

            except Exception as e:
                close_db_connection(connect, cursor)
                response = {'status': 'fail', 'description': str(e)}
                return response

    except Exception as e:
        log.error('An error occurred in set_dyn_wl: {}'.format(e))


def set_dyn_wl_v2(url, data, agent_ip, waf_id):
    try:

        ##
        # Check allowed command
        ##

        # init
        cmd = ''

        # search for allowed command in JSON
        for i in dyn_wl_erl_cmd:
            if i in data:
                cmd = i.lower()
                break

        # processing result
        if not cmd:

            # logging
            log.error('Required key missing in JSON from {} to {}'.format(agent_ip, url))

            # response
            response = {'status': 'fail', 'description': 'Required key missing'}
            return response

        ##
        # Processing JSON
        ##

        # init
        t = 'dyn_wl_ts'

        # create extended columns
        cols_ext = dyn_wl_cols + ['active']

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # processing command
        if cmd == 'add':

            # init
            k_list = []
            v_list = []

            # convert all keys to lowercase
            data[cmd] = {k.lower(): v for k, v in data[cmd].items()}

            # extract rl_id
            param = 'rl_id'
            if param not in data[cmd]:
                close_db_connection(connect, cursor)
                response = {'status': 'fail', 'description': 'No rl_id was received'}
                return response

            # processing data
            for k in cols_ext:

                if k in data[cmd]:

                    # v2 to v1 converter
                    if k in ['mz', 'extension']:

                        # value pre-validations
                        if not isinstance(data[cmd][k], list):
                            close_db_connection(connect, cursor)
                            response = {'status': 'fail', 'description': 'Incorrect value for {}'.format(k)}
                            return response

                        # list data validation
                        if any(x for x in data[cmd][k] if (not x) or (not isinstance(x, str))):
                            close_db_connection(connect, cursor)
                            response = {'status': 'fail', 'description': 'Incorrect value for {}'.format(k)}
                            return response

                        # data normalization to v1 format
                        data[cmd][k] = '|'.join(data[cmd][k])

                    # convert to str and replace newline/multiple space to space
                    v = data[cmd][k]
                    if not isinstance(v, (bool, type(None))):
                        v = re.sub(r'\s+', ' ', str(v))

                    # value validations
                    if not dyn_wl_value_validation(agent_ip, k, v):
                        close_db_connection(connect, cursor)
                        response = {'status': 'fail', 'description': 'Incorrect value for {}'.format(k)}
                        return response

                    # update key/value lists
                    k_list.append(k)
                    v_list.append(v)

            # check JSON parameters
            if not k_list:
                close_db_connection(connect, cursor)
                response = {'status': 'fail', 'description': 'Parameters are missing in JSON'}
                return response

            ##
            # Create row
            ##

            try:

                # init
                k_list.extend(['waf_id', 'modify'])
                v_list.extend([waf_id, str(datetime.now())])

                # insert the data
                query = """ INSERT INTO dyn_wl ({}) VALUES %s RETURNING id """.format(','.join(k_list))
                cursor.execute(query, (tuple(v_list),))
                res = cursor.fetchone()

                # update ts
                set_hash(agent_ip, url, t, waf_id)

                # set response
                if res:
                    # if ID was returned
                    response = {'status': 'success', 'id': res[0]}
                else:
                    # if ID was not returned
                    response = {'status': 'fail', 'description': 'ID was not returned during the data saving process'}

                # close database connection
                close_db_connection(connect, cursor)

                # response
                return response

            except Exception as e:

                # set error message
                err_msg = str(e)

                # processing err_msg
                if 'duplicate key value' in err_msg:
                    err_msg = 'Duplicate key value'

                # response
                close_db_connection(connect, cursor)
                response = {'status': 'fail', 'description': err_msg}
                return response

        if cmd == 'upd':

            # init
            k_list = []
            v_list = []

            # convert all keys/values to lowercase
            data[cmd] = {k.lower(): v for k, v in data[cmd].items()}

            # extract value
            param = 'id'
            if data[cmd].get(param):
                try:
                    rid = int(data[cmd][param])
                except Exception:
                    response = {'status': 'fail', 'description': 'Parameter "id" is not an integer'}
                    return response

            # if no ID in the request
            else:

                # response
                close_db_connection(connect, cursor)
                err_msg = 'No ID received'
                response = {'status': 'fail', 'description': err_msg}
                return response

            ##
            # Extract RL ID
            ##

            query = """ SELECT id FROM dyn_wl WHERE (waf_id = %s AND id = %s) """
            cursor.execute(query, (waf_id, rid,))
            res = cursor.fetchone()

            # set type
            if not res:

                # response
                close_db_connection(connect, cursor)
                err_msg = 'ID {} not found in database'.format(rid)
                response = {'status': 'fail', 'description': err_msg}
                return response

            # processing data
            for k in cols_ext:

                if k in data[cmd]:

                    # v2 to v1 converter
                    if k in ['mz', 'extension']:

                        # value pre-validations
                        if not isinstance(data[cmd][k], list):
                            close_db_connection(connect, cursor)
                            response = {'status': 'fail', 'description': 'Incorrect value for {}'.format(k)}
                            return response

                        # data normalization to v1 format
                        data[cmd][k] = '|'.join(data[cmd][k])

                    # convert to str and replace newline/multiple space to space
                    v = data[cmd][k]
                    if not isinstance(v, (bool, type(None))):
                        v = re.sub(r'\s+', ' ', str(v))

                    # value validations
                    if not dyn_wl_value_validation(agent_ip, k, v):
                        close_db_connection(connect, cursor)
                        response = {'status': 'fail', 'description': 'Incorrect value for {}'.format(k)}
                        return response

                    # update key/value lists
                    k_list.append(k)
                    v_list.append(v)

            # check JSON parameters
            if not k_list:
                close_db_connection(connect, cursor)
                response = {'status': 'fail', 'description': 'Parameters are missing in JSON'}
                return response

            ##
            # Update data in the database
            ##

            try:

                ##
                # Processing request data
                ##

                # search in the database
                query = """ SELECT waf_id FROM dyn_wl WHERE (waf_id = %s AND id = %s) """
                cursor.execute(query, (waf_id, rid,))
                res = cursor.fetchone()

                # check result
                if not res:

                    """
                    Row is not found, return response
                    """

                    # response
                    close_db_connection(connect, cursor)
                    err_msg = 'ID {} not found in database'.format(rid)
                    response = {'status': 'fail', 'description': err_msg}
                    return response

                # init
                k_list.append('modify')
                v_list.append(str(datetime.now()))

                # update data
                query = """ UPDATE dyn_wl SET ({}) = %s WHERE (waf_id = %s AND id = %s) """.format(','.join(k_list))
                cursor.execute(query, (tuple(v_list), waf_id, rid,))

                # update ts
                set_hash(agent_ip, url, t, waf_id)

                # close database connection
                close_db_connection(connect, cursor)

                # response
                response = {'status': 'success'}
                return response

            except Exception as e:

                # response
                close_db_connection(connect, cursor)
                response = {'status': 'fail', 'description': str(e)}
                return response

        # processing command
        elif cmd == 'del':

            # convert all keys to lowercase
            data[cmd] = {k.lower(): v for k, v in data[cmd].items()}

            # extract value
            param = 'id'
            if data[cmd].get(param):
                try:
                    rid = int(data[cmd][param])
                except Exception:
                    close_db_connection(connect, cursor)
                    response = {'status': 'fail', 'description': 'Parameter "id" is not an integer'}
                    return response

            # if no ID in the request
            else:

                # response
                close_db_connection(connect, cursor)
                err_msg = 'No ID received'
                response = {'status': 'fail', 'description': err_msg}
                return response

            try:

                # search data
                query = """ SELECT waf_id FROM dyn_wl WHERE (waf_id = %s AND id = %s) """
                cursor.execute(query, (waf_id, rid,))
                res = cursor.fetchone()

                # check result
                if not res:

                    """
                    Row is not found, return response
                    """

                    # response
                    close_db_connection(connect, cursor)
                    err_msg = 'ID {} not found in database'.format(rid)
                    response = {'status': 'fail', 'description': err_msg}
                    return response

                query = """ DELETE FROM dyn_wl WHERE (id = %s) AND (waf_id = %s) """
                cursor.execute(query, (rid, waf_id,))

                # update ts
                set_hash(agent_ip, url, t, waf_id)

                # close database connection
                close_db_connection(connect, cursor)

                # response
                response = {'status': 'success'}
                return response

            except Exception as e:

                # response
                close_db_connection(connect, cursor)
                response = {'status': 'fail', 'description': str(e)}
                return response

    except Exception as e:
        log.error('An error occurred in set_dyn_wl_v2: {}'.format(e))


def get_dyn_bl(request, agent_ip, waf_id):
    try:

        # init
        resp_in_json = False

        # extract output format
        try:
            args = request.query_params.get('format').lower()
            if args == 'json':
                resp_in_json = True
        except Exception:
            pass

        # search in the database
        query = """ SELECT active, bl FROM dyn_bl WHERE (waf_id = %s) """
        connect = init_db_connection()
        cursor = connect.cursor()
        cursor.execute(query, (waf_id,))
        res = cursor.fetchone()
        close_db_connection(connect, cursor)

        ##
        # Response in JSON format
        ##

        # processing result
        if not res:

            # init
            response = {'status': 'none'}

            # format result
            if resp_in_json:
                response = json.dumps(response)
            else:
                response = str(response)

            # response
            return response

        else:

            # init
            status = res[0]
            bl = literal_eval(res[1])

            # idna normalization
            bl = dyn_bl_idna_normalization(bl, agent_ip)

            if resp_in_json:

                # convert to JSON
                response = json.dumps(
                    {'status': 'success', 'active': status, 'bl': bl}
                )

            else:

                # init
                response = str({'status': 'success'})

                # processing BL
                if len(bl):
                    response = response + '\n' + '\n'.join(bl)

            # response
            return response

    except Exception as e:
        log.error('An error occurred in get_dyn_bl: {}'.format(e))


def get_dyn_bl_v2(agent_ip, waf_id):
    try:

        ##
        # Extract data from the database
        ##

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # search in the database
        query = """ SELECT active, extract(epoch from modify), bl FROM dyn_bl WHERE (waf_id = %s) """
        cursor.execute(query, (waf_id,))
        res = cursor.fetchone()

        # close database connection
        close_db_connection(connect, cursor)

        ##
        # Response in JSON format
        ##

        # processing result
        if not res:

            # init
            response = {'status': 'none'}

            # response
            return response

        else:

            # init
            status = res[0]
            modify = int(res[1])
            bl = literal_eval(res[2])

            # idna normalization
            bl = dyn_bl_idna_normalization(bl, agent_ip)

            # convert to JSON
            response = {
                'status': 'success',
                'modify': modify,
                'active': status,
                'bl': bl
            }

            # response
            return response

    except Exception as e:
        log.error('An error occurred in get_dyn_bl_v2: {}'.format(e))


def set_dyn_bl(url, data, agent_ip, waf_id):
    try:

        ##
        # Check allowed command
        ##

        # init
        cmd = ''

        # search for allowed command in JSON
        for i in settings_cmd:
            if i in data:
                cmd = i.lower()
                break

        # processing result
        if not cmd:

            # logging
            log.error('Required key missing in JSON from {} to {}'.format(agent_ip, url))

            # response
            response = {'status': 'fail', 'description': 'Required key missing'}
            return response

        ##
        # Processing JSON
        ##

        # create extended columns
        cols_ext = dyn_bl_cols + ['active']

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # processing command
        if cmd == 'set':

            # init
            k_list = []
            v_list = []

            # convert all keys/values to lowercase
            data[cmd] = {k.lower(): v for k, v in data[cmd].items()}

            # processing data
            for k in cols_ext:

                if k in data[cmd]:

                    # init
                    v = data[cmd][k]

                    # value validations
                    if not dyn_bl_value_validation(agent_ip, k, v):
                        close_db_connection(connect, cursor)
                        response = {'status': 'fail', 'description': 'Incorrect value for {}'.format(k)}
                        return response

                    # sort unique and replace quotes in list
                    if isinstance(data[cmd][k], list):
                        v = list(set(data[cmd][k]))
                        v = str(v).replace("'", '"')

                    # update key/value lists
                    k_list.append(k)
                    v_list.append(v)

            # check JSON parameters
            if not k_list:
                close_db_connection(connect, cursor)
                response = {'status': 'fail', 'description': 'Parameters are missing in JSON'}
                return response

            ##
            # Update data in the database
            ##

            try:

                ##
                # Processing request data
                ##

                # prepare SQL data
                k_list.extend(['waf_id', 'modify'])
                v_list.extend([waf_id, str(datetime.now())])

                # insert the data
                query = """ INSERT INTO dyn_bl ({}) VALUES %s ON CONFLICT (waf_id) DO UPDATE SET ({}) = %s """.format(
                    ','.join(k_list), ','.join(k_list)
                )
                cursor.execute(query, (tuple(v_list), tuple(v_list),))

                # close database connection
                close_db_connection(connect, cursor)

                # response
                response = {'status': 'success'}
                return response

            except Exception as e:
                close_db_connection(connect, cursor)
                response = {'status': 'fail', 'description': str(e)}
                return response

        # processing command
        elif cmd == 'del':
            try:

                # delete data
                query = """ DELETE FROM dyn_bl WHERE (waf_id = %s) """
                cursor.execute(query, (waf_id,))

                # close database connection
                close_db_connection(connect, cursor)

                # response
                response = {'status': 'success'}
                return response

            except Exception as e:
                close_db_connection(connect, cursor)
                response = {'status': 'fail', 'description': str(e)}
                return response

    except Exception as e:
        log.error('An error occurred in set_dyn_bl: {}'.format(e))


def set_dyn_bl_v2(url, data, agent_ip, waf_id):
    try:

        ##
        # Check allowed command
        ##

        # init
        cmd = ''

        # search for allowed command in JSON
        for i in settings_cmd:
            if i in data:
                cmd = i.lower()
                break

        # processing result
        if not cmd:

            # logging
            log.error('Required key missing in JSON from {} to {}'.format(agent_ip, url))

            # response
            response = {'status': 'fail', 'description': 'Required key missing'}
            return response

        ##
        # Processing JSON
        ##

        # create extended columns
        cols_ext = dyn_bl_cols + ['active']

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # processing command
        if cmd == 'set':

            # init
            k_list = []
            v_list = []

            # convert all keys/values to lowercase
            data[cmd] = {k.lower(): v for k, v in data[cmd].items()}

            # processing data
            for k in cols_ext:

                if k in data[cmd]:

                    # init
                    v = data[cmd][k]

                    # value validations
                    if not dyn_bl_value_validation(agent_ip, k, v):
                        close_db_connection(connect, cursor)
                        response = {'status': 'fail', 'description': 'Incorrect value for {}'.format(k)}
                        return response

                    # sort unique and replace quotes in list
                    if isinstance(data[cmd][k], list):
                        v = list(set(data[cmd][k]))
                        v = str(v).replace("'", '"')

                    # update key/value lists
                    k_list.append(k)
                    v_list.append(v)

            # check JSON parameters
            if not k_list:
                close_db_connection(connect, cursor)
                response = {'status': 'fail', 'description': 'Parameters are missing in JSON'}
                return response

            ##
            # Update data in the database
            ##

            try:

                ##
                # Processing request data
                ##

                # prepare SQL data
                k_list.extend(['waf_id', 'modify'])
                v_list.extend([waf_id, str(datetime.now())])

                # insert the data
                query = """ INSERT INTO dyn_bl ({}) VALUES %s ON CONFLICT (waf_id) DO UPDATE SET ({}) = %s """.format(
                    ','.join(k_list), ','.join(k_list)
                )
                cursor.execute(query, (tuple(v_list), tuple(v_list),))

                # close database connection
                close_db_connection(connect, cursor)

                # response
                response = {'status': 'success'}
                return response

            except Exception as e:

                # response
                close_db_connection(connect, cursor)
                response = {'status': 'fail', 'description': str(e)}
                return response

        # processing command
        elif cmd == 'del':
            try:

                # delete data
                query = """ DELETE FROM dyn_bl WHERE (waf_id = %s) """
                cursor.execute(query, (waf_id,))

                # close database connection
                close_db_connection(connect, cursor)

                # response
                response = {'status': 'success'}
                return response

            except Exception as e:

                # response
                close_db_connection(connect, cursor)
                response = {'status': 'fail', 'description': str(e)}
                return response

    except Exception as e:
        log.error('An error occurred in set_dyn_bl_v2: {}'.format(e))


def get_nws_target(waf_id):
    try:

        ##
        # Extract data from the database
        ##

        # init
        targets_list = []
        response = {'status': 'success', 'targets': targets_list}
        order_by = 'ORDER BY target'

        # search data
        query = """ SELECT to_jsonb(array_agg(nws_target {})) FROM nws_target WHERE (waf_id = %s) """.format(order_by)
        connect = init_db_connection()
        cursor = connect.cursor()
        cursor.execute(query, (waf_id,))
        res = cursor.fetchone()[0]
        close_db_connection(connect, cursor)

        # result processing
        if res:

            # update targets
            for r in res:

                # del columns
                del r['waf_id']

                # dict/list normalization
                for k in ['auth_params', 'exclude_modules']:
                    if k in r and r[k]:
                        r[k] = literal_eval(r[k])

                # update targets list
                targets_list.append(r)

            # update the response
            response['targets'] = targets_list

        # response
        return response

    except Exception as e:
        log.error('An error occurred in get_nws_target: {}'.format(e))


def set_nws_target(data, key, agent_ip, waf_id):
    try:

        ##
        # Processing request data
        ##

        # init
        k_list = []
        v_list = []

        # extract the target
        k = 'target'
        if k in data and isinstance(data[k], str) and data[k]:
            target = data[k]
        else:
            log.error('Required key "{}" is missing or incorrect'.format(k))
            response = {'status': 'fail', 'description': 'Required key missing'}
            return response

        # processing data
        for k in nws_target_cols:

            # auth_param processing
            if k == 'auth_params' and k in data:

                # convert all keys to lowercase and exclude unknown keys
                try:
                    data[k] = {ik.lower(): iv for ik, iv in data[k].items() if ik in nws_target_ap_cols}
                except Exception as e:
                    log.error(
                        'An error occurred while processing {}: {} ({})'
                        .format(k, e, data)
                    )

            # processing data
            if k in data:

                # value validations
                if not nws_target_value_validation(agent_ip, k, data[k]):
                    response = {'status': 'fail', 'description': 'Incorrect value for {}'.format(k)}
                    return response

                # replace quotes in list
                if isinstance(data[k], dict) or isinstance(data[k], list):
                    v = str(data[k])
                else:
                    v = data[k]

                # convert to str and replace newline/multiple space to space
                if isinstance(v, str):
                    v = re.sub(r'\s+', ' ', v)

                # update key/value lists
                k_list.append(k)
                v_list.append(v)

        ##
        # Save data to the database
        ##

        # check JSON parameters
        if not k_list:
            response = {'status': 'fail', 'description': 'Parameters are missing in JSON'}
            return response

        try:

            # init database connection
            connect = init_db_connection()
            cursor = connect.cursor()

            try:

                # insert row if it does not exist
                query = """ INSERT INTO nws_target (target, waf_id) VALUES (%s, %s) ON CONFLICT DO NOTHING """
                cursor.execute(query, (target, waf_id,))

                # update data
                query = """ UPDATE nws_target SET ({}) = %s WHERE (target = %s AND waf_id = %s) """.format(','.join(k_list))
                cursor.execute(query, (tuple(v_list), target, waf_id,))

            except Exception as e:

                # logging
                log.info('An error occurred while processing data for key "{}" in nws_target: {}'.format(get_log_key(key), e))

                # response
                close_db_connection(connect, cursor)
                response = {'status': 'fail', 'description': str(e)}
                return response

            # close database connection
            close_db_connection(connect, cursor)

            # response
            response = {'status': 'success'}
            return response

        except Exception as e:

            # response
            response = {'status': 'fail', 'description': str(e)}
            return response

    except Exception as e:
        log.error('An error occurred in set_nws_target: {}'.format(e))


def del_nws_target(data, key, waf_id):
    try:

        ##
        # Processing request data
        ##

        # extract the name
        k = 'target'
        if k in data and isinstance(data[k], str) and data[k]:
            target = data[k]
        else:
            log.error('Required key "{}" is missing or incorrect'.format(k))
            response = {'status': 'fail', 'description': 'Required key missing'}
            return response

        try:

            # init database connection
            connect = init_db_connection()
            cursor = connect.cursor()

            try:

                # delete data
                query = """ DELETE FROM nws_target WHERE (target = %s) AND (waf_id = %s) """
                cursor.execute(query, (target, waf_id,))

            except Exception as e:

                # logging
                log.info('An error occurred while processing data for key "{}" in nws_target: {}'.format(get_log_key(key), e))

                # response
                close_db_connection(connect, cursor)
                response = {'status': 'fail', 'description': str(e)}
                return response

            # close database connection
            close_db_connection(connect, cursor)

            # response
            response = {'status': 'success'}
            return response

        except Exception as e:

            # response
            response = {'status': 'fail', 'description': str(e)}
            return response

    except Exception as e:
        log.error('An error occurred in del_nws_target: {}'.format(e))
